@@ -1,10 +1,15 @@
+#[cfg(feature = "dtype-categorical")]
 use crate::datatypes::CategoricalChunked;
 use crate::prelude::*;
 use crate::utils::CustomIterTools;
-use arrow::array::{Array, BooleanArray, LargeListArray, LargeStringArray};
+use arrow::array::*;
 use std::convert::TryFrom;
+#[cfg(feature = "dtype-categorical")]
 use std::ops::Deref;
 
+type LargeStringArray = Utf8Array<i64>;
+type LargeListArray = ListArray<i64>;
+
 // If parallel feature is enable, then, activate the parallel module.
 #[cfg(feature = "parallel")]
 #[cfg_attr(docsrs, doc(cfg(feature = "parallel")))]
@@ -31,10 +36,16 @@ where
     type Item = Option<T::Native>;
     type IntoIter = Box<dyn PolarsIterator<Item = Self::Item> + 'a>;
     fn into_iter(self) -> Self::IntoIter {
-        Box::new(self.downcast_iter().flatten().trust_my_length(self.len()))
+        Box::new(
+            self.downcast_iter()
+                .flatten()
+                .map(|x| x.copied())
+                .trust_my_length(self.len()),
+        )
     }
 }
 
+#[cfg(feature = "dtype-categorical")]
 impl<'a> IntoIterator for &'a CategoricalChunked {
     type Item = Option<u32>;
     type IntoIter = Box<dyn PolarsIterator<Item = Self::Item> + 'a>;
@@ -117,7 +128,7 @@ impl BooleanChunked {
            + DoubleEndedIterator
            + TrustedLen {
         self.downcast_iter()
-            .map(|bool_arr| BoolIterNoNull::new(bool_arr))
+            .map(BoolIterNoNull::new)
             .flatten()
             .trust_my_length(self.len())
     }
@@ -195,7 +206,7 @@ impl Utf8Chunked {
            + DoubleEndedIterator
            + TrustedLen {
         self.downcast_iter()
-            .map(|arr| Utf8IterNoNull::new(arr))
+            .map(Utf8IterNoNull::new)
             .flatten()
             .trust_my_length(self.len())
     }
@@ -281,7 +292,7 @@ impl ListChunked {
            + DoubleEndedIterator
            + TrustedLen {
         self.downcast_iter()
-            .map(|arr| ListIterNoNull::new(arr))
+            .map(ListIterNoNull::new)
             .flatten()
             .trust_my_length(self.len())
     }
@@ -354,6 +365,7 @@ where
 
 impl<I> ExactSizeIterator for SomeIterator<I> where I: ExactSizeIterator {}
 
+#[cfg(feature = "dtype-categorical")]
 impl CategoricalChunked {
     #[allow(clippy::wrong_self_convention)]
     pub fn into_no_null_iter(
