@@ -1,10 +1,8 @@
 //! The typed heart of every Series column.
 use crate::prelude::*;
-use arrow::{array::*, bitmap::Bitmap, datatypes::TimeUnit};
+use arrow::{array::*, bitmap::Bitmap};
 use itertools::Itertools;
 use polars_arrow::prelude::ValueSize;
-use std::convert::TryFrom;
-use std::iter::{Copied, Map};
 use std::marker::PhantomData;
 use std::sync::Arc;
 
@@ -21,6 +19,13 @@ pub mod kernels;
 #[cfg(feature = "ndarray")]
 mod ndarray;
 
+mod bitwise;
+#[cfg(feature = "dtype-categorical")]
+pub(crate) mod categorical;
+#[cfg(feature = "object")]
+mod drop;
+pub(crate) mod list;
+pub(crate) mod logical;
 #[cfg(feature = "object")]
 #[cfg_attr(docsrs, doc(cfg(feature = "object")))]
 pub mod object;
@@ -30,19 +35,25 @@ mod random;
 #[cfg(feature = "strings")]
 #[cfg_attr(docsrs, doc(cfg(feature = "strings")))]
 pub mod strings;
-#[cfg(feature = "temporal")]
+#[cfg(any(
+    feature = "temporal",
+    feature = "dtype-datetime",
+    feature = "dtype-date"
+))]
 #[cfg_attr(docsrs, doc(cfg(feature = "temporal")))]
 pub mod temporal;
 mod trusted_len;
 pub mod upstream_traits;
-use arrow::array::Array;
-pub(crate) mod list;
+
 use polars_arrow::prelude::*;
 
-use crate::chunked_array::builder::categorical::RevMapping;
+#[cfg(feature = "dtype-categorical")]
+use crate::chunked_array::categorical::RevMapping;
 use crate::utils::{slice_offsets, CustomIterTools};
 use std::mem;
-use std::ops::{Deref, DerefMut};
+
+#[cfg(not(feature = "dtype-categorical"))]
+pub struct RevMapping {}
 
 pub type ChunkIdIter<'a> = std::iter::Map<std::slice::Iter<'a, ArrayRef>, fn(&ArrayRef) -> usize>;
 
@@ -77,9 +88,7 @@ pub type ChunkIdIter<'a> = std::iter::Map<std::slice::Iter<'a, ArrayRef>, fn(&Ar
 /// ```rust
 /// # use polars_core::prelude::*;
 /// fn apply_cosine_and_cast(ca: &Float32Chunked) -> Float64Chunked {
-///     ca.cast::<Float64Type>()
-///         .unwrap()
-///         .apply(|v| v.cos())
+///     ca.apply_cast_numeric(|v| v.cos() as f64)
 /// }
 /// ```
 ///
@@ -150,16 +159,15 @@ pub struct ChunkedArray<T> {
 }
 
 impl<T> ChunkedArray<T> {
-    #[cfg(feature = "asof_join")]
     pub(crate) fn is_sorted(&self) -> bool {
         self.bit_settings & 1 != 0
     }
 
-    #[cfg(feature = "asof_join")]
     pub(crate) fn is_sorted_reverse(&self) -> bool {
         self.bit_settings & 1 << 1 != 0
     }
 
+    /// Set the 'sorted' bit meta info.
     pub(crate) fn set_sorted(&mut self, reverse: bool) {
         if reverse {
             self.bit_settings |= 1 << 1
@@ -167,22 +175,18 @@ impl<T> ChunkedArray<T> {
             self.bit_settings |= 1
         }
     }
-    /// Get a reference to the mapping of categorical types to the string values.
-    pub fn get_categorical_map(&self) -> Option<&Arc<RevMapping>> {
-        self.categorical_map.as_ref()
-    }
 
     /// Get the index of the first non null value in this ChunkedArray.
     pub fn first_non_null(&self) -> Option<usize> {
         let mut offset = 0;
-        for (_, null_bitmap) in self.null_bits() {
-            if let Some(null_bitmap) = null_bitmap {
-                for (idx, is_valid) in null_bitmap.iter().enumerate() {
+        for validity in self.iter_validities() {
+            if let Some(validity) = validity {
+                for (idx, is_valid) in validity.iter().enumerate() {
                     if is_valid {
                         return Some(offset + idx);
                     }
                 }
-                offset += null_bitmap.len()
+                offset += validity.len()
             } else {
                 return Some(offset);
             }
@@ -191,15 +195,21 @@ impl<T> ChunkedArray<T> {
     }
 
     /// Get the buffer of bits representing null values
-    pub fn null_bits(&self) -> impl Iterator<Item = (usize, &Option<Bitmap>)> + '_ {
-        self.chunks
-            .iter()
-            .map(|arr| (arr.null_count(), arr.validity()))
+    #[inline]
+    pub fn iter_validities(&self) -> impl Iterator<Item = Option<&Bitmap>> + '_ {
+        self.chunks.iter().map(|arr| arr.validity())
+    }
+
+    #[inline]
+    /// Return if any the chunks in this `[ChunkedArray]` have a validity bitmap.
+    /// no bitmap means no null values.
+    pub fn has_validity(&self) -> bool {
+        self.iter_validities().any(|valid| valid.is_some())
     }
 
     /// Shrink the capacity of this array to fit it's length.
     pub fn shrink_to_fit(&mut self) {
-        self.chunks = vec![arrow::compute::concat::concatenate(
+        self.chunks = vec![arrow::compute::concatenate::concatenate(
             self.chunks.iter().map(|a| &**a).collect_vec().as_slice(),
         )
         .unwrap()
@@ -210,31 +220,26 @@ impl<T> ChunkedArray<T> {
     ///
     /// # Safety
     ///
-    /// This is unsafe as the dtype may be uncorrect and
+    /// This is unsafe as the dtype may be incorrect and
     /// is assumed to be correct in other safe code.
     pub(crate) unsafe fn unpack_series_matching_physical_type(
         &self,
         series: &Series,
-    ) -> Result<&ChunkedArray<T>> {
+    ) -> &ChunkedArray<T> {
         let series_trait = &**series;
         if self.dtype() == series.dtype() {
-            let ca = &*(series_trait as *const dyn SeriesTrait as *const ChunkedArray<T>);
-            Ok(ca)
+            &*(series_trait as *const dyn SeriesTrait as *const ChunkedArray<T>)
         } else {
             use DataType::*;
             match (self.dtype(), series.dtype()) {
-                (Int64, Date64) | (Int32, Date32) | (Int64, Duration(_)) | (Int64, Time64(_)) => {
-                    let ca = &*(series_trait as *const dyn SeriesTrait as *const ChunkedArray<T>);
-                    Ok(ca)
+                (Int64, Datetime) | (Int32, Date) => {
+                    &*(series_trait as *const dyn SeriesTrait as *const ChunkedArray<T>)
                 }
-                _ => Err(PolarsError::DataTypeMisMatch(
-                    format!(
-                        "cannot unpack series {:?} into matching type {:?}",
-                        series,
-                        self.dtype()
-                    )
-                    .into(),
-                )),
+                _ => panic!(
+                    "cannot unpack series {:?} into matching type {:?}",
+                    series,
+                    self.dtype()
+                ),
             }
         }
     }
@@ -244,9 +249,9 @@ impl<T> ChunkedArray<T> {
         if self.dtype() == series.dtype() {
             // Safety
             // dtype will be correct.
-            unsafe { self.unpack_series_matching_physical_type(series) }
+            Ok(unsafe { self.unpack_series_matching_physical_type(series) })
         } else {
-            Err(PolarsError::DataTypeMisMatch(
+            Err(PolarsError::SchemaMisMatch(
                 format!(
                     "cannot unpack series {:?} into matching type {:?}",
                     series,
@@ -279,7 +284,7 @@ impl<T> ChunkedArray<T> {
 
     /// Returns true if contains a single chunk and has no null values
     pub fn is_optimal_aligned(&self) -> bool {
-        self.chunks.len() == 1 && self.null_count() == 0
+        self.chunks.len() == 1 && !self.has_validity()
     }
 
     /// Count the null values.
@@ -313,7 +318,7 @@ impl<T> ChunkedArray<T> {
             self.chunks.push(other);
             Ok(())
         } else {
-            Err(PolarsError::DataTypeMisMatch(
+            Err(PolarsError::SchemaMisMatch(
                 format!(
                     "cannot append array of type {:?} in array of type {:?}",
                     other.data_type(),
@@ -372,7 +377,7 @@ impl<T> ChunkedArray<T> {
 
     /// Get a mask of the null values.
     pub fn is_null(&self) -> BooleanChunked {
-        if self.null_count() == 0 {
+        if !self.has_validity() {
             return BooleanChunked::full("is_null", false, self.len());
         }
         let chunks = self
@@ -381,18 +386,17 @@ impl<T> ChunkedArray<T> {
             .map(|arr| {
                 let bitmap = arr
                     .validity()
-                    .as_ref()
                     .map(|bitmap| !bitmap)
                     .unwrap_or_else(|| Bitmap::new_zeroed(arr.len()));
                 Arc::new(BooleanArray::from_data_default(bitmap, None)) as ArrayRef
             })
             .collect_vec();
-        BooleanChunked::new_from_chunks("is_null", chunks)
+        BooleanChunked::new_from_chunks(self.name(), chunks)
     }
 
     /// Get a mask of the valid values.
     pub fn is_not_null(&self) -> BooleanChunked {
-        if self.null_count() == 0 {
+        if !self.has_validity() {
             return BooleanChunked::full("is_not_null", true, self.len());
         }
         let chunks = self
@@ -401,12 +405,12 @@ impl<T> ChunkedArray<T> {
             .map(|arr| {
                 let bitmap = arr
                     .validity()
-                    .clone()
+                    .cloned()
                     .unwrap_or_else(|| !(&Bitmap::new_zeroed(arr.len())));
                 Arc::new(BooleanArray::from_data_default(bitmap, None)) as ArrayRef
             })
             .collect_vec();
-        BooleanChunked::new_from_chunks("is_not_null", chunks)
+        BooleanChunked::new_from_chunks(self.name(), chunks)
     }
 
     /// Get data type of ChunkedArray.
@@ -431,31 +435,6 @@ impl<T> ChunkedArray<T> {
         self.slice(-(len as i64), len)
     }
 
-    /// Append in place.
-    pub fn append(&mut self, other: &Self)
-    where
-        Self: std::marker::Sized,
-    {
-        if let (Some(rev_map_l), Some(rev_map_r)) = (
-            self.categorical_map.as_ref(),
-            other.categorical_map.as_ref(),
-        ) {
-            // first assertion checks if the global string cache is equal,
-            // the second checks if we append a slice from this array to self
-            if !rev_map_l.same_src(rev_map_r) && !Arc::ptr_eq(rev_map_l, rev_map_r) {
-                panic!("Appending categorical data can only be done if they are made under the same global string cache. \
-                Consider using a global string cache.")
-            }
-        }
-
-        // replace an empty array
-        if self.chunks.len() == 1 && self.is_empty() {
-            self.chunks = other.chunks.clone();
-        } else {
-            self.chunks.extend_from_slice(&other.chunks);
-        }
-    }
-
     /// Name of the ChunkedArray.
     pub fn name(&self) -> &str {
         self.field.name()
@@ -535,73 +514,11 @@ where
             bit_settings: 0,
         }
     }
-
-    #[inline]
-    unsafe fn arr_to_any_value(&self, arr: &dyn Array, idx: usize) -> AnyValue {
-        if arr.is_null(idx) {
-            return AnyValue::Null;
-        }
-
-        macro_rules! downcast_and_pack {
-            ($casttype:ident, $variant:ident) => {{
-                let arr = &*(arr as *const dyn Array as *const $casttype);
-                let v = arr.value(idx);
-                AnyValue::$variant(v)
-            }};
-        }
-        macro_rules! downcast {
-            ($casttype:ident) => {{
-                let arr = &*(arr as *const dyn Array as *const $casttype);
-                arr.value_unchecked(idx)
-            }};
-        }
-        // TODO: insert types
-        match T::get_dtype() {
-            DataType::Utf8 => downcast_and_pack!(LargeStringArray, Utf8),
-            DataType::Boolean => downcast_and_pack!(BooleanArray, Boolean),
-            DataType::UInt8 => downcast_and_pack!(UInt8Array, UInt8),
-            DataType::UInt16 => downcast_and_pack!(UInt16Array, UInt16),
-            DataType::UInt32 => downcast_and_pack!(UInt32Array, UInt32),
-            DataType::UInt64 => downcast_and_pack!(UInt64Array, UInt64),
-            DataType::Int8 => downcast_and_pack!(Int8Array, Int8),
-            DataType::Int16 => downcast_and_pack!(Int16Array, Int16),
-            DataType::Int32 => downcast_and_pack!(Int32Array, Int32),
-            DataType::Int64 => downcast_and_pack!(Int64Array, Int64),
-            DataType::Float32 => downcast_and_pack!(Float32Array, Float32),
-            DataType::Float64 => downcast_and_pack!(Float64Array, Float64),
-            DataType::Date32 => downcast_and_pack!(Int32Array, Date32),
-            DataType::Date64 => downcast_and_pack!(Int64Array, Date64),
-            DataType::Time64(TimeUnit::Nanosecond) => {
-                let v = downcast!(Int64Array);
-                AnyValue::Time64(v, TimeUnit::Nanosecond)
-            }
-            DataType::Duration(TimeUnit::Nanosecond) => {
-                let v = downcast!(Int64Array);
-                AnyValue::Duration(v, TimeUnit::Nanosecond)
-            }
-            DataType::Duration(TimeUnit::Millisecond) => {
-                let v = downcast!(Int64Array);
-                AnyValue::Duration(v, TimeUnit::Millisecond)
-            }
-            DataType::List(_) => {
-                let v: ArrayRef = downcast!(LargeListArray).into();
-                let s = Series::try_from(("", v));
-                AnyValue::List(s.unwrap())
-            }
-            DataType::Categorical => {
-                let v = downcast!(UInt32Array);
-                AnyValue::Utf8(self.categorical_map.as_ref().expect("should be set").get(v))
-            }
-            #[cfg(feature = "object")]
-            DataType::Object(_) => panic!("should not be here"),
-            _ => unimplemented!(),
-        }
-    }
 }
 
 impl<T> ChunkedArray<T>
 where
-    T: PolarsPrimitiveType,
+    T: PolarsNumericType,
 {
     /// Create a new ChunkedArray by taking ownership of the AlignedVec. This operation is zero copy.
     pub fn new_from_aligned_vec(name: &str, v: AlignedVec<T::Native>) -> Self {
@@ -661,7 +578,7 @@ where
 {
     /// Contiguous slice
     pub fn cont_slice(&self) -> Result<&[T::Native]> {
-        if self.chunks.len() == 1 && self.chunks[0].null_count() == 0 {
+        if self.chunks.len() == 1 && !self.chunks[0].has_validity() {
             Ok(self.downcast_iter().next().map(|arr| arr.values()).unwrap())
         } else {
             Err(PolarsError::NoSlice)
@@ -692,100 +609,6 @@ where
             .map(|v| *v)
             .trust_my_length(self.len())
     }
-
-    /// If [cont_slice](#method.cont_slice) is successful a closure is mapped over the elements.
-    ///
-    /// # Example
-    ///
-    /// ```
-    /// use polars_core::prelude::*;
-    /// fn multiply(ca: &UInt32Chunked) -> Result<Series> {
-    ///     let mapped = ca.map(|v| v * 2)?;
-    ///     Ok(mapped.collect())
-    /// }
-    /// ```
-    pub fn map<B, F>(&self, f: F) -> Result<Map<Copied<std::slice::Iter<T::Native>>, F>>
-    where
-        F: Fn(T::Native) -> B,
-    {
-        let slice = self.cont_slice()?;
-        Ok(slice.iter().copied().map(f))
-    }
-
-    /// If [cont_slice](#method.cont_slice) fails we can fallback on an iterator with null checks
-    /// and map a closure over the elements.
-    ///
-    /// # Example
-    ///
-    /// ```
-    /// use polars_core::prelude::*;
-    /// use itertools::Itertools;
-    /// fn multiply(ca: &UInt32Chunked) -> Series {
-    ///     let mapped_result = ca.map(|v| v * 2);
-    ///
-    ///     if let Ok(mapped) = mapped_result {
-    ///         mapped.collect()
-    ///     } else {
-    ///         ca
-    ///         .map_null_checks(|opt_v| opt_v.map(|v |v * 2)).collect()
-    ///     }
-    /// }
-    /// ```
-    pub fn map_null_checks<'a, B, F>(
-        &'a self,
-        f: F,
-    ) -> Map<Box<dyn PolarsIterator<Item = Option<T::Native>> + 'a>, F>
-    where
-        F: Fn(Option<T::Native>) -> B,
-    {
-        self.into_iter().map(f)
-    }
-
-    /// If [cont_slice](#method.cont_slice) is successful a closure can be applied as aggregation
-    ///
-    /// # Example
-    ///
-    /// ```
-    /// use polars_core::prelude::*;
-    /// fn compute_sum(ca: &UInt32Chunked) -> Result<u32> {
-    ///     ca.fold(0, |acc, value| acc + value)
-    /// }
-    /// ```
-    pub fn fold<F, B>(&self, init: B, f: F) -> Result<B>
-    where
-        F: Fn(B, T::Native) -> B,
-    {
-        let slice = self.cont_slice()?;
-        Ok(slice.iter().copied().fold(init, f))
-    }
-
-    /// If [cont_slice](#method.cont_slice) fails we can fallback on an iterator with null checks
-    /// and a closure for aggregation
-    ///
-    /// # Example
-    ///
-    /// ```
-    /// use polars_core::prelude::*;
-    /// fn compute_sum(ca: &UInt32Chunked) -> u32 {
-    ///     match ca.fold(0, |acc, value| acc + value) {
-    ///         // faster sum without null checks was successful
-    ///         Ok(sum) => sum,
-    ///         // Null values or multiple chunks in ChunkedArray, we need to do more bounds checking
-    ///         Err(_) => ca.fold_null_checks(0, |acc, opt_value| {
-    ///             match opt_value {
-    ///                 Some(v) => acc + v,
-    ///                 None => acc
-    ///             }
-    ///         })
-    ///     }
-    /// }
-    /// ```
-    pub fn fold_null_checks<F, B>(&self, init: B, f: F) -> B
-    where
-        F: Fn(B, Option<T::Native>) -> B,
-    {
-        self.into_iter().fold(init, f)
-    }
 }
 
 impl<T> Clone for ChunkedArray<T> {
@@ -806,37 +629,6 @@ impl<T> AsRef<ChunkedArray<T>> for ChunkedArray<T> {
     }
 }
 
-impl Deref for CategoricalChunked {
-    type Target = UInt32Chunked;
-
-    fn deref(&self) -> &Self::Target {
-        let ptr = self as *const CategoricalChunked;
-        let ptr = ptr as *const UInt32Chunked;
-        unsafe { &*ptr }
-    }
-}
-
-impl DerefMut for CategoricalChunked {
-    fn deref_mut(&mut self) -> &mut Self::Target {
-        let ptr = self as *mut CategoricalChunked;
-        let ptr = ptr as *mut UInt32Chunked;
-        unsafe { &mut *ptr }
-    }
-}
-
-impl From<UInt32Chunked> for CategoricalChunked {
-    fn from(ca: UInt32Chunked) -> Self {
-        ca.cast().unwrap()
-    }
-}
-
-impl CategoricalChunked {
-    fn set_state<T>(mut self, other: &ChunkedArray<T>) -> Self {
-        self.categorical_map = other.categorical_map.clone();
-        self
-    }
-}
-
 impl ValueSize for ListChunked {
     fn get_values_size(&self) -> usize {
         self.chunks
@@ -857,20 +649,20 @@ impl ListChunked {
     /// Get the inner data type of the list.
     pub fn inner_dtype(&self) -> DataType {
         match self.dtype() {
-            DataType::List(dt) => dt.into(),
+            DataType::List(dt) => *dt.clone(),
             _ => unreachable!(),
         }
     }
 }
 
-pub(crate) fn to_primitive<T: PolarsPrimitiveType>(
+pub(crate) fn to_primitive<T: PolarsNumericType>(
     values: AlignedVec<T::Native>,
     validity: Option<Bitmap>,
 ) -> PrimitiveArray<T::Native> {
     PrimitiveArray::from_data(T::get_dtype().to_arrow(), values.into(), validity)
 }
 
-pub(crate) fn to_array<T: PolarsPrimitiveType>(
+pub(crate) fn to_array<T: PolarsNumericType>(
     values: AlignedVec<T::Native>,
     validity: Option<Bitmap>,
 ) -> ArrayRef {
@@ -886,7 +678,6 @@ impl<T: PolarsNumericType> From<PrimitiveArray<T::Native>> for ChunkedArray<T> {
 #[cfg(test)]
 pub(crate) mod test {
     use crate::prelude::*;
-    use crate::reset_string_cache;
 
     pub(crate) fn get_chunked_array() -> Int32Chunked {
         ChunkedArray::new_from_slice("a", &[1, 2, 3])
@@ -973,8 +764,8 @@ pub(crate) mod test {
     #[test]
     fn cast() {
         let a = get_chunked_array();
-        let b = a.cast::<Int64Type>().unwrap();
-        assert_eq!(b.field.data_type(), &ArrowDataType::Int64)
+        let b = a.cast(&DataType::Int64).unwrap();
+        assert_eq!(b.dtype(), &ArrowDataType::Int64)
     }
 
     fn assert_slice_equal<T>(ca: &ChunkedArray<T>, eq: &[T::Native])
@@ -1063,13 +854,16 @@ pub(crate) mod test {
     }
 
     #[test]
+    #[cfg(feature = "dtype-categorical")]
     fn test_iter_categorical() {
+        use crate::reset_string_cache;
         use crate::SINGLE_LOCK;
         let _lock = SINGLE_LOCK.lock();
         reset_string_cache();
         let ca =
             Utf8Chunked::new_from_opt_slice("", &[Some("foo"), None, Some("bar"), Some("ham")]);
-        let ca = ca.cast::<CategoricalType>().unwrap();
+        let ca = ca.cast(&DataType::Categorical).unwrap();
+        let ca = ca.categorical().unwrap();
         let v: Vec<_> = ca.into_iter().collect();
         assert_eq!(v, &[Some(0), None, Some(1), Some(2)]);
     }
