@@ -3,38 +3,62 @@ use std::marker::Sized;
 
 use arrow::array::ArrayRef;
 
-use crate::chunked_array::builder::get_list_builder;
+pub use self::take::*;
 #[cfg(feature = "object")]
 use crate::chunked_array::object::ObjectType;
 use crate::prelude::*;
-use crate::utils::NoNull;
-
-pub use self::take::*;
+use arrow::buffer::Buffer;
+#[cfg(feature = "dtype-categorical")]
+use std::ops::Deref;
 
+#[cfg(feature = "abs")]
+mod abs;
 pub(crate) mod aggregate;
-pub(crate) mod any_value;
-pub(crate) mod apply;
-pub(crate) mod bit_repr;
+mod any_value;
+mod append;
+mod apply;
+mod bit_repr;
 pub(crate) mod chunkops;
 pub(crate) mod compare_inner;
-pub(crate) mod cum_agg;
+#[cfg(feature = "concat_str")]
+mod concat_str;
+#[cfg(feature = "cum_agg")]
+mod cum_agg;
 pub(crate) mod downcast;
 pub(crate) mod explode;
-pub(crate) mod fill_none;
-pub(crate) mod filter;
+mod fill_null;
+mod filter;
+pub mod full;
+#[cfg(feature = "interpolate")]
+mod interpolate;
 #[cfg(feature = "is_in")]
-pub(crate) mod is_in;
-pub(crate) mod peaks;
+mod is_in;
+mod peaks;
 #[cfg(feature = "repeat_by")]
-pub(crate) mod repeat_by;
-pub(crate) mod set;
-pub(crate) mod shift;
+mod repeat_by;
+mod reverse;
+pub(crate) mod rolling_window;
+mod set;
+mod shift;
 pub(crate) mod sort;
 pub(crate) mod take;
 pub(crate) mod unique;
-pub(crate) mod window;
 #[cfg(feature = "zip_with")]
-pub(crate) mod zip;
+pub mod zip;
+
+#[cfg(feature = "to_list")]
+pub trait ToList<T: PolarsDataType> {
+    fn to_list(&self) -> Result<ListChunked> {
+        Err(PolarsError::InvalidOperation(
+            format!("to_list not supported for dtype: {:?}", T::get_dtype()).into(),
+        ))
+    }
+}
+
+#[cfg(feature = "interpolate")]
+pub trait Interpolate {
+    fn interpolate(&self) -> Self;
+}
 
 #[cfg(feature = "reinterpret")]
 pub trait Reinterpret {
@@ -69,18 +93,23 @@ pub trait ChunkAnyValue {
     fn get_any_value(&self, index: usize) -> AnyValue;
 }
 
+#[cfg(feature = "cum_agg")]
 pub trait ChunkCumAgg<T> {
     /// Get an array with the cumulative max computed at every element
-    fn cum_max(&self, _reverse: bool) -> ChunkedArray<T> {
-        panic!("operation cum_max not supported for this dtype")
+    fn cummax(&self, _reverse: bool) -> ChunkedArray<T> {
+        panic!("operation cummax not supported for this dtype")
     }
     /// Get an array with the cumulative min computed at every element
-    fn cum_min(&self, _reverse: bool) -> ChunkedArray<T> {
-        panic!("operation cum_min not supported for this dtype")
+    fn cummin(&self, _reverse: bool) -> ChunkedArray<T> {
+        panic!("operation cummin not supported for this dtype")
     }
     /// Get an array with the cumulative sum computed at every element
-    fn cum_sum(&self, _reverse: bool) -> ChunkedArray<T> {
-        panic!("operation cum_sum not supported for this dtype")
+    fn cumsum(&self, _reverse: bool) -> ChunkedArray<T> {
+        panic!("operation cumsum not supported for this dtype")
+    }
+    /// Get an array with the cumulative product computed at every element
+    fn cumprod(&self, _reverse: bool) -> ChunkedArray<T> {
+        panic!("operation cumprod not supported for this dtype")
     }
 }
 
@@ -95,159 +124,21 @@ pub trait ChunkExplode {
     fn explode(&self) -> Result<Series> {
         self.explode_and_offsets().map(|t| t.0)
     }
-    fn explode_and_offsets(&self) -> Result<(Series, &[i64])>;
+    fn explode_and_offsets(&self) -> Result<(Series, Buffer<i64>)>;
 }
 
 pub trait ChunkBytes {
     fn to_byte_slices(&self) -> Vec<&[u8]>;
 }
 
-/// Rolling window functions
-pub trait ChunkWindow {
-    /// apply a rolling sum (moving sum) over the values in this array.
-    /// a window of length `window_size` will traverse the array. the values that fill this window
-    /// will (optionally) be multiplied with the weights given by the `weight` vector. the resulting
-    /// values will be aggregated to their sum.
-    ///
-    /// # Arguments
-    ///
-    /// * `window_size` - The length of the window.
-    /// * `weight` - An optional slice with the same length of the window that will be multiplied
-    ///              elementwise with the values in the window.
-    /// * `ignore_null` - Toggle behavior of aggregation regarding null values in the window.
-    ///                     `true` -> Null values will be ignored.
-    ///                     `false` -> Any Null in the window leads to a Null in the aggregation result.
-    fn rolling_sum(
-        &self,
-        _window_size: u32,
-        _weight: Option<&[f64]>,
-        _ignore_null: bool,
-        _min_periods: u32,
-    ) -> Result<Self>
+/// This differs from ChunkWindowCustom and ChunkWindow
+/// by not using a fold aggregator, but reusing a `Series` wrapper and calling `Series` aggregators.
+/// This likely is a bit slower than ChunkWindow
+#[cfg(feature = "rolling_window")]
+pub trait ChunkRollApply {
+    fn rolling_apply(&self, _window_size: usize, _f: &dyn Fn(&Series) -> Series) -> Result<Self>
     where
-        Self: std::marker::Sized,
-    {
-        Err(PolarsError::InvalidOperation(
-            "rolling sum not supported for this datatype".into(),
-        ))
-    }
-    /// Apply a rolling mean (moving mean) over the values in this array.
-    /// A window of length `window_size` will traverse the array. The values that fill this window
-    /// will (optionally) be multiplied with the weights given by the `weight` vector. The resulting
-    /// values will be aggregated to their mean.
-    ///
-    /// # Arguments
-    ///
-    /// * `window_size` - The length of the window.
-    /// * `weight` - An optional slice with the same length of the window that will be multiplied
-    ///              elementwise with the values in the window.
-    /// * `ignore_null` - Toggle behavior of aggregation regarding null values in the window.
-    ///                     `true` -> Null values will be ignored.
-    ///                     `false` -> Any Null in the window leads to a Null in the aggregation result.
-    /// * `min_periods` -  Amount of elements in the window that should be filled before computing a result.
-    fn rolling_mean(
-        &self,
-        _window_size: u32,
-        _weight: Option<&[f64]>,
-        _ignore_null: bool,
-        _min_periods: u32,
-    ) -> Result<Self>
-    where
-        Self: std::marker::Sized,
-    {
-        Err(PolarsError::InvalidOperation(
-            "rolling mean not supported for this datatype".into(),
-        ))
-    }
-
-    /// Apply a rolling min (moving min) over the values in this array.
-    /// A window of length `window_size` will traverse the array. The values that fill this window
-    /// will (optionally) be multiplied with the weights given by the `weight` vector. The resulting
-    /// values will be aggregated to their min.
-    ///
-    /// # Arguments
-    ///
-    /// * `window_size` - The length of the window.
-    /// * `weight` - An optional slice with the same length of the window that will be multiplied
-    ///              elementwise with the values in the window.
-    /// * `ignore_null` - Toggle behavior of aggregation regarding null values in the window.
-    ///                     `true` -> Null values will be ignored.
-    ///                     `false` -> Any Null in the window leads to a Null in the aggregation result.
-    /// * `min_periods` -  Amount of elements in the window that should be filled before computing a result.
-    fn rolling_min(
-        &self,
-        _window_size: u32,
-        _weight: Option<&[f64]>,
-        _ignore_null: bool,
-        _min_periods: u32,
-    ) -> Result<Self>
-    where
-        Self: std::marker::Sized,
-    {
-        Err(PolarsError::InvalidOperation(
-            "rolling mean not supported for this datatype".into(),
-        ))
-    }
-
-    /// Apply a rolling max (moving max) over the values in this array.
-    /// A window of length `window_size` will traverse the array. The values that fill this window
-    /// will (optionally) be multiplied with the weights given by the `weight` vector. The resulting
-    /// values will be aggregated to their max.
-    ///
-    /// # Arguments
-    ///
-    /// * `window_size` - The length of the window.
-    /// * `weight` - An optional slice with the same length of the window that will be multiplied
-    ///              elementwise with the values in the window.
-    /// * `ignore_null` - Toggle behavior of aggregation regarding null values in the window.
-    ///                     `true` -> Null values will be ignored.
-    ///                     `false` -> Any Null in the window leads to a Null in the aggregation result.
-    /// * `min_periods` -  Amount of elements in the window that should be filled before computing a result.
-    fn rolling_max(
-        &self,
-        _window_size: u32,
-        _weight: Option<&[f64]>,
-        _ignore_null: bool,
-        _min_periods: u32,
-    ) -> Result<Self>
-    where
-        Self: std::marker::Sized,
-    {
-        Err(PolarsError::InvalidOperation(
-            "rolling mean not supported for this datatype".into(),
-        ))
-    }
-}
-
-/// Custom rolling window functions
-pub trait ChunkWindowCustom<T> {
-    /// Apply a rolling aggregation over the values in this array.
-    ///
-    /// A window of length `window_size` will traverse the array. The values that fill this window
-    /// will (optionally) be multiplied with the weights given by the `weight` vector. The resulting
-    /// values will be aggregated to their max.
-    ///
-    /// You can pass a custom closure that will be used in the `fold` operation to aggregate the window.
-    /// The closure/fn of type `Fn(Option<T>, Option<T>) -> Option<T>` takes an `accumulator` and
-    /// a `value` as argument.
-    ///
-    /// # Arguments
-    ///
-    /// * `window_size` - The length of the window.
-    /// * `weight` - An optional slice with the same length of the window that will be multiplied
-    ///              elementwise with the values in the window.
-    /// * `min_periods` -  Amount of elements in the window that should be filled before computing a result.
-    fn rolling_custom<F>(
-        &self,
-        _window_size: u32,
-        _weight: Option<&[f64]>,
-        _fold_fn: F,
-        _init_fold: InitFold,
-        _min_periods: u32,
-    ) -> Result<Self>
-    where
-        F: Fn(Option<T>, Option<T>) -> Option<T> + Copy,
-        Self: std::marker::Sized,
+        Self: Sized,
     {
         Err(PolarsError::InvalidOperation(
             "rolling mean not supported for this datatype".into(),
@@ -261,7 +152,8 @@ pub trait TakeRandom {
 
     /// Get a nullable value by index.
     ///
-    /// Out of bounds access doesn't Error but will return a Null value
+    /// # Panics
+    /// Panics if `index >= self.len()`
     fn get(&self, index: usize) -> Option<Self::Item>;
 
     /// Get a value by index and ignore the null bit.
@@ -282,7 +174,8 @@ pub trait TakeRandomUtf8 {
 
     /// Get a nullable value by index.
     ///
-    /// Out of bounds access doesn't Error but will return a Null value
+    /// # Panics
+    /// Panics if `index >= self.len()`
     fn get(self, index: usize) -> Option<Self::Item>;
 
     /// Get a value by index and ignore the null bit.
@@ -331,7 +224,7 @@ pub trait ChunkSet<'a, A, B> {
     ///
     /// ```rust
     /// # use polars_core::prelude::*;
-    /// let ca = Int32Chunked::new_from_slice("a", &[1, 2, 3]);
+    /// let ca = UInt32Chunked::new_from_slice("a", &[1, 2, 3]);
     /// let new = ca.set_at_idx(vec![0, 1], Some(10)).unwrap();
     ///
     /// assert_eq!(Vec::from(&new), &[Some(10), Some(10), Some(3)]);
@@ -380,7 +273,7 @@ pub trait ChunkSet<'a, A, B> {
     ///
     /// ```rust
     /// # use polars_core::prelude::*;
-    /// let ca = Int32Chunked::new_from_slice("a", &[1, 2, 3]);
+    /// let ca = UInt32Chunked::new_from_slice("a", &[1, 2, 3]);
     /// let mask = BooleanChunked::new_from_slice("mask", &[false, true, false]);
     /// let new = ca.set_with(&mask, |opt_v| opt_v.map(
     ///     |v| v * 2
@@ -395,13 +288,8 @@ pub trait ChunkSet<'a, A, B> {
 
 /// Cast `ChunkedArray<T>` to `ChunkedArray<N>`
 pub trait ChunkCast {
-    /// Cast `ChunkedArray<T>` to `ChunkedArray<N>`
-    fn cast<N>(&self) -> Result<ChunkedArray<N>>
-    where
-        N: PolarsDataType,
-        Self: Sized;
-
-    fn cast_with_dtype(&self, data_type: &DataType) -> Result<Series>;
+    /// Cast a `[ChunkedArray]` to `[DataType]`
+    fn cast(&self, data_type: &DataType) -> Result<Series>;
 }
 
 /// Fastest way to do elementwise operations on a ChunkedArray<T> when the operation is cheaper than
@@ -439,6 +327,11 @@ pub trait ChunkApply<'a, A, B> {
     where
         F: Fn(A) -> B + Copy;
 
+    fn try_apply<F>(&'a self, f: F) -> Result<Self>
+    where
+        F: Fn(A) -> Result<B> + Copy,
+        Self: Sized;
+
     /// Apply a closure elementwise including null values.
     fn apply_on_opt<F>(&'a self, f: F) -> Self
     where
@@ -521,7 +414,7 @@ pub trait ChunkVar<T> {
 /// fn filter_all_ones(df: &DataFrame) -> Result<DataFrame> {
 ///     let mask = df
 ///     .column("column_a")?
-///     .eq(1);
+///     .equal(1);
 ///
 ///     df.filter(&mask)
 /// }
@@ -531,10 +424,10 @@ pub trait ChunkCompare<Rhs> {
     fn eq_missing(&self, rhs: Rhs) -> BooleanChunked;
 
     /// Check for equality.
-    fn eq(&self, rhs: Rhs) -> BooleanChunked;
+    fn equal(&self, rhs: Rhs) -> BooleanChunked;
 
     /// Check for inequality.
-    fn neq(&self, rhs: Rhs) -> BooleanChunked;
+    fn not_equal(&self, rhs: Rhs) -> BooleanChunked;
 
     /// Greater than comparison.
     fn gt(&self, rhs: Rhs) -> BooleanChunked;
@@ -603,8 +496,19 @@ pub trait ToDummies<T>: ChunkUnique<T> {
     }
 }
 
+#[derive(Default)]
+pub struct SortOptions {
+    descending: bool,
+    nulls_last: bool,
+}
+
 /// Sort operations on `ChunkedArray`.
 pub trait ChunkSort<T> {
+    #[allow(unused_variables)]
+    fn sort_with(&self, options: SortOptions) -> ChunkedArray<T> {
+        unimplemented!()
+    }
+
     /// Returned a sorted `ChunkedArray`.
     fn sort(&self, reverse: bool) -> ChunkedArray<T>;
 
@@ -623,7 +527,7 @@ pub trait ChunkSort<T> {
 }
 
 #[derive(Copy, Clone, Debug)]
-pub enum FillNoneStrategy {
+pub enum FillNullStrategy {
     /// previous value in array
     Backward,
     /// next value in array
@@ -645,21 +549,21 @@ pub enum FillNoneStrategy {
 }
 
 /// Replace None values with various strategies
-pub trait ChunkFillNone {
+pub trait ChunkFillNull {
     /// Replace None values with one of the following strategies:
     /// * Forward fill (replace None with the previous value)
     /// * Backward fill (replace None with the next value)
     /// * Mean fill (replace None with the mean of the whole array)
     /// * Min fill (replace None with the minimum of the whole array)
     /// * Max fill (replace None with the maximum of the whole array)
-    fn fill_none(&self, strategy: FillNoneStrategy) -> Result<Self>
+    fn fill_null(&self, strategy: FillNullStrategy) -> Result<Self>
     where
         Self: Sized;
 }
 /// Replace None values with a value
-pub trait ChunkFillNoneValue<T> {
+pub trait ChunkFillNullValue<T> {
     /// Replace None values with a give value `T`.
-    fn fill_none_with_value(&self, value: T) -> Result<Self>
+    fn fill_null_with_values(&self, value: T) -> Result<Self>
     where
         Self: Sized;
 }
@@ -678,150 +582,12 @@ pub trait ChunkFullNull {
         Self: std::marker::Sized;
 }
 
-impl<T> ChunkFull<T::Native> for ChunkedArray<T>
-where
-    T: PolarsPrimitiveType,
-{
-    fn full(name: &str, value: T::Native, length: usize) -> Self
-    where
-        T::Native: Copy,
-    {
-        let mut ca = (0..length)
-            .map(|_| value)
-            .collect::<NoNull<ChunkedArray<T>>>()
-            .into_inner();
-        ca.rename(name);
-        ca
-    }
-}
-
-impl<T> ChunkFullNull for ChunkedArray<T>
-where
-    T: PolarsPrimitiveType,
-{
-    fn full_null(name: &str, length: usize) -> Self {
-        let mut ca = (0..length).map(|_| None).collect::<Self>();
-        ca.rename(name);
-        ca
-    }
-}
-impl ChunkFull<bool> for BooleanChunked {
-    fn full(name: &str, value: bool, length: usize) -> Self {
-        let mut ca = (0..length).map(|_| value).collect::<BooleanChunked>();
-        ca.rename(name);
-        ca
-    }
-}
-
-impl ChunkFullNull for BooleanChunked {
-    fn full_null(name: &str, length: usize) -> Self {
-        let mut ca = (0..length).map(|_| None).collect::<Self>();
-        ca.rename(name);
-        ca
-    }
-}
-
-impl<'a> ChunkFull<&'a str> for Utf8Chunked {
-    fn full(name: &str, value: &'a str, length: usize) -> Self {
-        let mut builder = Utf8ChunkedBuilder::new(name, length, length * value.len());
-
-        for _ in 0..length {
-            builder.append_value(value);
-        }
-        builder.finish()
-    }
-}
-
-impl ChunkFullNull for Utf8Chunked {
-    fn full_null(name: &str, length: usize) -> Self {
-        let mut ca = (0..length)
-            .map::<Option<String>, _>(|_| None)
-            .collect::<Self>();
-        ca.rename(name);
-        ca
-    }
-}
-
-impl ChunkFull<&Series> for ListChunked {
-    fn full(name: &str, value: &Series, length: usize) -> ListChunked {
-        let mut builder = get_list_builder(value.dtype(), value.len() * length, length, name);
-        for _ in 0..length {
-            builder.append_series(value)
-        }
-        builder.finish()
-    }
-}
-
-impl ChunkFullNull for ListChunked {
-    fn full_null(name: &str, length: usize) -> ListChunked {
-        let mut ca = (0..length)
-            .map::<Option<Series>, _>(|_| None)
-            .collect::<Self>();
-        ca.rename(name);
-        ca
-    }
-}
-
-#[cfg(feature = "object")]
-impl<T: PolarsObject> ChunkFullNull for ObjectChunked<T> {
-    fn full_null(name: &str, length: usize) -> ObjectChunked<T> {
-        let mut ca: Self = (0..length).map(|_| None).collect();
-        ca.rename(name);
-        ca
-    }
-}
-
 /// Reverse a ChunkedArray<T>
 pub trait ChunkReverse<T> {
     /// Return a reversed version of this array.
     fn reverse(&self) -> ChunkedArray<T>;
 }
 
-impl<T> ChunkReverse<T> for ChunkedArray<T>
-where
-    T: PolarsNumericType,
-    ChunkedArray<T>: ChunkOps,
-{
-    fn reverse(&self) -> ChunkedArray<T> {
-        if let Ok(slice) = self.cont_slice() {
-            let ca: NoNull<ChunkedArray<T>> = slice.iter().rev().copied().collect();
-            let mut ca = ca.into_inner();
-            ca.rename(self.name());
-            ca
-        } else {
-            self.into_iter().rev().collect()
-        }
-    }
-}
-
-impl ChunkReverse<CategoricalType> for CategoricalChunked {
-    fn reverse(&self) -> ChunkedArray<CategoricalType> {
-        self.cast::<UInt32Type>().unwrap().reverse().cast().unwrap()
-    }
-}
-
-macro_rules! impl_reverse {
-    ($arrow_type:ident, $ca_type:ident) => {
-        impl ChunkReverse<$arrow_type> for $ca_type {
-            fn reverse(&self) -> Self {
-                unsafe { self.take_unchecked((0..self.len()).rev().into()) }
-            }
-        }
-    };
-}
-
-impl_reverse!(BooleanType, BooleanChunked);
-impl_reverse!(Utf8Type, Utf8Chunked);
-impl_reverse!(ListType, ListChunked);
-#[cfg(feature = "object")]
-impl<T: PolarsObject> ChunkReverse<ObjectType<T>> for ObjectChunked<T> {
-    fn reverse(&self) -> Self {
-        // Safety
-        // we we know we don't get out of bounds
-        unsafe { self.take_unchecked((0..self.len()).rev().into()) }
-    }
-}
-
 /// Filter values by a boolean mask.
 pub trait ChunkFilter<T> {
     /// Filter values in the ChunkedArray with a boolean mask.
@@ -858,7 +624,7 @@ macro_rules! impl_chunk_expand {
 impl<T> ChunkExpandAtIndex<T> for ChunkedArray<T>
 where
     ChunkedArray<T>: ChunkFull<T::Native> + TakeRandom<Item = T::Native>,
-    T: PolarsPrimitiveType,
+    T: PolarsNumericType,
 {
     fn expand_at_index(&self, index: usize, length: usize) -> ChunkedArray<T> {
         impl_chunk_expand!(self, length, index)
@@ -877,26 +643,32 @@ impl ChunkExpandAtIndex<Utf8Type> for Utf8Chunked {
     }
 }
 
+#[cfg(feature = "dtype-categorical")]
 impl ChunkExpandAtIndex<CategoricalType> for CategoricalChunked {
     fn expand_at_index(&self, index: usize, length: usize) -> CategoricalChunked {
-        self.cast::<UInt32Type>()
-            .unwrap()
-            .expand_at_index(index, length)
-            .cast()
-            .unwrap()
+        let ca: CategoricalChunked = self.deref().expand_at_index(index, length).into();
+        ca.set_state(self)
     }
 }
 
 impl ChunkExpandAtIndex<ListType> for ListChunked {
-    fn expand_at_index(&self, _index: usize, _length: usize) -> ListChunked {
-        unimplemented!()
+    fn expand_at_index(&self, index: usize, length: usize) -> ListChunked {
+        let opt_val = self.get(index);
+        match opt_val {
+            Some(val) => ListChunked::full(self.name(), &val, length),
+            None => ListChunked::full_null_with_dtype(self.name(), length, &self.inner_dtype()),
+        }
     }
 }
 
 #[cfg(feature = "object")]
-impl<T> ChunkExpandAtIndex<ObjectType<T>> for ObjectChunked<T> {
-    fn expand_at_index(&self, _index: usize, _length: usize) -> ObjectChunked<T> {
-        todo!()
+impl<T: PolarsObject> ChunkExpandAtIndex<ObjectType<T>> for ObjectChunked<T> {
+    fn expand_at_index(&self, index: usize, length: usize) -> ObjectChunked<T> {
+        let opt_val = self.get(index);
+        match opt_val {
+            Some(val) => ObjectChunked::<T>::full(self.name(), val.clone(), length),
+            None => ObjectChunked::<T>::full_null(self.name(), length),
+        }
     }
 }
 
@@ -919,7 +691,7 @@ pub trait ChunkZip<T> {
 }
 
 /// Apply kernels on the arrow array chunks in a ChunkedArray.
-pub trait ChunkApplyKernel<A> {
+pub trait ChunkApplyKernel<A: Array> {
     /// Apply kernel and return result as a new ChunkedArray.
     fn apply_kernel<F>(&self, f: F) -> Self
     where
@@ -979,6 +751,7 @@ pub trait RepeatBy {
 
 #[cfg(feature = "is_first")]
 #[cfg_attr(docsrs, doc(cfg(feature = "is_first")))]
+/// Mask the first unique values as `true`
 pub trait IsFirst<T: PolarsDataType> {
     fn is_first(&self) -> Result<BooleanChunked> {
         Err(PolarsError::InvalidOperation(
@@ -989,6 +762,7 @@ pub trait IsFirst<T: PolarsDataType> {
 
 #[cfg(feature = "is_first")]
 #[cfg_attr(docsrs, doc(cfg(feature = "is_first")))]
+/// Mask the last unique values as `true`
 pub trait IsLast<T: PolarsDataType> {
     fn is_last(&self) -> Result<BooleanChunked> {
         Err(PolarsError::InvalidOperation(
@@ -996,3 +770,14 @@ pub trait IsLast<T: PolarsDataType> {
         ))
     }
 }
+
+#[cfg(feature = "concat_str")]
+#[cfg_attr(docsrs, doc(cfg(feature = "concat_str")))]
+/// Concat the values into a string array.
+pub trait StrConcat {
+    /// Concat the values into a string array.
+    /// # Arguments
+    ///
+    /// * `delimiter` - A string that will act as delimiter between values.
+    fn str_concat(&self, delimiter: &str) -> Utf8Chunked;
+}
