@@ -1,12 +1,14 @@
 use crate::prelude::compare_inner::PartialOrdInner;
 use crate::prelude::*;
 use crate::utils::{CustomIterTools, NoNull};
+use arrow::{bitmap::MutableBitmap, buffer::Buffer};
 use itertools::Itertools;
 use polars_arrow::prelude::ValueSize;
 use polars_arrow::trusted_len::PushUnchecked;
 use rayon::prelude::*;
 use std::cmp::Ordering;
 use std::iter::FromIterator;
+#[cfg(feature = "dtype-categorical")]
 use std::ops::{Deref, DerefMut};
 
 /// Sort with null values, to reverse, swap the arguments.
@@ -59,7 +61,6 @@ fn order_reverse_null<T: PartialOrd>(a: &Option<T>, b: &Option<T>) -> Ordering {
 
 fn sort_branch<T, Fd, Fr>(
     slice: &mut [T],
-    sort_parallel: bool,
     reverse: bool,
     default_order_fn: Fd,
     reverse_order_fn: Fr,
@@ -68,17 +69,14 @@ fn sort_branch<T, Fd, Fr>(
     Fd: FnMut(&T, &T) -> Ordering + for<'r, 's> Fn(&'r T, &'s T) -> Ordering + Sync,
     Fr: FnMut(&T, &T) -> Ordering + for<'r, 's> Fn(&'r T, &'s T) -> Ordering + Sync,
 {
-    match (sort_parallel, reverse) {
-        (true, true) => slice.par_sort_unstable_by(reverse_order_fn),
-        (true, false) => slice.par_sort_unstable_by(default_order_fn),
-        (false, true) => slice.sort_unstable_by(reverse_order_fn),
-        (false, false) => slice.sort_unstable_by(default_order_fn),
+    match reverse {
+        true => slice.par_sort_unstable_by(reverse_order_fn),
+        false => slice.par_sort_unstable_by(default_order_fn),
     }
 }
 
 fn argsort_branch<T, Fd, Fr>(
     slice: &mut [T],
-    sort_parallel: bool,
     reverse: bool,
     default_order_fn: Fd,
     reverse_order_fn: Fr,
@@ -87,26 +85,14 @@ fn argsort_branch<T, Fd, Fr>(
     Fd: FnMut(&T, &T) -> Ordering + for<'r, 's> Fn(&'r T, &'s T) -> Ordering + Sync,
     Fr: FnMut(&T, &T) -> Ordering + for<'r, 's> Fn(&'r T, &'s T) -> Ordering + Sync,
 {
-    match (sort_parallel, reverse) {
-        (true, true) => slice.par_sort_by(reverse_order_fn),
-        (true, false) => slice.par_sort_by(default_order_fn),
-        (false, true) => slice.sort_by(reverse_order_fn),
-        (false, false) => slice.sort_by(default_order_fn),
+    match reverse {
+        true => slice.par_sort_by(reverse_order_fn),
+        false => slice.par_sort_by(default_order_fn),
     }
 }
 
-/// If the sort should be ran parallel or not.
-fn sort_parallel<T>(ca: &ChunkedArray<T>) -> bool {
-    ca.len()
-        > std::env::var("POLARS_PAR_SORT_BOUND")
-            .map(|v| v.parse::<usize>().expect("could not parse"))
-            .unwrap_or(1000000)
-}
-
 macro_rules! argsort {
     ($self:expr, $reverse:expr) => {{
-        let sort_parallel = sort_parallel($self);
-
         let mut vals = Vec::with_capacity($self.len());
         let mut count: u32 = 0;
         $self.downcast_iter().for_each(|arr| {
@@ -120,7 +106,6 @@ macro_rules! argsort {
 
         argsort_branch(
             vals.as_mut_slice(),
-            sort_parallel,
             $reverse,
             |(_, a), (_, b)| order_default_null(a, b),
             |(_, a), (_, b)| order_reverse_null(a, b),
@@ -145,9 +130,9 @@ where
 
     let mut offset = 0;
     ca.downcast_iter().for_each(|arr| {
-        let values = arr.values();
-        let len = arr.len();
-        (vals_slice[offset..len]).copy_from_slice(values);
+        let values = arr.values().as_slice();
+        let len = values.len();
+        (vals_slice[offset..offset + len]).copy_from_slice(values);
         offset += len;
     });
     vals
@@ -156,52 +141,114 @@ where
 impl<T> ChunkSort<T> for ChunkedArray<T>
 where
     T: PolarsNumericType,
-    T::Native: std::cmp::PartialOrd,
+    T::Native: Default,
 {
-    fn sort(&self, reverse: bool) -> ChunkedArray<T> {
-        let sort_parallel = sort_parallel(self);
+    fn sort_with(&self, options: SortOptions) -> ChunkedArray<T> {
+        if self.is_empty() {
+            return self.clone();
+        }
 
-        if self.null_count() == 0 {
+        if options.descending && self.is_sorted_reverse() || self.is_sorted() {
+            // there are nulls
+            if self.has_validity() {
+                // if the nulls are already last we can clone
+                if options.nulls_last && self.get(self.len() - 1).is_none()  ||
+                // if the nulls are already first we can clone
+                self.get(0).is_none()
+                {
+                    return self.clone();
+                }
+                // nulls are not at the right place
+                // continue w/ sorting
+                // TODO: we can optimize here and just put the null at the correct place
+            } else {
+                return self.clone();
+            }
+        }
+        if !self.has_validity() {
             let mut vals = memcpy_values(self);
-
             sort_branch(
                 vals.as_mut_slice(),
-                sort_parallel,
-                reverse,
+                options.descending,
                 order_default,
                 order_reverse,
             );
 
-            return ChunkedArray::new_from_aligned_vec(self.name(), vals);
+            ChunkedArray::new_from_aligned_vec(self.name(), vals)
         } else {
+            let null_count = self.null_count();
+            let len = self.len();
             let mut vals = Vec::with_capacity(self.len());
+
+            if !options.nulls_last {
+                let iter = std::iter::repeat(T::Native::default()).take(null_count);
+                vals.extend(iter);
+            }
+
             self.downcast_iter().for_each(|arr| {
-                let iter = arr.iter().map(|opt_v| opt_v.copied());
+                let iter = arr
+                    .iter()
+                    .filter_map(|v| v.copied())
+                    .trust_my_length(len - null_count);
                 vals.extend_trusted_len(iter);
             });
-            sort_branch(
-                vals.as_mut_slice(),
-                sort_parallel,
-                reverse,
-                order_default_null,
-                order_reverse_null,
-            );
-            let mut ca: Self = vals.into_iter().collect_trusted();
-            ca.rename(self.name());
-            ca.set_sorted(reverse);
+            let mut_slice = if options.nulls_last {
+                &mut vals[..len - null_count]
+            } else {
+                &mut vals[null_count..]
+            };
+            sort_branch(mut_slice, options.descending, order_default, order_reverse);
+
+            let mut ca: Self = if options.nulls_last {
+                vals.extend(std::iter::repeat(T::Native::default()).take(self.null_count()));
+                let mut validity = MutableBitmap::with_capacity(len);
+                validity.extend_constant(len - null_count, true);
+                validity.extend_constant(null_count, false);
+
+                (
+                    self.name(),
+                    PrimitiveArray::from_data(
+                        T::get_dtype().to_arrow(),
+                        vals.into(),
+                        Some(validity.into()),
+                    ),
+                )
+                    .into()
+            } else {
+                let mut validity = MutableBitmap::with_capacity(len);
+                validity.extend_constant(null_count, false);
+                validity.extend_constant(len - null_count, true);
+
+                (
+                    self.name(),
+                    PrimitiveArray::from_data(
+                        T::get_dtype().to_arrow(),
+                        vals.into(),
+                        Some(validity.into()),
+                    ),
+                )
+                    .into()
+            };
+
+            ca.set_sorted(options.descending);
             ca
         }
     }
 
+    fn sort(&self, reverse: bool) -> ChunkedArray<T> {
+        self.sort_with(SortOptions {
+            descending: reverse,
+            ..Default::default()
+        })
+    }
+
     fn sort_in_place(&mut self, reverse: bool) {
-        let sorted = self.sort(reverse);
-        self.chunks = sorted.chunks;
+        let mut sorted = self.sort(reverse);
+        self.chunks = std::mem::take(&mut sorted.chunks);
     }
 
     fn argsort(&self, reverse: bool) -> UInt32Chunked {
-        let sort_parallel = sort_parallel(self);
-
-        let ca: NoNull<UInt32Chunked> = if self.null_count() == 0 {
+        if !self.has_validity() {
             let mut vals = Vec::with_capacity(self.len());
             let mut count: u32 = 0;
             self.downcast_iter().for_each(|arr| {
@@ -216,37 +263,69 @@ where
 
             argsort_branch(
                 vals.as_mut_slice(),
-                sort_parallel,
                 reverse,
                 |(_, a), (_, b)| a.partial_cmp(b).unwrap(),
                 |(_, a), (_, b)| b.partial_cmp(a).unwrap(),
             );
 
-            vals.into_iter().map(|(idx, _v)| idx).collect_trusted()
+            let ca: NoNull<UInt32Chunked> = vals.into_iter().map(|(idx, _v)| idx).collect_trusted();
+            let mut ca = ca.into_inner();
+            ca.rename(self.name());
+            ca
         } else {
-            let mut vals = Vec::with_capacity(self.len());
+            let null_count = self.null_count();
+            let len = self.len();
+            let mut vals = Vec::with_capacity(len - null_count);
+
+            // if we sort reverse, the nulls are last
+            // and need to be extended to the indices in reverse order
+            let null_cap = if reverse {
+                null_count
+            // if we sort normally, the nulls are first
+            // and can be extended with the sorted indices
+            } else {
+                len
+            };
+            let mut nulls_idx = Vec::with_capacity(null_cap);
             let mut count: u32 = 0;
             self.downcast_iter().for_each(|arr| {
-                let iter = arr.iter().map(|v| {
+                let iter = arr.iter().filter_map(|v| {
                     let i = count;
                     count += 1;
-                    (i, v)
+                    match v {
+                        Some(v) => Some((i, *v)),
+                        None => {
+                            // Safety:
+                            // we allocated enough
+                            unsafe { nulls_idx.push_unchecked(i) };
+                            None
+                        }
+                    }
                 });
-                vals.extend_trusted_len(iter);
+                vals.extend(iter);
             });
 
             argsort_branch(
                 vals.as_mut_slice(),
-                sort_parallel,
                 reverse,
-                |(_, a), (_, b)| order_default_null(a, b),
-                |(_, a), (_, b)| order_reverse_null(a, b),
+                |(_, a), (_, b)| a.partial_cmp(b).unwrap(),
+                |(_, a), (_, b)| b.partial_cmp(a).unwrap(),
             );
-            vals.into_iter().map(|(idx, _v)| idx).collect_trusted()
-        };
-        let mut ca = ca.into_inner();
-        ca.rename(self.name());
-        ca
+
+            let iter = vals.into_iter().map(|(idx, _v)| idx);
+            let idx = if reverse {
+                let mut idx = Vec::with_capacity(len);
+                idx.extend(iter);
+                idx.extend(nulls_idx.into_iter().rev());
+                idx
+            } else {
+                nulls_idx.extend(iter);
+                nulls_idx
+            };
+
+            let arr = UInt32Array::from_data(ArrowDataType::UInt32, Buffer::from_vec(idx), None);
+            UInt32Chunked::new_from_chunks(self.name(), vec![Arc::new(arr)])
+        }
     }
 
     #[cfg(feature = "sort_multiple")]
@@ -271,7 +350,10 @@ where
 
         assert_eq!(other.len(), reverse.len() - 1);
 
-        let compare_inner: Vec<_> = other.iter().map(|s| s.into_partial_ord_inner()).collect();
+        let compare_inner: Vec<_> = other
+            .iter()
+            .map(|s| s.into_partial_ord_inner())
+            .collect_trusted();
 
         let mut count: u32 = 0;
         let mut vals: Vec<_> = self
@@ -281,7 +363,7 @@ where
                 count += 1;
                 (i, v)
             })
-            .collect();
+            .collect_trusted();
 
         vals.sort_by(
             |tpl_a, tpl_b| match (reverse[0], sort_with_nulls(&tpl_a.1, &tpl_b.1)) {
@@ -297,7 +379,7 @@ where
                 (_, ord) => ord,
             },
         );
-        let ca: NoNull<UInt32Chunked> = vals.into_iter().map(|(idx, _v)| idx).collect();
+        let ca: NoNull<UInt32Chunked> = vals.into_iter().map(|(idx, _v)| idx).collect_trusted();
         let mut ca = ca.into_inner();
         ca.set_sorted(reverse[0]);
         Ok(ca)
@@ -327,21 +409,24 @@ fn ordering_other_columns<'a>(
 macro_rules! sort {
     ($self:ident, $reverse:ident) => {{
         if $reverse {
-            $self.into_iter().sorted_by(|a, b| b.cmp(a)).collect()
+            $self
+                .into_iter()
+                .sorted_by(|a, b| b.cmp(a))
+                .collect_trusted()
         } else {
-            $self.into_iter().sorted_by(|a, b| a.cmp(b)).collect()
+            $self
+                .into_iter()
+                .sorted_by(|a, b| a.cmp(b))
+                .collect_trusted()
         }
     }};
 }
 
 impl ChunkSort<Utf8Type> for Utf8Chunked {
     fn sort(&self, reverse: bool) -> Utf8Chunked {
-        let sort_parallel = sort_parallel(self);
-
         let mut v = Vec::from_iter(self);
         sort_branch(
             v.as_mut_slice(),
-            sort_parallel,
             reverse,
             order_default_null,
             order_reverse_null,
@@ -356,8 +441,8 @@ impl ChunkSort<Utf8Type> for Utf8Chunked {
     }
 
     fn sort_in_place(&mut self, reverse: bool) {
-        let sorted = self.sort(reverse);
-        self.chunks = sorted.chunks;
+        let mut sorted = self.sort(reverse);
+        self.chunks = std::mem::take(&mut sorted.chunks);
     }
 
     fn argsort(&self, reverse: bool) -> UInt32Chunked {
@@ -390,8 +475,11 @@ impl ChunkSort<Utf8Type> for Utf8Chunked {
                 count += 1;
                 (i, v)
             })
-            .collect();
-        let compare_inner: Vec<_> = other.iter().map(|s| s.into_partial_ord_inner()).collect();
+            .collect_trusted();
+        let compare_inner: Vec<_> = other
+            .iter()
+            .map(|s| s.into_partial_ord_inner())
+            .collect_trusted();
 
         vals.sort_by(
             |tpl_a, tpl_b| match (reverse[0], sort_with_nulls(&tpl_a.1, &tpl_b.1)) {
@@ -407,16 +495,17 @@ impl ChunkSort<Utf8Type> for Utf8Chunked {
                 (_, ord) => ord,
             },
         );
-        let ca: NoNull<UInt32Chunked> = vals.into_iter().map(|(idx, _v)| idx).collect();
+        let ca: NoNull<UInt32Chunked> = vals.into_iter().map(|(idx, _v)| idx).collect_trusted();
         let mut ca = ca.into_inner();
         ca.set_sorted(reverse[0]);
         Ok(ca)
     }
 }
 
+#[cfg(feature = "dtype-categorical")]
 impl ChunkSort<CategoricalType> for CategoricalChunked {
     fn sort(&self, reverse: bool) -> Self {
-        self.as_ref().sort(reverse).cast().unwrap()
+        self.deref().sort(reverse).into()
     }
 
     fn sort_in_place(&mut self, reverse: bool) {
@@ -428,43 +517,14 @@ impl ChunkSort<CategoricalType> for CategoricalChunked {
     }
 }
 
-impl ChunkSort<ListType> for ListChunked {
-    fn sort(&self, _reverse: bool) -> Self {
-        unimplemented!()
-    }
-
-    fn sort_in_place(&mut self, _reverse: bool) {
-        unimplemented!()
-    }
-
-    fn argsort(&self, _reverse: bool) -> UInt32Chunked {
-        unimplemented!()
-    }
-}
-
-#[cfg(feature = "object")]
-impl<T> ChunkSort<ObjectType<T>> for ObjectChunked<T> {
-    fn sort(&self, _reverse: bool) -> Self {
-        unimplemented!()
-    }
-
-    fn sort_in_place(&mut self, _reverse: bool) {
-        unimplemented!()
-    }
-
-    fn argsort(&self, _reverse: bool) -> UInt32Chunked {
-        unimplemented!()
-    }
-}
-
 impl ChunkSort<BooleanType> for BooleanChunked {
     fn sort(&self, reverse: bool) -> BooleanChunked {
         sort!(self, reverse)
     }
 
     fn sort_in_place(&mut self, reverse: bool) {
-        let sorted = self.sort(reverse);
-        self.chunks = sorted.chunks;
+        let mut sorted = self.sort(reverse);
+        self.chunks = std::mem::take(&mut sorted.chunks);
     }
 
     fn argsort(&self, reverse: bool) -> UInt32Chunked {
@@ -485,7 +545,15 @@ pub(crate) fn prepare_argsort(
             use DataType::*;
             match s.dtype() {
                 Float32 | Float64 | Int32 | Int64 | Utf8 | UInt32 | UInt64 => s.clone(),
-                _ => s.cast::<Int32Type>().unwrap(),
+                _ => {
+                    // small integers i8, u8 etc are casted to reduce compiler bloat
+                    // not that we don't expect any logical types at this point
+                    if s.bit_repr_is_large() {
+                        s.cast(&DataType::Int64).unwrap()
+                    } else {
+                        s.cast(&DataType::Int32).unwrap()
+                    }
+                }
             }
         })
         .collect::<Vec<_>>();
@@ -505,6 +573,85 @@ pub(crate) fn prepare_argsort(
 mod test {
     use crate::prelude::*;
 
+    #[test]
+    fn test_argsort() {
+        let a = Int32Chunked::new_from_opt_slice(
+            "a",
+            &[
+                Some(1), // 0
+                Some(5), // 1
+                None,    // 2
+                Some(1), // 3
+                None,    // 4
+                Some(4), // 5
+                Some(3), // 6
+                Some(1), // 7
+            ],
+        );
+        let idx = a.argsort(false);
+        let idx = idx.cont_slice().unwrap();
+
+        let expected = [2, 4, 0, 3, 7, 6, 5, 1];
+        assert_eq!(idx, expected);
+
+        let idx = a.argsort(true);
+        let idx = idx.cont_slice().unwrap();
+        // the duplicates are in reverse order of appearance, so we cannot reverse expected
+        let expected = [1, 5, 6, 0, 3, 7, 4, 2];
+        assert_eq!(idx, expected);
+    }
+
+    #[test]
+    fn test_sort() {
+        let a = Int32Chunked::new_from_opt_slice(
+            "a",
+            &[
+                Some(1),
+                Some(5),
+                None,
+                Some(1),
+                None,
+                Some(4),
+                Some(3),
+                Some(1),
+            ],
+        );
+        let out = a.sort_with(SortOptions {
+            descending: false,
+            nulls_last: false,
+        });
+        assert_eq!(
+            Vec::from(&out),
+            &[
+                None,
+                None,
+                Some(1),
+                Some(1),
+                Some(1),
+                Some(3),
+                Some(4),
+                Some(5)
+            ]
+        );
+        let out = a.sort_with(SortOptions {
+            descending: false,
+            nulls_last: true,
+        });
+        assert_eq!(
+            Vec::from(&out),
+            &[
+                Some(1),
+                Some(1),
+                Some(1),
+                Some(3),
+                Some(4),
+                Some(5),
+                None,
+                None
+            ]
+        );
+    }
+
     #[test]
     #[cfg(feature = "sort_multiple")]
     #[cfg_attr(miri, ignore)]
