@@ -3,14 +3,11 @@ use crate::chunked_array::builder::get_list_builder;
 #[cfg(feature = "object")]
 use crate::chunked_array::object::ObjectArray;
 use crate::prelude::*;
-use crate::utils::arrow::buffer::MutableBuffer;
 use crate::utils::NoNull;
 use crate::utils::{get_iter_capacity, CustomIterTools};
-use arrow::array::{ArrayData, BooleanArray, LargeStringArray, PrimitiveArray};
-use arrow::buffer::Buffer;
-#[cfg(feature = "object")]
-use polars_arrow::prelude::BooleanBufferBuilder;
-use polars_arrow::trusted_len::trusted_len_unzip_extend;
+use arrow::array::{BooleanArray, PrimitiveArray, Utf8Array};
+use arrow::buffer::MutableBuffer;
+use polars_arrow::utils::TrustMyLength;
 use rayon::iter::{FromParallelIterator, IntoParallelIterator};
 use rayon::prelude::*;
 use std::borrow::{Borrow, Cow};
@@ -35,12 +32,12 @@ impl<T> Default for ChunkedArray<T> {
 
 impl<T> FromIterator<Option<T::Native>> for ChunkedArray<T>
 where
-    T: PolarsPrimitiveType,
+    T: PolarsNumericType,
 {
     fn from_iter<I: IntoIterator<Item = Option<T::Native>>>(iter: I) -> Self {
         let iter = iter.into_iter();
 
-        let arr: PrimitiveArray<T> = match iter.size_hint() {
+        let arr: PrimitiveArray<T::Native> = match iter.size_hint() {
             (a, Some(b)) if a == b => {
                 // 2021-02-07: ~40% faster than builder.
                 // It is unsafe because we cannot be certain that the iterators length can be trusted.
@@ -49,18 +46,18 @@ where
                 // This will not lead to UB, but will panic.
                 #[cfg(feature = "performant")]
                 unsafe {
-                    let arr = PrimitiveArray::from_trusted_len_iter(iter);
+                    let arr = PrimitiveArray::from_trusted_len_iter_unchecked(iter)
+                        .to(T::get_dtype().to_arrow());
                     assert_eq!(arr.len(), a);
                     arr
                 }
                 #[cfg(not(feature = "performant"))]
-                PrimitiveArray::from_iter(iter)
-            }
-            _ => {
-                // 2021-02-07: ~1.5% slower than builder. Will still use this as it is more idiomatic and will
-                // likely improve over time.
-                PrimitiveArray::from_iter(iter)
+                iter.collect::<PrimitiveArray<T::Native>>()
+                    .to(T::get_dtype().to_arrow())
             }
+            _ => iter
+                .collect::<PrimitiveArray<T::Native>>()
+                .to(T::get_dtype().to_arrow()),
         };
         ChunkedArray::new_from_chunks("", vec![Arc::new(arr)])
     }
@@ -69,30 +66,14 @@ where
 // NoNull is only a wrapper needed for specialization
 impl<T> FromIterator<T::Native> for NoNull<ChunkedArray<T>>
 where
-    T: PolarsPrimitiveType,
+    T: PolarsNumericType,
 {
     // We use AlignedVec because it is way faster than Arrows builder. We can do this because we
     // know we don't have null values.
     fn from_iter<I: IntoIterator<Item = T::Native>>(iter: I) -> Self {
-        let iter = iter.into_iter();
-        #[cfg(feature = "performant")]
-        {
-            let len = iter.size_hint().0;
-            let buffer = unsafe { Buffer::from_trusted_len_iter(iter) };
-
-            let data = ArrayData::new(T::DATA_TYPE, len, None, None, 0, vec![buffer], vec![]);
-            NoNull::new(ChunkedArray::new_from_chunks(
-                "",
-                vec![Arc::new(PrimitiveArray::<T>::from(data))],
-            ))
-        }
-        #[cfg(not(feature = "performant"))]
-        {
-            // 2021-02-07: aligned vec was ~2x faster than arrow collect.
-            let mut av = AlignedVec::with_capacity(0);
-            av.extend(iter);
-            NoNull::new(ChunkedArray::new_from_aligned_vec("", av))
-        }
+        // 2021-02-07: aligned vec was ~2x faster than arrow collect.
+        let av = iter.into_iter().collect::<AlignedVec<T::Native>>();
+        NoNull::new(ChunkedArray::new_from_aligned_vec("", av))
     }
 }
 
@@ -125,8 +106,7 @@ where
     Ptr: AsRef<str>,
 {
     fn from_iter<I: IntoIterator<Item = Option<Ptr>>>(iter: I) -> Self {
-        // 2021-02-07: this was ~30% faster than with the builder.
-        let arr = LargeStringArray::from_iter(iter);
+        let arr = Utf8Array::<i64>::from_iter(iter);
         Self::new_from_chunks("", vec![Arc::new(arr)])
     }
 }
@@ -145,7 +125,7 @@ where
     Ptr: PolarsAsRef<str>,
 {
     fn from_iter<I: IntoIterator<Item = Ptr>>(iter: I) -> Self {
-        let arr = LargeStringArray::from_iter_values(iter);
+        let arr = Utf8Array::<i64>::from_iter_values(iter.into_iter());
         Self::new_from_chunks("", vec![Arc::new(arr)])
     }
 }
@@ -158,8 +138,11 @@ where
         let mut it = iter.into_iter();
         let capacity = get_iter_capacity(&it);
 
-        // first take one to get the dtype. We panic if we have an empty iterator
-        let v = it.next().unwrap();
+        // first take one to get the dtype.
+        let v = match it.next() {
+            Some(v) => v,
+            None => return ListChunked::full_null("", 0),
+        };
         // We don't know the needed capacity. We arbitrarily choose an average of 5 elements per series.
         let mut builder = get_list_builder(v.borrow().dtype(), capacity * 5, capacity, "collected");
 
@@ -194,7 +177,7 @@ where
                 // end of iterator
                 None => {
                     // type is not known
-                    panic!("Type of Series cannot be determined as they are all null")
+                    return ListChunked::full_null("", 0);
                 }
             }
         }
@@ -226,38 +209,34 @@ where
 #[cfg(feature = "object")]
 impl<T: PolarsObject> FromIterator<Option<T>> for ObjectChunked<T> {
     fn from_iter<I: IntoIterator<Item = Option<T>>>(iter: I) -> Self {
+        use arrow::bitmap::Bitmap;
+        use arrow::bitmap::MutableBitmap;
+
         let iter = iter.into_iter();
         let size = iter.size_hint().0;
-        let mut null_mask_builder = BooleanBufferBuilder::new(size);
+        let mut null_mask_builder = MutableBitmap::with_capacity(size);
 
-        let mut null_count = 0;
         let values: Vec<T> = iter
             .map(|value| match value {
                 Some(value) => {
-                    null_mask_builder.append(true);
+                    null_mask_builder.push(true);
                     value
                 }
                 None => {
-                    null_count += 1;
-                    null_mask_builder.append(false);
+                    null_mask_builder.push(false);
                     T::default()
                 }
             })
             .collect();
 
-        let null_bit_buffer = null_mask_builder.finish();
-        let null_bitmap = arrow::bitmap::Bitmap::from(null_bit_buffer);
-        let null_bitmap = match null_count {
-            0 => None,
-            _ => Some(Arc::new(null_bitmap)),
-        };
+        let null_bit_buffer: Option<Bitmap> = null_mask_builder.into();
+        let null_bitmap = null_bit_buffer;
 
         let len = values.len();
 
         let arr = Arc::new(ObjectArray {
             values: Arc::new(values),
             null_bitmap,
-            null_count,
             offset: 0,
             len,
         });
@@ -293,20 +272,13 @@ fn collect_into_linked_list<I>(par_iter: I) -> LinkedList<Vec<I::Item>>
 where
     I: IntoParallelIterator,
 {
-    use crate::POOL;
     let it = par_iter.into_par_iter();
-    if let Some(len) = it.opt_len() {
-        it.fold(
-            || Vec::with_capacity(len / POOL.current_num_threads()),
-            vec_push,
-        )
+    // be careful optimizing allocations. Its hard to figure out the size
+    // needed
+    // https://github.com/pola-rs/polars/issues/1562
+    it.fold(Vec::new, vec_push)
         .map(as_list)
         .reduce(LinkedList::new, list_append)
-    } else {
-        it.fold(Vec::new, vec_push)
-            .map(as_list)
-            .reduce(LinkedList::new, list_append)
-    }
 }
 
 fn get_capacity_from_par_results<T>(ll: &LinkedList<Vec<T>>) -> usize {
@@ -315,41 +287,35 @@ fn get_capacity_from_par_results<T>(ll: &LinkedList<Vec<T>>) -> usize {
 
 impl<T> FromParallelIterator<T::Native> for NoNull<ChunkedArray<T>>
 where
-    T: PolarsPrimitiveType,
+    T: PolarsNumericType,
 {
     fn from_par_iter<I: IntoParallelIterator<Item = T::Native>>(iter: I) -> Self {
         // Get linkedlist filled with different vec result from different threads
         let vectors = collect_into_linked_list(iter);
         let capacity: usize = get_capacity_from_par_results(&vectors);
 
-        let mut av = AlignedVec::with_capacity(capacity);
+        let mut av = MutableBuffer::<T::Native>::with_capacity(capacity);
         for v in vectors {
             av.extend_from_slice(&v)
         }
-        let arr = av.into_primitive_array::<T>(None);
-        NoNull::new(ChunkedArray::new_from_chunks("", vec![Arc::new(arr)]))
+        let arr = to_array::<T>(av, None);
+        NoNull::new(ChunkedArray::new_from_chunks("", vec![arr]))
     }
 }
 
 impl<T> FromParallelIterator<Option<T::Native>> for ChunkedArray<T>
 where
-    T: PolarsPrimitiveType,
+    T: PolarsNumericType,
 {
     fn from_par_iter<I: IntoParallelIterator<Item = Option<T::Native>>>(iter: I) -> Self {
         // Get linkedlist filled with different vec result from different threads
         let vectors = collect_into_linked_list(iter);
 
         let capacity: usize = get_capacity_from_par_results(&vectors);
-        // zeroed immediately correct for nulls
-        let mut validity = MutableBuffer::from_len_zeroed(capacity.saturating_add(7) / 8);
-        let mut values = AlignedVec::with_capacity(capacity);
 
-        for vec in vectors {
-            // Safety:
-            // vec's length can be trusted
-            unsafe { trusted_len_unzip_extend(vec.into_iter(), &mut values, &mut validity) }
-        }
-        let arr: PrimitiveArray<T> = values.into_primitive_array(Some(validity.into()));
+        let iter = TrustMyLength::new(vectors.into_iter().flatten(), capacity);
+        let arr =
+            PrimitiveArray::<T::Native>::from_trusted_len_iter(iter).to(T::get_dtype().to_arrow());
         Self::new_from_chunks("", vec![Arc::new(arr)])
     }
 }
