@@ -6,16 +6,24 @@
 //! [See the AnyValue variants](enum.AnyValue.html#variants) for the data types that
 //! are currently supported.
 //!
+#[cfg(feature = "dtype-categorical")]
+use crate::chunked_array::categorical::RevMapping;
+pub use crate::chunked_array::logical::*;
 #[cfg(feature = "object")]
 use crate::chunked_array::object::PolarsObjectSafe;
 use crate::prelude::*;
 use ahash::RandomState;
+use arrow::compute::comparison::Simd8;
+use arrow::datatypes::IntegerType;
 pub use arrow::datatypes::{DataType as ArrowDataType, TimeUnit};
+use arrow::types::simd::Simd;
 use arrow::types::NativeType;
+use num::{Bounded, FromPrimitive, Num, NumCast, Zero};
 #[cfg(feature = "serde")]
 use serde::{Deserialize, Serialize};
 use std::cmp::Ordering;
 use std::fmt::{Display, Formatter};
+use std::ops::{Add, AddAssign, Div, Mul, Rem, Sub};
 
 pub struct Utf8Type {}
 
@@ -52,32 +60,9 @@ impl_polars_datatype!(Int32Type, Int32, i32);
 impl_polars_datatype!(Int64Type, Int64, i64);
 impl_polars_datatype!(Float32Type, Float32, f32);
 impl_polars_datatype!(Float64Type, Float64, f64);
-impl_polars_datatype!(Date32Type, Date32, i32);
-impl_polars_datatype!(Date64Type, Date64, i64);
-
-pub struct Time64NanosecondType {}
-
-impl PolarsDataType for Time64NanosecondType {
-    fn get_dtype() -> DataType {
-        DataType::Time64(TimeUnit::Nanosecond)
-    }
-}
-
-pub struct DurationNanosecondType {}
-
-impl PolarsDataType for DurationNanosecondType {
-    fn get_dtype() -> DataType {
-        DataType::Duration(TimeUnit::Nanosecond)
-    }
-}
-
-pub struct DurationMillisecondType {}
-
-impl PolarsDataType for DurationMillisecondType {
-    fn get_dtype() -> DataType {
-        DataType::Duration(TimeUnit::Millisecond)
-    }
-}
+impl_polars_datatype!(DateType, Date, i32);
+impl_polars_datatype!(DatetimeType, Datetime, i64);
+impl_polars_datatype!(TimeType, Time, i64);
 
 impl PolarsDataType for Utf8Type {
     fn get_dtype() -> DataType {
@@ -96,7 +81,7 @@ impl PolarsDataType for BooleanType {
 impl PolarsDataType for ListType {
     fn get_dtype() -> DataType {
         // null as we cannot no anything without self.
-        DataType::List(ArrowDataType::Null)
+        DataType::List(Box::new(DataType::Null))
     }
 }
 
@@ -140,84 +125,71 @@ pub type Int64Chunked = ChunkedArray<Int64Type>;
 pub type Float32Chunked = ChunkedArray<Float32Type>;
 pub type Float64Chunked = ChunkedArray<Float64Type>;
 pub type Utf8Chunked = ChunkedArray<Utf8Type>;
-pub type Date32Chunked = ChunkedArray<Date32Type>;
-pub type Date64Chunked = ChunkedArray<Date64Type>;
-pub type DurationNanosecondChunked = ChunkedArray<DurationNanosecondType>;
-pub type DurationMillisecondChunked = ChunkedArray<DurationMillisecondType>;
-pub type Time64NanosecondChunked = ChunkedArray<Time64NanosecondType>;
 pub type CategoricalChunked = ChunkedArray<CategoricalType>;
 
-pub trait PolarsPrimitiveType: Send + Sync + PolarsDataType + 'static {
-    type Native: NativeType;
+pub trait NumericNative:
+    PartialOrd
+    + NativeType
+    + Num
+    + NumCast
+    + Zero
+    + Simd
+    + Simd8
+    + std::iter::Sum<Self>
+    + Add<Output = Self>
+    + Sub<Output = Self>
+    + Mul<Output = Self>
+    + Div<Output = Self>
+    + Rem<Output = Self>
+    + AddAssign
+    + Bounded
+    + FromPrimitive
+{
 }
-impl PolarsPrimitiveType for UInt8Type {
+impl NumericNative for i8 {}
+impl NumericNative for i16 {}
+impl NumericNative for i32 {}
+impl NumericNative for i64 {}
+impl NumericNative for u8 {}
+impl NumericNative for u16 {}
+impl NumericNative for u32 {}
+impl NumericNative for u64 {}
+impl NumericNative for f32 {}
+impl NumericNative for f64 {}
+
+pub trait PolarsNumericType: Send + Sync + PolarsDataType + 'static {
+    type Native: NumericNative;
+}
+impl PolarsNumericType for UInt8Type {
     type Native = u8;
 }
-impl PolarsPrimitiveType for UInt16Type {
+impl PolarsNumericType for UInt16Type {
     type Native = u16;
 }
-impl PolarsPrimitiveType for UInt32Type {
+impl PolarsNumericType for UInt32Type {
     type Native = u32;
 }
-impl PolarsPrimitiveType for UInt64Type {
+impl PolarsNumericType for UInt64Type {
     type Native = u64;
 }
-impl PolarsPrimitiveType for Int8Type {
+impl PolarsNumericType for Int8Type {
     type Native = i8;
 }
-impl PolarsPrimitiveType for Int16Type {
+impl PolarsNumericType for Int16Type {
     type Native = i16;
 }
-impl PolarsPrimitiveType for Int32Type {
+impl PolarsNumericType for Int32Type {
     type Native = i32;
 }
-impl PolarsPrimitiveType for Int64Type {
+impl PolarsNumericType for Int64Type {
     type Native = i64;
 }
-impl PolarsPrimitiveType for Float32Type {
+impl PolarsNumericType for Float32Type {
     type Native = f32;
 }
-impl PolarsPrimitiveType for Float64Type {
+impl PolarsNumericType for Float64Type {
     type Native = f64;
 }
-impl PolarsPrimitiveType for Date32Type {
-    type Native = i32;
-}
-impl PolarsPrimitiveType for Date64Type {
-    type Native = i64;
-}
-impl PolarsPrimitiveType for Time64NanosecondType {
-    type Native = i64;
-}
-impl PolarsPrimitiveType for DurationNanosecondType {
-    type Native = i64;
-}
-impl PolarsPrimitiveType for DurationMillisecondType {
-    type Native = i64;
-}
-
-macro_rules! impl_polars_numeric {
-    ($ca:ident, $physical:ty) => {
-        impl PolarsNumericType for $ca {}
-    };
-}
-
-pub trait PolarsNumericType: PolarsPrimitiveType {}
-impl_polars_numeric!(UInt8Type, u8);
-impl_polars_numeric!(UInt16Type, u16);
-impl_polars_numeric!(UInt32Type, u32);
-impl_polars_numeric!(UInt64Type, u64);
-impl_polars_numeric!(Int8Type, i8);
-impl_polars_numeric!(Int16Type, i16);
-impl_polars_numeric!(Int32Type, i32);
-impl_polars_numeric!(Int64Type, i64);
-impl_polars_numeric!(Float32Type, f32);
-impl_polars_numeric!(Float64Type, f64);
-impl_polars_numeric!(Date32Type, i32);
-impl_polars_numeric!(Date64Type, i64);
-impl_polars_numeric!(Time64NanosecondType, i64);
-impl_polars_numeric!(DurationNanosecondType, i64);
-impl_polars_numeric!(DurationMillisecondType, i64);
 
 pub trait PolarsIntegerType: PolarsNumericType {}
 impl PolarsIntegerType for UInt8Type {}
@@ -228,41 +200,11 @@ impl PolarsIntegerType for Int8Type {}
 impl PolarsIntegerType for Int16Type {}
 impl PolarsIntegerType for Int32Type {}
 impl PolarsIntegerType for Int64Type {}
-impl PolarsIntegerType for Date32Type {}
-impl PolarsIntegerType for Date64Type {}
-impl PolarsIntegerType for Time64NanosecondType {}
-impl PolarsIntegerType for DurationNanosecondType {}
-impl PolarsIntegerType for DurationMillisecondType {}
 
 pub trait PolarsFloatType: PolarsNumericType {}
 impl PolarsFloatType for Float32Type {}
 impl PolarsFloatType for Float64Type {}
 
-pub trait ToPolarsType {
-    fn to_polars_type() -> &'static dyn PolarsDataType;
-}
-
-macro_rules! to_polars_type {
-    ($native:ty, $polars:ident) => {
-        impl ToPolarsType for $native {
-            fn to_polars_type() -> &'static dyn PolarsDataType {
-                &$polars {}
-            }
-        }
-    };
-}
-
-to_polars_type!(i8, Int8Type);
-to_polars_type!(i16, Int16Type);
-to_polars_type!(i32, Int32Type);
-to_polars_type!(i64, Int64Type);
-to_polars_type!(u8, UInt8Type);
-to_polars_type!(u16, UInt16Type);
-to_polars_type!(u32, UInt32Type);
-to_polars_type!(u64, UInt64Type);
-to_polars_type!(f32, Float32Type);
-to_polars_type!(f64, Float64Type);
-
 #[derive(Debug, Clone)]
 pub enum AnyValue<'a> {
     Null,
@@ -292,14 +234,17 @@ pub enum AnyValue<'a> {
     Float64(f64),
     /// A 32-bit date representing the elapsed time since UNIX epoch (1970-01-01)
     /// in days (32 bits).
-    Date32(i32),
+    #[cfg(feature = "dtype-date")]
+    Date(i32),
     /// A 64-bit date representing the elapsed time since UNIX epoch (1970-01-01)
     /// in milliseconds (64 bits).
-    Date64(i64),
-    /// A 64-bit time representing the elapsed time since midnight in the unit of `TimeUnit`.
-    Time64(i64, TimeUnit),
-    /// A 32-bit time representing the elapsed time since midnight in the unit of `TimeUnit`.
-    Duration(i64, TimeUnit),
+    #[cfg(feature = "dtype-datetime")]
+    Datetime(i64),
+    /// A 64-bit time representing the elapsed time since midnight in nanoseconds
+    #[cfg(feature = "dtype-time")]
+    Time(i64),
+    #[cfg(feature = "dtype-categorical")]
+    Categorical(u32, &'a RevMapping),
     /// Nested type, contains arrays that are filled with one of the datetypes.
     List(Series),
     #[cfg(feature = "object")]
@@ -372,7 +317,28 @@ where
 }
 
 impl<'a> AnyValue<'a> {
-    pub fn add<'b>(&self, rhs: &AnyValue<'b>) -> AnyValue<'a> {
+    #[cfg(any(feature = "dtype-date", feature = "dtype-datetime"))]
+    pub(crate) fn into_date(self) -> Self {
+        match self {
+            #[cfg(feature = "dtype-date")]
+            AnyValue::Int32(v) => AnyValue::Date(v),
+            #[cfg(feature = "dtype-datetime")]
+            AnyValue::Int64(v) => AnyValue::Datetime(v),
+            AnyValue::Null => AnyValue::Null,
+            dt => panic!("cannot create date from other type. dtype: {}", dt),
+        }
+    }
+
+    #[cfg(feature = "dtype-time")]
+    pub(crate) fn into_time(self) -> Self {
+        match self {
+            AnyValue::Int64(v) => AnyValue::Time(v),
+            AnyValue::Null => AnyValue::Null,
+            dt => panic!("cannot create date from other type. dtype: {}", dt),
+        }
+    }
+
+    pub fn add<'b>(&self, rhs: &AnyValue<'b>) -> Self {
         use AnyValue::*;
         match (self, rhs) {
             (Null, _) => Null,
@@ -417,16 +383,13 @@ impl Display for DataType {
             DataType::Float32 => "f32",
             DataType::Float64 => "f64",
             DataType::Utf8 => "str",
-            DataType::Date32 => "date32(days)",
-            DataType::Date64 => "date64(ms)",
-            DataType::Time64(TimeUnit::Nanosecond) => "time64(ns)",
-            DataType::Duration(TimeUnit::Nanosecond) => "duration(ns)",
-            DataType::Duration(TimeUnit::Millisecond) => "duration(ms)",
-            DataType::List(tp) => return write!(f, "list [{}]", DataType::from(tp)),
+            DataType::Date => "date",
+            DataType::Datetime => "datetime",
+            DataType::Time => "time",
+            DataType::List(tp) => return write!(f, "list [{}]", tp),
             #[cfg(feature = "object")]
             DataType::Object(s) => s,
             DataType::Categorical => "cat",
-            _ => panic!("{:?} not implemented", self),
         };
         f.write_str(s)
     }
@@ -448,16 +411,28 @@ impl PartialEq for AnyValue<'_> {
             (Int64(l), Int64(r)) => l == r,
             (Float32(l), Float32(r)) => l == r,
             (Float64(l), Float64(r)) => l == r,
-            (Date32(l), Date32(r)) => l == r,
-            (Date64(l), Date64(r)) => l == r,
-            (Time64(l, _), Time64(r, _)) => l == r,
-            (Duration(l, _), Duration(r, _)) => l == r,
+            #[cfg(feature = "dtype-time")]
+            (Time(l), Time(r)) => l == r,
+            #[cfg(all(feature = "dtype-datetime", feature = "dtype-date"))]
+            (Date(l), Date(r)) => l == r,
+            #[cfg(all(feature = "dtype-datetime", feature = "dtype-date"))]
+            (Datetime(l), Datetime(r)) => l == r,
             (Boolean(l), Boolean(r)) => l == r,
             (List(_), List(_)) => panic!("eq between list series not supported"),
             #[cfg(feature = "object")]
             (Object(_), Object(_)) => panic!("eq between object not supported"),
             // should it?
             (Null, Null) => true,
+            #[cfg(feature = "dtype-categorical")]
+            (Categorical(idx_l, rev_l), Categorical(idx_r, rev_r)) => match (rev_l, rev_r) {
+                (RevMapping::Global(_, _, id_l), RevMapping::Global(_, _, id_r)) => {
+                    id_l == id_r && idx_l == idx_r
+                }
+                (RevMapping::Local(arr_l), RevMapping::Local(arr_r)) => {
+                    std::ptr::eq(arr_l, arr_r) && idx_l == idx_r
+                }
+                _ => false,
+            },
             _ => false,
         }
     }
@@ -496,16 +471,17 @@ pub enum DataType {
     Int64,
     Float32,
     Float64,
+    /// String data
     Utf8,
     /// A 32-bit date representing the elapsed time since UNIX epoch (1970-01-01)
     /// in days (32 bits).
-    Date32,
+    Date,
     /// A 64-bit date representing the elapsed time since UNIX epoch (1970-01-01)
     /// in milliseconds (64 bits).
-    Date64,
-    Time64(TimeUnit),
-    List(ArrowDataType),
-    Duration(TimeUnit),
+    Datetime,
+    /// A 64-bit time representing the elapsed time since midnight in nanoseconds
+    Time,
+    List(Box<DataType>),
     #[cfg(feature = "object")]
     /// A generic type that can be used in a `Series`
     /// &'static str can be used to determine/set inner type
@@ -515,6 +491,27 @@ pub enum DataType {
 }
 
 impl DataType {
+    pub fn inner_dtype(&self) -> Option<&DataType> {
+        if let DataType::List(inner) = self {
+            Some(&*inner)
+        } else {
+            None
+        }
+    }
+
+    /// Convert to the physical data type
+    pub fn to_physical(&self) -> DataType {
+        use DataType::*;
+        match self {
+            Date => Int32,
+            Datetime => Int64,
+            Time => Int64,
+            Categorical => UInt32,
+            _ => self.clone(),
+        }
+    }
+
+    /// Convert to an Arrow data type.
     pub fn to_arrow(&self) -> ArrowDataType {
         use DataType::*;
         match self {
@@ -530,15 +527,14 @@ impl DataType {
             Float32 => ArrowDataType::Float32,
             Float64 => ArrowDataType::Float64,
             Utf8 => ArrowDataType::LargeUtf8,
-            Date32 => ArrowDataType::Date32,
-            Date64 => ArrowDataType::Date64,
-            Time64(tu) => ArrowDataType::Time64(*tu),
+            Date => ArrowDataType::Date32,
+            Datetime => ArrowDataType::Timestamp(TimeUnit::Millisecond, None),
+            Time => ArrowDataType::Time64(TimeUnit::Nanosecond),
             List(dt) => ArrowDataType::LargeList(Box::new(arrow::datatypes::Field::new(
                 "",
-                dt.clone(),
+                dt.to_arrow(),
                 true,
             ))),
-            Duration(tu) => ArrowDataType::Duration(*tu),
             Null => ArrowDataType::Null,
             #[cfg(feature = "object")]
             Object(_) => panic!("cannot convert object to arrow"),
@@ -579,22 +575,20 @@ impl Field {
         self.data_type = dtype;
     }
 
+    pub fn set_name(&mut self, name: String) {
+        self.name = name;
+    }
+
     pub fn to_arrow(&self) -> ArrowField {
         ArrowField::new(&self.name, self.data_type.to_arrow(), true)
     }
 }
 
-#[derive(Clone, Debug, PartialEq, Hash)]
+#[derive(Clone, Debug, PartialEq, Hash, Default)]
 pub struct Schema {
     fields: Vec<Field>,
 }
 
-impl Default for Schema {
-    fn default() -> Self {
-        Schema { fields: vec![] }
-    }
-}
-
 impl Schema {
     pub fn rename<I, J, T, S>(&self, old_names: I, new_names: J) -> Result<Schema>
     where
@@ -675,11 +669,19 @@ impl Schema {
                         f.name(),
                         ArrowDataType::LargeList(Box::new(ArrowField::new(
                             "item",
-                            dt.clone(),
+                            dt.to_arrow(),
                             true,
                         ))),
                         true,
                     ),
+                    DataType::Categorical => ArrowField::new(
+                        f.name(),
+                        ArrowDataType::Dictionary(
+                            IntegerType::UInt32,
+                            Box::new(ArrowDataType::LargeUtf8),
+                        ),
+                        true,
+                    ),
                     _ => f.to_arrow(),
                 }
             })
@@ -732,21 +734,27 @@ impl From<&ArrowDataType> for DataType {
             ArrowDataType::Int16 => DataType::Int16,
             ArrowDataType::Int32 => DataType::Int32,
             ArrowDataType::Int64 => DataType::Int64,
-            ArrowDataType::LargeUtf8 => DataType::Utf8,
             ArrowDataType::Boolean => DataType::Boolean,
             ArrowDataType::Float32 => DataType::Float32,
             ArrowDataType::Float64 => DataType::Float64,
-            ArrowDataType::LargeList(f) => DataType::List(f.data_type().clone()),
-            ArrowDataType::Date32 => DataType::Date32,
-            ArrowDataType::Date64 => DataType::Date64,
-            ArrowDataType::Time64(TimeUnit::Nanosecond) => DataType::Time64(TimeUnit::Nanosecond),
-            ArrowDataType::Duration(TimeUnit::Nanosecond) => {
-                DataType::Duration(TimeUnit::Nanosecond)
-            }
-            ArrowDataType::Duration(TimeUnit::Millisecond) => {
-                DataType::Duration(TimeUnit::Millisecond)
-            }
+            ArrowDataType::LargeList(f) => DataType::List(Box::new(f.data_type().into())),
+            ArrowDataType::List(f) => DataType::List(Box::new(f.data_type().into())),
+            ArrowDataType::Date32 => DataType::Date,
+            ArrowDataType::Timestamp(_, _) | ArrowDataType::Date64 => DataType::Datetime,
+            ArrowDataType::LargeUtf8 => DataType::Utf8,
             ArrowDataType::Utf8 => DataType::Utf8,
+            ArrowDataType::Time64(_) | ArrowDataType::Time32(_) => DataType::Time,
+            ArrowDataType::Dictionary(_, _) => DataType::Categorical,
+            ArrowDataType::Extension(name, _, _) if name == "POLARS_EXTENSION_TYPE" => {
+                #[cfg(feature = "object")]
+                {
+                    DataType::Object("extension")
+                }
+                #[cfg(not(feature = "object"))]
+                {
+                    panic!("activate the 'object' feature to be able to load POLARS_EXTENSION_TYPE")
+                }
+            }
             dt => panic!("Arrow datatype {:?} not supported by Polars", dt),
         }
     }
@@ -778,3 +786,79 @@ impl From<ArrowSchema> for Schema {
 pub type PlHashMap<K, V> = hashbrown::HashMap<K, V, RandomState>;
 #[cfg(feature = "private")]
 pub type PlHashSet<V> = hashbrown::HashSet<V, RandomState>;
+
+#[cfg(test)]
+mod test {
+    use super::*;
+
+    #[test]
+    fn test_arrow_dtypes_to_polars() {
+        let dtypes = [
+            (ArrowDataType::Date64, DataType::Datetime),
+            (
+                ArrowDataType::Timestamp(TimeUnit::Nanosecond, None),
+                DataType::Datetime,
+            ),
+            (
+                ArrowDataType::Timestamp(TimeUnit::Microsecond, None),
+                DataType::Datetime,
+            ),
+            (
+                ArrowDataType::Timestamp(TimeUnit::Millisecond, None),
+                DataType::Datetime,
+            ),
+            (
+                ArrowDataType::Timestamp(TimeUnit::Second, None),
+                DataType::Datetime,
+            ),
+            (
+                ArrowDataType::Timestamp(TimeUnit::Second, Some("".to_string())),
+                DataType::Datetime,
+            ),
+            (ArrowDataType::LargeUtf8, DataType::Utf8),
+            (ArrowDataType::Utf8, DataType::Utf8),
+            (ArrowDataType::Time64(TimeUnit::Nanosecond), DataType::Time),
+            (ArrowDataType::Time64(TimeUnit::Millisecond), DataType::Time),
+            (ArrowDataType::Time64(TimeUnit::Microsecond), DataType::Time),
+            (ArrowDataType::Time64(TimeUnit::Second), DataType::Time),
+            (ArrowDataType::Time32(TimeUnit::Nanosecond), DataType::Time),
+            (ArrowDataType::Time32(TimeUnit::Millisecond), DataType::Time),
+            (ArrowDataType::Time32(TimeUnit::Microsecond), DataType::Time),
+            (ArrowDataType::Time32(TimeUnit::Second), DataType::Time),
+            (
+                ArrowDataType::List(Box::new(ArrowField::new(
+                    "item",
+                    ArrowDataType::Float64,
+                    true,
+                ))),
+                DataType::List(DataType::Float64.into()),
+            ),
+            (
+                ArrowDataType::LargeList(Box::new(ArrowField::new(
+                    "item",
+                    ArrowDataType::Float64,
+                    true,
+                ))),
+                DataType::List(DataType::Float64.into()),
+            ),
+            (
+                ArrowDataType::Dictionary(IntegerType::UInt32, ArrowDataType::Utf8.into()),
+                DataType::Categorical,
+            ),
+            (
+                ArrowDataType::Dictionary(IntegerType::UInt32, ArrowDataType::LargeUtf8.into()),
+                DataType::Categorical,
+            ),
+            (
+                ArrowDataType::Dictionary(IntegerType::UInt64, ArrowDataType::LargeUtf8.into()),
+                DataType::Categorical,
+            ),
+        ];
+
+        for (dt_a, dt_p) in dtypes {
+            let dt: DataType = (&dt_a).into();
+
+            assert_eq!(dt_p, dt);
+        }
+    }
+}
