@@ -1,9 +1,7 @@
 use crate::prelude::*;
 
-#[cfg(feature = "temporal")]
-use crate::chunked_array::temporal::{
-    date32_as_datetime, date64_as_datetime, time64_nanosecond_as_time,
-};
+#[cfg(any(feature = "dtype-date", feature = "dtype-datetime"))]
+use arrow::temporal_conversions::{date32_to_date, timestamp_ms_to_datetime};
 use num::{Num, NumCast};
 use std::{
     fmt,
@@ -15,61 +13,10 @@ const LIMIT: usize = 25;
 use comfy_table::presets::{ASCII_FULL, UTF8_FULL};
 #[cfg(feature = "pretty_fmt")]
 use comfy_table::*;
-
 #[cfg(all(feature = "plain_fmt", not(feature = "pretty_fmt")))]
 use prettytable::{Cell, Row, Table};
-
-/// Some unit functions that just pass the integer values if we don't want all chrono functionality
-#[cfg(not(feature = "temporal"))]
-mod temporal {
-    pub struct DateTime<T>(T)
-    where
-        T: Copy;
-
-    impl<T> DateTime<T>
-    where
-        T: Copy,
-    {
-        pub fn date(&self) -> T {
-            self.0
-        }
-    }
-
-    pub fn date32_as_datetime(v: i32) -> DateTime<i32> {
-        DateTime(v)
-    }
-    pub fn date64_as_datetime(v: i64) -> DateTime<i64> {
-        DateTime(v)
-    }
-    pub fn time32_millisecond_as_time(v: i32) -> i32 {
-        v
-    }
-    pub fn time32_second_as_time(v: i32) -> i32 {
-        v
-    }
-    pub fn time64_nanosecond_as_time(v: i64) -> i64 {
-        v
-    }
-    pub fn time64_microsecond_as_time(v: i64) -> i64 {
-        v
-    }
-    pub fn timestamp_nanoseconds_as_datetime(v: i64) -> i64 {
-        v
-    }
-    pub fn timestamp_microseconds_as_datetime(v: i64) -> i64 {
-        v
-    }
-    pub fn timestamp_milliseconds_as_datetime(v: i64) -> i64 {
-        v
-    }
-    pub fn timestamp_seconds_as_datetime(v: i64) -> i64 {
-        v
-    }
-}
 #[cfg(any(feature = "plain_fmt", feature = "pretty_fmt"))]
 use std::borrow::Cow;
-#[cfg(not(feature = "temporal"))]
-use temporal::*;
 
 macro_rules! format_array {
     ($limit:expr, $f:ident, $a:expr, $dtype:expr, $name:expr, $array_type:expr) => {{
@@ -192,6 +139,7 @@ impl Debug for ListChunked {
     }
 }
 
+#[cfg(feature = "dtype-categorical")]
 impl Debug for CategoricalChunked {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         let limit = set_limit!(self);
@@ -288,43 +236,19 @@ impl Debug for Series {
             DataType::Float64 => {
                 format_array!(limit, f, self.f64().unwrap(), "f64", self.name(), "Series")
             }
-            DataType::Date32 => format_array!(
+            DataType::Date => format_array!(
                 limit,
                 f,
-                self.date32().unwrap(),
-                "date32",
+                self.date().unwrap(),
+                "date",
                 self.name(),
                 "Series"
             ),
-            DataType::Date64 => format_array!(
+            DataType::Datetime => format_array!(
                 limit,
                 f,
-                self.date64().unwrap(),
-                "date64",
-                self.name(),
-                "Series"
-            ),
-            DataType::Time64(TimeUnit::Nanosecond) => format_array!(
-                limit,
-                f,
-                self.time64_nanosecond().unwrap(),
-                "time64(ns)",
-                self.name(),
-                "Series"
-            ),
-            DataType::Duration(TimeUnit::Nanosecond) => format_array!(
-                limit,
-                f,
-                self.duration_nanosecond().unwrap(),
-                "duration(ns)",
-                self.name(),
-                "Series"
-            ),
-            DataType::Duration(TimeUnit::Millisecond) => format_array!(
-                limit,
-                f,
-                self.duration_millisecond().unwrap(),
-                "duration(ms)",
+                self.datetime().unwrap(),
+                "datetime",
                 self.name(),
                 "Series"
             ),
@@ -340,6 +264,7 @@ impl Debug for Series {
             DataType::Object(_) => {
                 format_object_array(limit, f, self.as_ref(), self.name(), "Series")
             }
+            #[cfg(feature = "dtype-categorical")]
             DataType::Categorical => format_array!(
                 limit,
                 f,
@@ -399,20 +324,27 @@ fn prepare_row(row: Vec<Cow<'_, str>>, n_first: usize, n_last: usize) -> Vec<Str
 impl Display for DataFrame {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         let height = self.height();
-        if !self.columns.iter().all(|s| s.len() == height) {
-            panic!("The columns lengths in the DataFrame are not equal.");
-        }
+        assert!(
+            self.columns.iter().all(|s| s.len() == height),
+            "The columns lengths in the DataFrame are not equal."
+        );
 
         let max_n_cols = std::env::var("POLARS_FMT_MAX_COLS")
             .unwrap_or_else(|_| "8".to_string())
             .parse()
             .unwrap_or(8);
         #[cfg(any(feature = "plain_fmt", feature = "pretty_fmt"))]
-        let max_n_rows = std::env::var("POLARS_FMT_MAX_ROWS")
-            .unwrap_or_else(|_| "8".to_string())
-            .parse()
-            .unwrap_or(8);
-
+        let max_n_rows = {
+            let max_n_rows = std::env::var("POLARS_FMT_MAX_ROWS")
+                .unwrap_or_else(|_| "8".to_string())
+                .parse()
+                .unwrap_or(8);
+            if max_n_rows < 2 {
+                2
+            } else {
+                max_n_rows
+            }
+        };
         let (n_first, n_last) = if self.width() > max_n_cols {
             ((max_n_cols + 1) / 2, max_n_cols / 2)
         } else {
@@ -569,54 +501,45 @@ impl Display for AnyValue<'_> {
             AnyValue::Float32(v) => fmt_float(f, width, *v),
             AnyValue::Float64(v) => fmt_float(f, width, *v),
             AnyValue::Boolean(v) => write!(f, "{}", *v),
-            AnyValue::Utf8(v) => write!(f, "{}", format!("\"{}\"", v)),
-            #[cfg(feature = "temporal")]
-            AnyValue::Date32(v) => write!(f, "{}", date32_as_datetime(*v).date()),
-            #[cfg(feature = "temporal")]
-            AnyValue::Date64(v) => write!(f, "{}", date64_as_datetime(*v)),
-            AnyValue::Time64(v, TimeUnit::Nanosecond) => {
-                write!(f, "{}", time64_nanosecond_as_time(*v))
+            AnyValue::Utf8(v) => write!(f, "{}", format_args!("\"{}\"", v)),
+            #[cfg(feature = "dtype-date")]
+            AnyValue::Date(v) => write!(f, "{}", date32_to_date(*v)),
+            #[cfg(feature = "dtype-datetime")]
+            AnyValue::Datetime(v) => write!(f, "{}", timestamp_ms_to_datetime(*v)),
+            #[cfg(feature = "dtype-time")]
+            AnyValue::Time(_) => {
+                let nt: chrono::NaiveTime = self.into();
+                write!(f, "{}", nt)
+            }
+            #[cfg(feature = "dtype-categorical")]
+            AnyValue::Categorical(idx, rev) => {
+                let s = rev.get(*idx);
+                write!(f, "\"{}\"", s)
             }
-            AnyValue::Duration(v, TimeUnit::Nanosecond) => write!(f, "{}", v),
-            AnyValue::Duration(v, TimeUnit::Millisecond) => write!(f, "{}", v),
             AnyValue::List(s) => write!(f, "{}", s.fmt_list()),
             #[cfg(feature = "object")]
             AnyValue::Object(_) => write!(f, "object"),
-            _ => unimplemented!(),
         }
     }
 }
 
-macro_rules! fmt_option {
-    ($opt:expr) => {{
-        match $opt {
-            Some(v) => format!("{}", v),
-            None => "null".to_string(),
-        }
-    }};
-}
-
 macro_rules! impl_fmt_list {
     ($self:ident) => {{
         match $self.len() {
             0 => format!("[]"),
-            1 => format!("[{}]", fmt_option!($self.get(0))),
-            2 => format!(
-                "[{}, {}]",
-                fmt_option!($self.get(0)),
-                fmt_option!($self.get(1))
-            ),
+            1 => format!("[{}]", $self.get_any_value(0)),
+            2 => format!("[{}, {}]", $self.get_any_value(0), $self.get_any_value(1)),
             3 => format!(
                 "[{}, {}, {}]",
-                fmt_option!($self.get(0)),
-                fmt_option!($self.get(1)),
-                fmt_option!($self.get(2))
+                $self.get_any_value(0),
+                $self.get_any_value(1),
+                $self.get_any_value(2)
             ),
             _ => format!(
                 "[{}, {}, ... {}]",
-                fmt_option!($self.get(0)),
-                fmt_option!($self.get(1)),
-                fmt_option!($self.get($self.len() - 1))
+                $self.get_any_value(0),
+                $self.get_any_value(1),
+                $self.get_any_value($self.len() - 1)
             ),
         }
     }};
@@ -654,31 +577,53 @@ impl FmtList for ListChunked {
     }
 }
 
+#[cfg(feature = "dtype-categorical")]
 impl FmtList for CategoricalChunked {
     fn fmt_list(&self) -> String {
         impl_fmt_list!(self)
     }
 }
 
-#[cfg(feature = "object")]
-impl<T> FmtList for ObjectChunked<T> {
+#[cfg(feature = "dtype-date")]
+impl FmtList for DateChunked {
     fn fmt_list(&self) -> String {
-        todo!()
+        impl_fmt_list!(self)
+    }
+}
+
+#[cfg(feature = "dtype-datetime")]
+impl FmtList for DatetimeChunked {
+    fn fmt_list(&self) -> String {
+        impl_fmt_list!(self)
+    }
+}
+
+#[cfg(feature = "dtype-time")]
+impl FmtList for TimeChunked {
+    fn fmt_list(&self) -> String {
+        impl_fmt_list!(self)
+    }
+}
+
+#[cfg(feature = "object")]
+impl<T: PolarsObject> FmtList for ObjectChunked<T> {
+    fn fmt_list(&self) -> String {
+        impl_fmt_list!(self)
     }
 }
 
 #[cfg(all(
     test,
     feature = "temporal",
-    feature = "dtype-date32",
-    feature = "dtype-date64"
+    feature = "dtype-date",
+    feature = "dtype-datetime"
 ))]
 mod test {
     use crate::prelude::*;
 
     #[test]
     fn test_fmt_list() {
-        let mut builder = ListPrimitiveChunkedBuilder::<Int32Type>::new("a", 10, 10);
+        let mut builder = ListPrimitiveChunkedBuilder::<i32>::new("a", 10, 10, DataType::Int32);
         builder.append_slice(Some(&[1, 2, 3]));
         builder.append_slice(None);
         let list = builder.finish().into_series();
@@ -696,12 +641,11 @@ Series: 'a' [list]
     }
 
     #[test]
-    #[cfg(feature = "dtype-time64-ns")]
     fn test_fmt_temporal() {
-        let s = Date32Chunked::new_from_opt_slice("date32", &[Some(1), None, Some(3)]);
+        let s = Int32Chunked::new_from_opt_slice("Date", &[Some(1), None, Some(3)]).into_date();
         assert_eq!(
             r#"shape: (3,)
-Series: 'date32' [date32]
+Series: 'Date' [date]
 [
 	1970-01-02
 	null
@@ -710,10 +654,11 @@ Series: 'date32' [date32]
             format!("{:?}", s.into_series())
         );
 
-        let s = Date64Chunked::new_from_opt_slice("", &[Some(1), None, Some(1_000_000_000_000)]);
+        let s = Int64Chunked::new_from_opt_slice("", &[Some(1), None, Some(1_000_000_000_000)])
+            .into_date();
         assert_eq!(
             r#"shape: (3,)
-Series: '' [date64]
+Series: '' [datetime]
 [
 	1970-01-01 00:00:00.001
 	null
@@ -721,28 +666,15 @@ Series: '' [date64]
 ]"#,
             format!("{:?}", s.into_series())
         );
-        let s = Time64NanosecondChunked::new_from_slice(
-            "",
-            &[1_000_000, 37_800_005_000_000, 86_399_210_000_000],
-        );
-        assert_eq!(
-            r#"shape: (3,)
-Series: '' [time64(ns)]
-[
-	00:00:00.001
-	10:30:00.005
-	23:59:59.210
-]"#,
-            format!("{:?}", s.into_series())
-        )
     }
+
     #[test]
     fn test_fmt_chunkedarray() {
-        let ca = Int32Chunked::new_from_opt_slice("date32", &[Some(1), None, Some(3)]);
+        let ca = Int32Chunked::new_from_opt_slice("Date", &[Some(1), None, Some(3)]);
         println!("{:?}", ca);
         assert_eq!(
             r#"shape: (3,)
-ChunkedArray: 'date32' [Int32]
+ChunkedArray: 'Date' [Int32]
 [
 	1
 	null
