@@ -7,41 +7,15 @@ use std::hash::Hash;
 
 use arrow::types::{simd::Simd, NativeType};
 
-use crate::chunked_array::kernels::take_agg::{
-    take_agg_no_null_primitive_iter_unchecked, take_agg_primitive_iter_unchecked,
-    take_agg_primitive_iter_unchecked_count_nulls,
-};
+#[cfg(feature = "object")]
+use crate::chunked_array::object::extension::create_extension;
 use crate::prelude::*;
+use crate::series::implementations::SeriesWrap;
 use crate::utils::NoNull;
 use arrow::buffer::MutableBuffer;
+use polars_arrow::kernels::take_agg::*;
 use polars_arrow::trusted_len::PushUnchecked;
-
-pub(crate) trait NumericAggSync {
-    fn agg_mean(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        None
-    }
-    fn agg_min(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        None
-    }
-    fn agg_max(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        None
-    }
-    fn agg_sum(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        None
-    }
-    fn agg_std(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        None
-    }
-    fn agg_var(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        None
-    }
-
-    /// Count the valid values. That is length - null_count
-    /// Used in partitioned aggregation to compute the mean values.
-    fn agg_valid_count(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        None
-    }
-}
+use std::ops::Deref;
 
 fn agg_helper<T, F>(groups: &[(u32, Vec<u32>)], f: F) -> Option<Series>
 where
@@ -53,33 +27,256 @@ where
     Some(ca.into_series())
 }
 
-impl NumericAggSync for BooleanChunked {
-    fn agg_min(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        self.cast::<UInt32Type>().unwrap().agg_min(groups)
+impl BooleanChunked {
+    pub(crate) fn agg_min(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        self.cast(&DataType::UInt32).unwrap().agg_min(groups)
     }
-    fn agg_max(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        self.cast::<UInt32Type>().unwrap().agg_max(groups)
+    pub(crate) fn agg_max(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        self.cast(&DataType::UInt32).unwrap().agg_max(groups)
     }
-    fn agg_sum(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        self.cast::<UInt32Type>().unwrap().agg_sum(groups)
+    pub(crate) fn agg_sum(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        self.cast(&DataType::UInt32).unwrap().agg_sum(groups)
     }
 }
-impl NumericAggSync for Utf8Chunked {}
-impl NumericAggSync for ListChunked {}
-impl NumericAggSync for CategoricalChunked {}
-#[cfg(feature = "object")]
-impl<T> NumericAggSync for ObjectChunked<T> {}
 
-impl<T> NumericAggSync for ChunkedArray<T>
+impl<T> ChunkedArray<T>
+where
+    ChunkedArray<T>: ChunkTake,
+    T: PolarsDataType + Sync,
+{
+    #[cfg(feature = "lazy")]
+    pub(crate) fn agg_valid_count(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        agg_helper::<UInt32Type, _>(groups, |(_first, idx)| {
+            debug_assert!(idx.len() <= self.len());
+            if idx.is_empty() {
+                None
+            } else if !self.has_validity() {
+                Some(idx.len() as u32)
+            } else {
+                let take = unsafe { self.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
+                Some((take.len() - take.null_count()) as u32)
+            }
+        })
+    }
+}
+
+impl<T> ChunkedArray<T>
 where
     T: PolarsNumericType + Sync,
-    T::Native: NativeType + PartialOrd + Num + NumCast + Zero + Simd + Bounded,
+    T::Native:
+        NativeType + PartialOrd + Num + NumCast + Zero + Simd + Bounded + std::iter::Sum<T::Native>,
     <T::Native as Simd>::Simd: std::ops::Add<Output = <T::Native as Simd>::Simd>
         + arrow::compute::aggregate::Sum<T::Native>
         + arrow::compute::aggregate::SimdOrd<T::Native>,
     ChunkedArray<T>: IntoSeries,
 {
-    fn agg_mean(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+    pub(crate) fn agg_min(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        agg_helper::<T, _>(groups, |(first, idx)| {
+            debug_assert!(idx.len() <= self.len());
+            if idx.is_empty() {
+                None
+            } else if idx.len() == 1 {
+                self.get(*first as usize)
+            } else {
+                match (self.has_validity(), self.chunks.len()) {
+                    (false, 1) => Some(unsafe {
+                        take_agg_no_null_primitive_iter_unchecked(
+                            self.downcast_iter().next().unwrap(),
+                            idx.iter().map(|i| *i as usize),
+                            |a, b| if a < b { a } else { b },
+                            T::Native::max_value(),
+                        )
+                    }),
+                    (_, 1) => unsafe {
+                        take_agg_primitive_iter_unchecked::<T::Native, _, _>(
+                            self.downcast_iter().next().unwrap(),
+                            idx.iter().map(|i| *i as usize),
+                            |a, b| if a < b { a } else { b },
+                            T::Native::max_value(),
+                        )
+                    },
+                    _ => {
+                        let take =
+                            unsafe { self.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
+                        take.min()
+                    }
+                }
+            }
+        })
+    }
+
+    pub(crate) fn agg_max(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        agg_helper::<T, _>(groups, |(first, idx)| {
+            debug_assert!(idx.len() <= self.len());
+            if idx.is_empty() {
+                None
+            } else if idx.len() == 1 {
+                self.get(*first as usize)
+            } else {
+                match (self.has_validity(), self.chunks.len()) {
+                    (false, 1) => Some(unsafe {
+                        take_agg_no_null_primitive_iter_unchecked(
+                            self.downcast_iter().next().unwrap(),
+                            idx.iter().map(|i| *i as usize),
+                            |a, b| if a > b { a } else { b },
+                            T::Native::min_value(),
+                        )
+                    }),
+                    (_, 1) => unsafe {
+                        take_agg_primitive_iter_unchecked::<T::Native, _, _>(
+                            self.downcast_iter().next().unwrap(),
+                            idx.iter().map(|i| *i as usize),
+                            |a, b| if a > b { a } else { b },
+                            T::Native::min_value(),
+                        )
+                    },
+                    _ => {
+                        let take =
+                            unsafe { self.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
+                        take.max()
+                    }
+                }
+            }
+        })
+    }
+
+    pub(crate) fn agg_sum(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        agg_helper::<T, _>(groups, |(first, idx)| {
+            debug_assert!(idx.len() <= self.len());
+            if idx.is_empty() {
+                None
+            } else if idx.len() == 1 {
+                self.get(*first as usize)
+            } else {
+                match (self.has_validity(), self.chunks.len()) {
+                    (false, 1) => Some(unsafe {
+                        take_agg_no_null_primitive_iter_unchecked(
+                            self.downcast_iter().next().unwrap(),
+                            idx.iter().map(|i| *i as usize),
+                            |a, b| a + b,
+                            T::Native::zero(),
+                        )
+                    }),
+                    (_, 1) => unsafe {
+                        take_agg_primitive_iter_unchecked::<T::Native, _, _>(
+                            self.downcast_iter().next().unwrap(),
+                            idx.iter().map(|i| *i as usize),
+                            |a, b| a + b,
+                            T::Native::zero(),
+                        )
+                    },
+                    _ => {
+                        let take =
+                            unsafe { self.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
+                        take.sum()
+                    }
+                }
+            }
+        })
+    }
+}
+
+impl<T> SeriesWrap<ChunkedArray<T>>
+where
+    T: PolarsFloatType,
+    ChunkedArray<T>: IntoSeries,
+    T::Native: NativeType + PartialOrd + Num + NumCast + Simd + std::iter::Sum<T::Native>,
+    <T::Native as Simd>::Simd: std::ops::Add<Output = <T::Native as Simd>::Simd>
+        + arrow::compute::aggregate::Sum<T::Native>
+        + arrow::compute::aggregate::SimdOrd<T::Native>,
+{
+    pub(crate) fn agg_mean(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        agg_helper::<T, _>(groups, |(first, idx)| {
+            // this can fail due to a bug in lazy code.
+            // here users can create filters in aggregations
+            // and thereby creating shorter columns than the original group tuples.
+            // the group tuples are modified, but if that's done incorrect there can be out of bounds
+            // access
+            debug_assert!(idx.len() <= self.len());
+            let out = if idx.is_empty() {
+                None
+            } else if idx.len() == 1 {
+                self.get(*first as usize).map(|sum| sum.to_f64().unwrap())
+            } else {
+                match (self.has_validity(), self.chunks.len()) {
+                    (false, 1) => unsafe {
+                        take_agg_no_null_primitive_iter_unchecked(
+                            self.downcast_iter().next().unwrap(),
+                            idx.iter().map(|i| *i as usize),
+                            |a, b| a + b,
+                            T::Native::zero(),
+                        )
+                    }
+                    .to_f64()
+                    .map(|sum| sum / idx.len() as f64),
+                    (_, 1) => unsafe {
+                        take_agg_primitive_iter_unchecked_count_nulls::<T::Native, _, _>(
+                            self.downcast_iter().next().unwrap(),
+                            idx.iter().map(|i| *i as usize),
+                            |a, b| a + b,
+                            T::Native::zero(),
+                        )
+                    }
+                    .map(|(sum, null_count)| {
+                        sum.to_f64()
+                            .map(|sum| sum / (idx.len() as f64 - null_count as f64))
+                            .unwrap()
+                    }),
+                    _ => {
+                        let take =
+                            unsafe { self.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
+                        let opt_sum: Option<T::Native> = take.sum();
+                        opt_sum.map(|sum| sum.to_f64().unwrap() / idx.len() as f64)
+                    }
+                }
+            };
+            out.map(|flt| NumCast::from(flt).unwrap())
+        })
+    }
+
+    pub(crate) fn agg_var(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        let ca = &self.0;
+        agg_helper::<T, _>(groups, |(_first, idx)| {
+            debug_assert!(idx.len() <= ca.len());
+            if idx.is_empty() {
+                return None;
+            }
+            let take = unsafe { ca.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
+            take.into_series()
+                .var_as_series()
+                .unpack::<T>()
+                .unwrap()
+                .get(0)
+        })
+    }
+    pub(crate) fn agg_std(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        let ca = &self.0;
+        agg_helper::<T, _>(groups, |(_first, idx)| {
+            debug_assert!(idx.len() <= ca.len());
+            if idx.is_empty() {
+                return None;
+            }
+            let take = unsafe { ca.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
+            take.into_series()
+                .std_as_series()
+                .unpack::<T>()
+                .unwrap()
+                .get(0)
+        })
+    }
+}
+
+impl<T> ChunkedArray<T>
+where
+    T: PolarsIntegerType,
+    ChunkedArray<T>: IntoSeries,
+    T::Native:
+        NativeType + PartialOrd + Num + NumCast + Zero + Simd + Bounded + std::iter::Sum<T::Native>,
+    <T::Native as Simd>::Simd: std::ops::Add<Output = <T::Native as Simd>::Simd>
+        + arrow::compute::aggregate::Sum<T::Native>
+        + arrow::compute::aggregate::SimdOrd<T::Native>,
+{
+    pub(crate) fn agg_mean(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
         agg_helper::<Float64Type, _>(groups, |(first, idx)| {
             // this can fail due to a bug in lazy code.
             // here users can create filters in aggregations
@@ -92,8 +289,8 @@ where
             } else if idx.len() == 1 {
                 self.get(*first as usize).map(|sum| sum.to_f64().unwrap())
             } else {
-                match (self.null_count(), self.chunks.len()) {
-                    (0, 1) => unsafe {
+                match (self.has_validity(), self.chunks.len()) {
+                    (false, 1) => unsafe {
                         take_agg_no_null_primitive_iter_unchecked(
                             self.downcast_iter().next().unwrap(),
                             idx.iter().map(|i| *i as usize),
@@ -104,7 +301,7 @@ where
                     .to_f64()
                     .map(|sum| sum / idx.len() as f64),
                     (_, 1) => unsafe {
-                        take_agg_primitive_iter_unchecked_count_nulls::<T, _, _>(
+                        take_agg_primitive_iter_unchecked_count_nulls::<T::Native, _, _>(
                             self.downcast_iter().next().unwrap(),
                             idx.iter().map(|i| *i as usize),
                             |a, b| a + b,
@@ -127,111 +324,7 @@ where
         })
     }
 
-    fn agg_min(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        agg_helper::<T, _>(groups, |(first, idx)| {
-            debug_assert!(idx.len() <= self.len());
-            if idx.is_empty() {
-                None
-            } else if idx.len() == 1 {
-                self.get(*first as usize)
-            } else {
-                match (self.null_count(), self.chunks.len()) {
-                    (0, 1) => Some(unsafe {
-                        take_agg_no_null_primitive_iter_unchecked(
-                            self.downcast_iter().next().unwrap(),
-                            idx.iter().map(|i| *i as usize),
-                            |a, b| if a < b { a } else { b },
-                            T::Native::max_value(),
-                        )
-                    }),
-                    (_, 1) => unsafe {
-                        take_agg_primitive_iter_unchecked::<T, _, _>(
-                            self.downcast_iter().next().unwrap(),
-                            idx.iter().map(|i| *i as usize),
-                            |a, b| if a < b { a } else { b },
-                            T::Native::max_value(),
-                        )
-                    },
-                    _ => {
-                        let take =
-                            unsafe { self.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
-                        take.min()
-                    }
-                }
-            }
-        })
-    }
-
-    fn agg_max(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        agg_helper::<T, _>(groups, |(first, idx)| {
-            debug_assert!(idx.len() <= self.len());
-            if idx.is_empty() {
-                None
-            } else if idx.len() == 1 {
-                self.get(*first as usize)
-            } else {
-                match (self.null_count(), self.chunks.len()) {
-                    (0, 1) => Some(unsafe {
-                        take_agg_no_null_primitive_iter_unchecked(
-                            self.downcast_iter().next().unwrap(),
-                            idx.iter().map(|i| *i as usize),
-                            |a, b| if a > b { a } else { b },
-                            T::Native::min_value(),
-                        )
-                    }),
-                    (_, 1) => unsafe {
-                        take_agg_primitive_iter_unchecked::<T, _, _>(
-                            self.downcast_iter().next().unwrap(),
-                            idx.iter().map(|i| *i as usize),
-                            |a, b| if a > b { a } else { b },
-                            T::Native::min_value(),
-                        )
-                    },
-                    _ => {
-                        let take =
-                            unsafe { self.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
-                        take.max()
-                    }
-                }
-            }
-        })
-    }
-
-    fn agg_sum(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        agg_helper::<T, _>(groups, |(first, idx)| {
-            debug_assert!(idx.len() <= self.len());
-            if idx.is_empty() {
-                None
-            } else if idx.len() == 1 {
-                self.get(*first as usize)
-            } else {
-                match (self.null_count(), self.chunks.len()) {
-                    (0, 1) => Some(unsafe {
-                        take_agg_no_null_primitive_iter_unchecked(
-                            self.downcast_iter().next().unwrap(),
-                            idx.iter().map(|i| *i as usize),
-                            |a, b| a + b,
-                            T::Native::zero(),
-                        )
-                    }),
-                    (_, 1) => unsafe {
-                        take_agg_primitive_iter_unchecked::<T, _, _>(
-                            self.downcast_iter().next().unwrap(),
-                            idx.iter().map(|i| *i as usize),
-                            |a, b| a + b,
-                            T::Native::zero(),
-                        )
-                    },
-                    _ => {
-                        let take =
-                            unsafe { self.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
-                        take.sum()
-                    }
-                }
-            }
-        })
-    }
-    fn agg_var(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+    pub(crate) fn agg_var(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
         agg_helper::<Float64Type, _>(groups, |(_first, idx)| {
             debug_assert!(idx.len() <= self.len());
             if idx.is_empty() {
@@ -245,7 +338,7 @@ where
                 .get(0)
         })
     }
-    fn agg_std(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+    pub(crate) fn agg_std(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
         agg_helper::<Float64Type, _>(groups, |(_first, idx)| {
             debug_assert!(idx.len() <= self.len());
             if idx.is_empty() {
@@ -259,178 +352,32 @@ where
                 .get(0)
         })
     }
-    #[cfg(feature = "lazy")]
-    fn agg_valid_count(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        agg_helper::<UInt32Type, _>(groups, |(_first, idx)| {
-            debug_assert!(idx.len() <= self.len());
+}
+
+impl<T> ChunkedArray<T>
+where
+    ChunkedArray<T>: ChunkTake + IntoSeries,
+{
+    pub(crate) fn agg_first(&self, groups: &[(u32, Vec<u32>)]) -> Series {
+        let iter = groups.iter().map(|(first, idx)| {
             if idx.is_empty() {
                 None
-            } else if self.null_count() == 0 {
-                Some(idx.len() as u32)
             } else {
-                let take = unsafe { self.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
-                Some((take.len() - take.null_count()) as u32)
+                Some(*first as usize)
             }
-        })
+        });
+        unsafe { self.take_unchecked(iter.into()) }.into_series()
     }
-}
 
-pub(crate) trait AggFirst {
-    fn agg_first(&self, _groups: &[(u32, Vec<u32>)]) -> Series;
-}
-
-macro_rules! impl_agg_first {
-    ($self:ident, $groups:ident, $ca_type:ty) => {{
-        let mut ca = $groups
-            .iter()
-            .map(|(first, idx)| {
-                debug_assert!(idx.len() <= $self.len());
-                if idx.is_empty() {
-                    return None;
-                }
-                $self.get(*first as usize)
-            })
-            .collect::<$ca_type>();
-
-        ca.categorical_map = $self.categorical_map.clone();
-        ca.into_series()
-    }};
-}
-
-impl<T> AggFirst for ChunkedArray<T>
-where
-    T: PolarsNumericType + Send,
-    ChunkedArray<T>: IntoSeries,
-{
-    fn agg_first(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        impl_agg_first!(self, groups, ChunkedArray<T>)
-    }
-}
-
-impl AggFirst for BooleanChunked {
-    fn agg_first(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        impl_agg_first!(self, groups, BooleanChunked)
-    }
-}
-
-impl AggFirst for Utf8Chunked {
-    fn agg_first(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        impl_agg_first!(self, groups, Utf8Chunked)
-    }
-}
-
-impl AggFirst for ListChunked {
-    fn agg_first(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        impl_agg_first!(self, groups, ListChunked)
-    }
-}
-
-impl AggFirst for CategoricalChunked {
-    fn agg_first(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        let out = self
-            .cast::<UInt32Type>()
-            .unwrap()
-            .agg_first(groups)
-            .cast::<CategoricalType>()
-            .unwrap();
-
-        debug_assert!(out.categorical().unwrap().categorical_map.is_some());
-        out
-    }
-}
-
-#[cfg(feature = "object")]
-impl<T: PolarsObject> AggFirst for ObjectChunked<T> {
-    fn agg_first(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        let ca: Self = groups
-            .iter()
-            .map(|(first, idx)| {
-                if idx.is_empty() {
-                    return None;
-                }
-                self.get(*first as usize).cloned()
-            })
-            .collect();
-
-        ca.into_series()
-    }
-}
-
-pub(crate) trait AggLast {
-    fn agg_last(&self, _groups: &[(u32, Vec<u32>)]) -> Series;
-}
-
-macro_rules! impl_agg_last {
-    ($self:ident, $groups:ident, $ca_type:ty) => {{
-        let mut ca = $groups
-            .iter()
-            .map(|(_first, idx)| {
-                debug_assert!(idx.len() <= $self.len());
-                if idx.is_empty() {
-                    return None;
-                }
-
-                $self.get(idx[idx.len() - 1] as usize)
-            })
-            .collect::<$ca_type>();
-
-        ca.categorical_map = $self.categorical_map.clone();
-        ca.into_series()
-    }};
-}
-
-impl<T> AggLast for ChunkedArray<T>
-where
-    T: PolarsNumericType + Send,
-    ChunkedArray<T>: IntoSeries,
-{
-    fn agg_last(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        impl_agg_last!(self, groups, ChunkedArray<T>)
-    }
-}
-
-impl AggLast for BooleanChunked {
-    fn agg_last(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        impl_agg_last!(self, groups, BooleanChunked)
-    }
-}
-
-impl AggLast for Utf8Chunked {
-    fn agg_last(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        impl_agg_last!(self, groups, Utf8Chunked)
-    }
-}
-
-impl AggLast for CategoricalChunked {
-    fn agg_last(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        self.cast::<UInt32Type>()
-            .unwrap()
-            .agg_last(groups)
-            .cast::<CategoricalType>()
-            .unwrap()
-    }
-}
-
-impl AggLast for ListChunked {
-    fn agg_last(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        impl_agg_last!(self, groups, ListChunked)
-    }
-}
-
-#[cfg(feature = "object")]
-impl<T: PolarsObject> AggLast for ObjectChunked<T> {
-    fn agg_last(&self, groups: &[(u32, Vec<u32>)]) -> Series {
-        let ca: Self = groups
-            .iter()
-            .map(|(_first, idx)| {
-                if idx.is_empty() {
-                    return None;
-                }
-                self.get((idx.len() - 1) as usize).cloned()
-            })
-            .collect();
-
-        ca.into_series()
+    pub(crate) fn agg_last(&self, groups: &[(u32, Vec<u32>)]) -> Series {
+        let iter = groups.iter().map(|(_, idx)| {
+            if idx.is_empty() {
+                None
+            } else {
+                Some(idx[idx.len() - 1] as usize)
+            }
+        });
+        unsafe { self.take_unchecked(iter.into()) }.into_series()
     }
 }
 
@@ -451,21 +398,12 @@ macro_rules! impl_agg_n_unique {
                 }
                 let taker = $self.take_rand();
 
-                if $self.null_count() == 0 {
-                    let mut set = HashSet::with_hasher(RandomState::new());
-                    for i in idx {
-                        let v = unsafe { taker.get_unchecked(*i as usize) };
-                        set.insert(v);
-                    }
-                    set.len() as u32
-                } else {
-                    let mut set = HashSet::with_hasher(RandomState::new());
-                    for i in idx {
-                        let opt_v = taker.get(*i as usize);
-                        set.insert(opt_v);
-                    }
-                    set.len() as u32
+                let mut set = HashSet::with_hasher(RandomState::new());
+                for i in idx {
+                    let v = unsafe { taker.get_unchecked(*i as usize) };
+                    set.insert(v);
                 }
+                set.len() as u32
             })
             .collect::<$ca_type>()
             .into_inner()
@@ -482,13 +420,21 @@ where
     }
 }
 
-// todo! could use mantissa method here
-impl AggNUnique for Float32Chunked {}
-impl AggNUnique for Float64Chunked {}
+impl AggNUnique for Float32Chunked {
+    fn agg_n_unique(&self, groups: &[(u32, Vec<u32>)]) -> Option<UInt32Chunked> {
+        self.bit_repr_small().agg_n_unique(groups)
+    }
+}
+impl AggNUnique for Float64Chunked {
+    fn agg_n_unique(&self, groups: &[(u32, Vec<u32>)]) -> Option<UInt32Chunked> {
+        self.bit_repr_large().agg_n_unique(groups)
+    }
+}
 impl AggNUnique for ListChunked {}
+#[cfg(feature = "dtype-categorical")]
 impl AggNUnique for CategoricalChunked {
     fn agg_n_unique(&self, groups: &[(u32, Vec<u32>)]) -> Option<UInt32Chunked> {
-        self.cast::<UInt32Type>()
+        self.cast(&DataType::UInt32)
             .unwrap()
             .agg_n_unique(groups)
             .map(|mut ca| {
@@ -557,8 +503,12 @@ where
                 ListArray::<i64>::from_data(data_type, offsets.into(), Arc::new(array), None)
             }
             _ => {
-                let mut builder =
-                    ListPrimitiveChunkedBuilder::<T>::new(self.name(), groups.len(), self.len());
+                let mut builder = ListPrimitiveChunkedBuilder::<T::Native>::new(
+                    self.name(),
+                    groups.len(),
+                    self.len(),
+                    self.dtype().clone(),
+                );
                 for (_first, idx) in groups {
                     let s = unsafe {
                         self.take_unchecked(idx.iter().map(|i| *i as usize).into())
@@ -605,9 +555,130 @@ impl AggList for Utf8Chunked {
     }
 }
 
-impl AggList for ListChunked {}
+impl AggList for ListChunked {
+    fn agg_list(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        let mut can_fast_explode = true;
+        let mut offsets = MutableBuffer::<i64>::with_capacity(groups.len() + 1);
+        let mut length_so_far = 0i64;
+        offsets.push(length_so_far);
+
+        let mut list_values = Vec::with_capacity(groups.len());
+        groups.iter().for_each(|(_, idx)| {
+            let idx_len = idx.len();
+            if idx_len == 0 {
+                can_fast_explode = false;
+            }
+
+            length_so_far += idx_len as i64;
+            // Safety:
+            // group tuples are in bounds
+            unsafe {
+                let mut s = self.take_unchecked((idx.iter().map(|idx| *idx as usize)).into());
+                let arr = s.chunks.pop().unwrap();
+                list_values.push(arr);
+
+                // Safety:
+                // we know that offsets has allocated enough slots
+                offsets.push_unchecked(length_so_far);
+            }
+        });
+        if groups.is_empty() {
+            list_values.push(self.chunks[0].slice(0, 0).into())
+        }
+        let arrays = list_values.iter().map(|arr| &**arr).collect::<Vec<_>>();
+        let list_values: ArrayRef = arrow::compute::concatenate::concatenate(&arrays)
+            .unwrap()
+            .into();
+        let data_type = ListArray::<i64>::default_datatype(list_values.data_type().clone());
+        let arr = Arc::new(ListArray::<i64>::from_data(
+            data_type,
+            offsets.into(),
+            list_values,
+            None,
+        )) as ArrayRef;
+        let mut listarr = ListChunked::new_from_chunks(self.name(), vec![arr]);
+        if can_fast_explode {
+            listarr.set_fast_explode()
+        }
+        Some(listarr.into_series())
+    }
+}
+impl AggList for CategoricalChunked {
+    fn agg_list(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        match self.deref().agg_list(groups) {
+            None => None,
+            Some(s) => {
+                let ca = s.list().unwrap();
+                let mut out = ListChunked::new_from_chunks(ca.name(), ca.chunks.clone());
+                out.field = Arc::new(Field::new(
+                    ca.name(),
+                    DataType::List(Box::new(DataType::Categorical)),
+                ));
+                out.categorical_map = self.categorical_map.clone();
+                out.bit_settings = ca.bit_settings;
+                Some(out.into_series())
+            }
+        }
+    }
+}
 #[cfg(feature = "object")]
-impl<T> AggList for ObjectChunked<T> {}
+impl<T: PolarsObject> AggList for ObjectChunked<T> {
+    fn agg_list(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        let mut can_fast_explode = true;
+        let mut offsets = MutableBuffer::<i64>::with_capacity(groups.len() + 1);
+        let mut length_so_far = 0i64;
+        offsets.push(length_so_far);
+
+        let iter = groups
+            .iter()
+            .map(|(_, idx)| {
+                // Safety:
+                // group tuples always in bounds
+                let group_vals =
+                    unsafe { self.take_unchecked((idx.iter().map(|idx| *idx as usize)).into()) };
+
+                let idx_len = idx.len();
+                if idx_len == 0 {
+                    can_fast_explode = false;
+                }
+                length_so_far += idx_len as i64;
+                // Safety:
+                // we know that offsets has allocated enough slots
+                unsafe {
+                    offsets.push_unchecked(length_so_far);
+                }
+
+                let arr = group_vals.downcast_iter().next().unwrap().clone();
+                arr.into_iter_cloned()
+            })
+            .flatten()
+            .trust_my_length(self.len());
+
+        let mut pe = create_extension(iter);
+
+        // Safety:
+        // this is safe because we just created the PolarsExtension
+        // meaning that the sentinel is heap allocated and the dereference of the
+        // pointer does not fail
+        unsafe { pe.set_to_series_fn::<T>(self.name()) };
+        let extension_array = Arc::new(pe.take_and_forget()) as ArrayRef;
+        let extension_dtype = extension_array.data_type();
+
+        let data_type = ListArray::<i64>::default_datatype(extension_dtype.clone());
+        let arr = Arc::new(ListArray::<i64>::from_data(
+            data_type,
+            offsets.into(),
+            extension_array,
+            None,
+        )) as ArrayRef;
+
+        let mut listarr = ListChunked::new_from_chunks(self.name(), vec![arr]);
+        if can_fast_explode {
+            listarr.set_fast_explode()
+        }
+        Some(listarr.into_series())
+    }
+}
 
 pub(crate) trait AggQuantile {
     fn agg_quantile(&self, _groups: &[(u32, Vec<u32>)], _quantile: f64) -> Option<Series> {
@@ -622,7 +693,7 @@ pub(crate) trait AggQuantile {
 impl<T> AggQuantile for ChunkedArray<T>
 where
     T: PolarsNumericType + Sync,
-    T::Native: PartialOrd + Num + NumCast + Zero + Simd,
+    T::Native: PartialOrd + Num + NumCast + Zero + Simd + std::iter::Sum<T::Native>,
     <T::Native as Simd>::Simd: std::ops::Add<Output = <T::Native as Simd>::Simd>
         + arrow::compute::aggregate::Sum<T::Native>
         + arrow::compute::aggregate::SimdOrd<T::Native>,
@@ -654,6 +725,7 @@ where
 impl AggQuantile for Utf8Chunked {}
 impl AggQuantile for BooleanChunked {}
 impl AggQuantile for ListChunked {}
+#[cfg(feature = "dtype-categorical")]
 impl AggQuantile for CategoricalChunked {}
 #[cfg(feature = "object")]
 impl<T> AggQuantile for ObjectChunked<T> {}
