@@ -7,14 +7,16 @@ use crate::utils::Wrap;
 use crate::utils::{
     accumulate_dataframes_vertical, copy_from_slice_unchecked, set_partition_size, split_ca, NoNull,
 };
-use crate::vector_hasher::{AsU64, StrHash};
+use crate::vector_hasher::{get_null_hash_value, AsU64, StrHash};
 use crate::POOL;
-use ahash::RandomState;
+use ahash::{CallHasher, RandomState};
 use hashbrown::HashMap;
 use num::NumCast;
 use rayon::prelude::*;
 use std::fmt::Debug;
-use std::hash::{BuildHasher, Hash, Hasher};
+use std::hash::Hash;
+#[cfg(feature = "dtype-categorical")]
+use std::ops::Deref;
 
 pub mod aggregations;
 pub(crate) mod hashing;
@@ -47,17 +49,20 @@ where
     T::Native: Hash + Eq + Send + AsU64,
     Option<T::Native>: AsU64,
 {
+    #[cfg(feature = "dtype-categorical")]
     let group_size_hint = if let Some(m) = &ca.categorical_map {
         ca.len() / m.len()
     } else {
         0
     };
+    #[cfg(not(feature = "dtype-categorical"))]
+    let group_size_hint = 0;
     if multithreaded && group_multithreaded(ca) {
         let n_partitions = set_partition_size() as u64;
 
         // use the arrays as iterators
         if ca.chunks.len() == 1 {
-            if ca.null_count() == 0 {
+            if !ca.has_validity() {
                 let keys = vec![ca.cont_slice().unwrap()];
                 groupby_threaded_num(keys, group_size_hint, n_partitions)
             } else {
@@ -68,14 +73,14 @@ where
                 groupby_threaded_num(keys, group_size_hint, n_partitions)
             }
             // use the polars-iterators
-        } else if ca.null_count() == 0 {
+        } else if !ca.has_validity() {
             let keys = vec![ca.into_no_null_iter().collect::<Vec<_>>()];
             groupby_threaded_num(keys, group_size_hint, n_partitions)
         } else {
             let keys = vec![ca.into_iter().collect::<Vec<_>>()];
             groupby_threaded_num(keys, group_size_hint, n_partitions)
         }
-    } else if ca.null_count() == 0 {
+    } else if !ca.has_validity() {
         groupby(ca.into_no_null_iter())
     } else {
         groupby(ca.into_iter())
@@ -112,37 +117,41 @@ where
                 num_group_tuples(&ca, multithreaded)
             }
             _ => {
-                let ca = self.cast::<UInt32Type>().unwrap();
-                num_group_tuples(&ca, multithreaded)
+                let ca = self.cast(&DataType::UInt32).unwrap();
+                let ca = ca.u32().unwrap();
+                num_group_tuples(ca, multithreaded)
             }
         }
     }
 }
 impl IntoGroupTuples for BooleanChunked {
     fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
-        self.cast::<UInt32Type>()
-            .unwrap()
-            .group_tuples(multithreaded)
+        let ca = self.cast(&DataType::UInt32).unwrap();
+        let ca = ca.u32().unwrap();
+        ca.group_tuples(multithreaded)
     }
 }
 
 impl IntoGroupTuples for Utf8Chunked {
     fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
         let hb = RandomState::default();
+        let null_h = get_null_hash_value(hb.clone());
+
         if multithreaded {
             let n_partitions = set_partition_size();
 
-            let splitted = split_ca(self, n_partitions).unwrap();
+            let split = split_ca(self, n_partitions).unwrap();
 
             let str_hashes = POOL.install(|| {
-                splitted
+                split
                     .par_iter()
                     .map(|ca| {
                         ca.into_iter()
                             .map(|opt_s| {
-                                let mut state = hb.build_hasher();
-                                opt_s.hash(&mut state);
-                                let hash = state.finish();
+                                let hash = match opt_s {
+                                    Some(s) => str::get_hash(s, &hb),
+                                    None => null_h,
+                                };
                                 StrHash::new(opt_s, hash)
                             })
                             .collect::<Vec<_>>()
@@ -154,9 +163,10 @@ impl IntoGroupTuples for Utf8Chunked {
             let str_hashes = self
                 .into_iter()
                 .map(|opt_s| {
-                    let mut state = hb.build_hasher();
-                    opt_s.hash(&mut state);
-                    let hash = state.finish();
+                    let hash = match opt_s {
+                        Some(s) => str::get_hash(s, &hb),
+                        None => null_h,
+                    };
                     StrHash::new(opt_s, hash)
                 })
                 .collect::<Vec<_>>();
@@ -165,11 +175,10 @@ impl IntoGroupTuples for Utf8Chunked {
     }
 }
 
+#[cfg(feature = "dtype-categorical")]
 impl IntoGroupTuples for CategoricalChunked {
     fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
-        self.cast::<UInt32Type>()
-            .unwrap()
-            .group_tuples(multithreaded)
+        self.deref().group_tuples(multithreaded)
     }
 }
 
@@ -297,7 +306,6 @@ fn pack_u32_u64_tuples(opt_l: Option<u32>, opt_r: Option<u64>) -> [u8; 13] {
 
 impl DataFrame {
     pub fn groupby_with_series(&self, by: Vec<Series>, multithreaded: bool) -> Result<GroupBy> {
-        use polars_arrow::utils::CustomIterTools;
         macro_rules! finish_packed_bit_path {
             ($ca0:expr, $ca1:expr, $pack_fn:expr) => {{
                 let n_partitions = set_partition_size();
@@ -342,15 +350,16 @@ impl DataFrame {
         let keys_df = DataFrame::new(
             by.iter()
                 .map(|s| match s.dtype() {
-                    Categorical | Int8 | UInt8 | Int16 | UInt16 => s.cast::<UInt32Type>().unwrap(),
+                    Categorical | Int8 | UInt8 | Int16 | UInt16 => {
+                        s.cast(&DataType::UInt32).unwrap()
+                    }
                     Float32 => s.bit_repr_small().into_series(),
                     // otherwise we use the vec hash for float
-                    #[cfg(feature = "dtype-u64")]
                     Float64 => s.bit_repr_large().into_series(),
                     _ => {
                         // is date like
                         if !s.is_numeric() && s.is_numeric_physical() {
-                            s.to_physical_repr()
+                            s.to_physical_repr().into_owned()
                         } else {
                             s.clone()
                         }
@@ -450,7 +459,7 @@ impl DataFrame {
 /// // date format
 /// let fmt = "%Y-%m-%d";
 /// // create date series
-/// let s0 = Date32Chunked::parse_from_str_slice("date", dates, fmt)
+/// let s0 = DateChunked::parse_from_str_slice("date", dates, fmt)
 ///         .into_series();
 /// // create temperature series
 /// let s1 = Series::new("temp", [20, 10, 7, 9, 1].as_ref());
@@ -467,7 +476,7 @@ impl DataFrame {
 /// +------------+------+------+
 /// | date       | temp | rain |
 /// | ---        | ---  | ---  |
-/// | date32     | i32  | f64  |
+/// | Date     | i32  | f64  |
 /// +============+======+======+
 /// | 2020-08-21 | 20   | 0.2  |
 /// +------------+------+------+
@@ -584,7 +593,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
     /// +------------+-----------+-----------+
     /// | date       | temp_mean | rain_mean |
     /// | ---        | ---       | ---       |
-    /// | date32     | f64       | f64       |
+    /// | Date     | f64       | f64       |
     /// +============+===========+===========+
     /// | 2020-08-23 | 9         | 0.1       |
     /// +------------+-----------+-----------+
@@ -623,7 +632,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
     /// +------------+----------+
     /// | date       | temp_sum |
     /// | ---        | ---      |
-    /// | date32     | i32      |
+    /// | Date     | i32      |
     /// +============+==========+
     /// | 2020-08-23 | 9        |
     /// +------------+----------+
@@ -662,7 +671,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
     /// +------------+----------+
     /// | date       | temp_min |
     /// | ---        | ---      |
-    /// | date32     | i32      |
+    /// | Date     | i32      |
     /// +============+==========+
     /// | 2020-08-23 | 9        |
     /// +------------+----------+
@@ -700,7 +709,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
     /// +------------+----------+
     /// | date       | temp_max |
     /// | ---        | ---      |
-    /// | date32     | i32      |
+    /// | Date     | i32      |
     /// +============+==========+
     /// | 2020-08-23 | 9        |
     /// +------------+----------+
@@ -738,7 +747,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
     /// +------------+------------+
     /// | date       | temp_first |
     /// | ---        | ---        |
-    /// | date32     | i32        |
+    /// | Date     | i32        |
     /// +============+============+
     /// | 2020-08-23 | 9          |
     /// +------------+------------+
@@ -774,7 +783,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
     /// +------------+------------+
     /// | date       | temp_last |
     /// | ---        | ---        |
-    /// | date32     | i32        |
+    /// | Date     | i32        |
     /// +============+============+
     /// | 2020-08-23 | 9          |
     /// +------------+------------+
@@ -810,7 +819,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
     /// +------------+---------------+
     /// | date       | temp_n_unique |
     /// | ---        | ---           |
-    /// | date32     | u32           |
+    /// | Date     | u32           |
     /// +============+===============+
     /// | 2020-08-23 | 1             |
     /// +------------+---------------+
@@ -927,7 +936,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
     /// +------------+------------+
     /// | date       | temp_count |
     /// | ---        | ---        |
-    /// | date32     | u32        |
+    /// | Date     | u32        |
     /// +============+============+
     /// | 2020-08-23 | 1          |
     /// +------------+------------+
@@ -967,7 +976,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
     /// +--------------+------------+
     /// | date         | groups     |
     /// | ---          | ---        |
-    /// | date32(days) | list [u32] |
+    /// | Date(days) | list [u32] |
     /// +==============+============+
     /// | 2020-08-23   | "[3]"      |
     /// +--------------+------------+
@@ -1020,7 +1029,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
     ///  +--------------+---------------+----------+----------+
     ///  | date         | temp_n_unique | temp_sum | temp_min |
     ///  | ---          | ---           | ---      | ---      |
-    ///  | date32(days) | u32           | i32      | i32      |
+    ///  | Date(days) | u32           | i32      | i32      |
     ///  +==============+===============+==========+==========+
     ///  | 2020-08-23   | 1             | 9        | 9        |
     ///  +--------------+---------------+----------+----------+
@@ -1116,7 +1125,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
     /// +------------+------------------------+
     /// | date       | temp_agg_list          |
     /// | ---        | ---                    |
-    /// | date32     | list [i32]             |
+    /// | Date     | list [i32]             |
     /// +============+========================+
     /// | 2020-08-23 | "[Some(9)]"            |
     /// +------------+------------------------+
@@ -1137,25 +1146,28 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
         DataFrame::new(cols)
     }
 
-    /// Apply a closure over the groups as a new DataFrame.
-    pub fn apply<F>(&self, f: F) -> Result<DataFrame>
-    where
-        F: Fn(DataFrame) -> Result<DataFrame> + Send + Sync,
-    {
-        let df = if let Some(agg) = &self.selected_agg {
+    fn prepare_apply(&self) -> Result<DataFrame> {
+        if let Some(agg) = &self.selected_agg {
             if agg.is_empty() {
-                self.df.clone()
+                Ok(self.df.clone())
             } else {
                 let mut new_cols = Vec::with_capacity(self.selected_keys.len() + agg.len());
                 new_cols.extend_from_slice(&self.selected_keys);
                 let cols = self.df.select_series(agg)?;
                 new_cols.extend(cols.into_iter());
-                DataFrame::new_no_checks(new_cols)
+                Ok(DataFrame::new_no_checks(new_cols))
             }
         } else {
-            self.df.clone()
-        };
+            Ok(self.df.clone())
+        }
+    }
 
+    /// Apply a closure over the groups as a new DataFrame in parallel.
+    pub fn par_apply<F>(&self, f: F) -> Result<DataFrame>
+    where
+        F: Fn(DataFrame) -> Result<DataFrame> + Send + Sync,
+    {
+        let df = self.prepare_apply()?;
         let dfs = self
             .get_groups()
             .par_iter()
@@ -1169,6 +1181,26 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
         df.as_single_chunk();
         Ok(df)
     }
+
+    /// Apply a closure over the groups as a new DataFrame.
+    pub fn apply<F>(&self, f: F) -> Result<DataFrame>
+    where
+        F: Fn(DataFrame) -> Result<DataFrame> + Send + Sync,
+    {
+        let df = self.prepare_apply()?;
+        let dfs = self
+            .get_groups()
+            .iter()
+            .map(|t| {
+                let sub_df = unsafe { df.take_iter_unchecked(t.1.iter().map(|i| *i as usize)) };
+                f(sub_df)
+            })
+            .collect::<Result<Vec<_>>>()?;
+
+        let mut df = accumulate_dataframes_vertical(dfs)?;
+        df.as_single_chunk();
+        Ok(df)
+    }
 }
 
 #[derive(Copy, Clone)]
@@ -1220,10 +1252,10 @@ mod test {
     use num::traits::FloatConst;
 
     #[test]
-    #[cfg(feature = "dtype-date32")]
+    #[cfg(feature = "dtype-date")]
     #[cfg_attr(miri, ignore)]
     fn test_group_by() {
-        let s0 = Date32Chunked::parse_from_str_slice(
+        let s0 = DateChunked::parse_from_str_slice(
             "date",
             &[
                 "2020-08-21",
@@ -1314,7 +1346,7 @@ mod test {
         let gb = df.groupby("date").unwrap().n_unique().unwrap();
         println!("{:?}", df.groupby("date").unwrap().n_unique().unwrap());
         // check the group by column is filtered out.
-        assert_eq!(gb.width(), 2);
+        assert_eq!(gb.width(), 3);
         println!(
             "{:?}",
             df.groupby("date")
@@ -1434,6 +1466,7 @@ mod test {
 
     #[test]
     #[cfg_attr(miri, ignore)]
+    #[cfg(feature = "dtype-categorical")]
     fn test_groupby_categorical() {
         let mut df = df! {"foo" => ["a", "a", "b", "b", "c"],
                     "ham" => ["a", "a", "b", "b", "c"],
@@ -1441,7 +1474,7 @@ mod test {
         }
         .unwrap();
 
-        df.apply("foo", |s| s.cast::<CategoricalType>().unwrap())
+        df.apply("foo", |s| s.cast(&DataType::Categorical).unwrap())
             .unwrap();
 
         // check multiple keys and categorical
@@ -1479,13 +1512,13 @@ mod test {
             vec![1, 2, 3, 4, 4, 4, 2, 1, 1],
             vec![1, 2, 3, 4, 4, 4],
         ] {
-            let ca = UInt32Chunked::new_from_slice("", &slice);
-            let splitted = split_ca(&ca, 4).unwrap();
+            let ca = UInt32Chunked::new_from_slice("", slice);
+            let split = split_ca(&ca, 4).unwrap();
 
             let a = groupby(ca.into_iter()).into_iter().sorted().collect_vec();
 
-            let keys = splitted.iter().map(|ca| ca.cont_slice().unwrap()).collect();
-            let b = groupby_threaded_num(keys, 0, splitted.len() as u64)
+            let keys = split.iter().map(|ca| ca.cont_slice().unwrap()).collect();
+            let b = groupby_threaded_num(keys, 0, split.len() as u64)
                 .into_iter()
                 .sorted()
                 .collect_vec();
@@ -1539,6 +1572,7 @@ mod test {
 
     #[test]
     #[cfg_attr(miri, ignore)]
+    #[cfg(feature = "dtype-categorical")]
     fn test_groupby_null_group() -> Result<()> {
         // check if null is own group
         let mut df = df![
@@ -1547,7 +1581,7 @@ mod test {
             "int" => [1, 2, 3, 1, 1]
         ]?;
 
-        df.may_apply("g", |s| s.cast::<CategoricalType>())?;
+        df.may_apply("g", |s| s.cast(&DataType::Categorical))?;
 
         let out = df.groupby("g")?.sum()?;
         dbg!(out);
