@@ -1,4 +1,4 @@
-mod multiple_keys;
+pub(crate) mod multiple_keys;
 use polars_arrow::utils::CustomIterTools;
 
 use crate::frame::hash_join::multiple_keys::{
@@ -26,9 +26,11 @@ use unsafe_unwrap::UnsafeUnwrap;
 #[cfg(feature = "private")]
 pub use self::multiple_keys::private_left_join_multiple_keys;
 use crate::frame::groupby::hashing::HASHMAP_INIT_SIZE;
+use crate::utils::series::to_physical;
 
 /// If Categorical types are created without a global string cache or under
 /// a different global string cache the mapping will be incorrect.
+#[cfg(feature = "dtype-categorical")]
 pub(crate) fn check_categorical_src(l: &Series, r: &Series) -> Result<()> {
     if let (Ok(l), Ok(r)) = (l.categorical(), r.categorical()) {
         let l = l.categorical_map.as_ref().unwrap();
@@ -69,7 +71,7 @@ pub enum JoinType {
     Cross,
 }
 
-unsafe fn get_hash_tbl_threaded_join_partitioned<T, H>(
+pub(crate) unsafe fn get_hash_tbl_threaded_join_partitioned<T, H>(
     h: u64,
     hash_tables: &[HashMap<T, Vec<u32>, H>],
     len: u64,
@@ -85,11 +87,12 @@ unsafe fn get_hash_tbl_threaded_join_partitioned<T, H>(
     hash_tables.get_unchecked(idx)
 }
 
+#[allow(clippy::type_complexity)]
 unsafe fn get_hash_tbl_threaded_join_mut_partitioned<T, H>(
     h: u64,
-    hash_tables: &mut [HashMap<T, Vec<u32>, H>],
+    hash_tables: &mut [HashMap<T, (bool, Vec<u32>), H>],
     len: u64,
-) -> &mut HashMap<T, Vec<u32>, H> {
+) -> &mut HashMap<T, (bool, Vec<u32>), H> {
     let mut idx = 0;
     for i in 0..len {
         // can only be done for powers of two.
@@ -314,7 +317,7 @@ where
 /// Probe the build table and add tuples to the results (inner join)
 fn probe_outer<T, F, G, H>(
     probe_hashes: &[Vec<(u64, T)>],
-    hash_tbls: &mut [PlHashMap<T, Vec<u32>>],
+    hash_tbls: &mut [PlHashMap<T, (bool, Vec<u32>)>],
     results: &mut Vec<(Option<u32>, Option<u32>)>,
     n_tables: u64,
     // Function that get index_a, index_b when there is a match and pushes to result
@@ -348,8 +351,9 @@ fn probe_outer<T, F, G, H>(
 
             match entry {
                 // match and remove
-                RawEntryMut::Occupied(occupied) => {
-                    let indexes_b = occupied.remove();
+                RawEntryMut::Occupied(mut occupied) => {
+                    let (tracker, indexes_b) = occupied.get_mut();
+                    *tracker = true;
                     results.extend(indexes_b.iter().map(|&idx_b| swap_fn_match(idx_a, idx_b)))
                 }
                 // no match
@@ -360,9 +364,11 @@ fn probe_outer<T, F, G, H>(
     }
 
     for hash_tbl in hash_tbls {
-        hash_tbl.iter().for_each(|(_k, indexes_b)| {
+        hash_tbl.iter().for_each(|(_k, (tracker, indexes_b))| {
             // remaining joined values from the right table
-            results.extend(indexes_b.iter().map(|&idx_b| swap_fn_drain(idx_b)))
+            if !*tracker {
+                results.extend(indexes_b.iter().map(|&idx_b| swap_fn_drain(idx_b)))
+            }
         });
     }
 }
@@ -476,16 +482,15 @@ impl HashJoin<Float64Type> for Float64Chunked {
     }
 }
 
-impl HashJoin<ListType> for ListChunked {}
 impl HashJoin<CategoricalType> for CategoricalChunked {
     fn hash_join_inner(&self, other: &CategoricalChunked) -> Vec<(u32, u32)> {
-        self.deref().hash_join_inner(&other.cast().unwrap())
+        self.deref().hash_join_inner(other.deref())
     }
     fn hash_join_left(&self, other: &CategoricalChunked) -> Vec<(u32, Option<u32>)> {
-        self.deref().hash_join_left(&other.cast().unwrap())
+        self.deref().hash_join_left(other.deref())
     }
     fn hash_join_outer(&self, other: &CategoricalChunked) -> Vec<(Option<u32>, Option<u32>)> {
-        self.deref().hash_join_outer(&other.cast().unwrap())
+        self.deref().hash_join_outer(other.deref())
     }
 }
 
@@ -500,12 +505,12 @@ where
     let splitted_a = split_ca(a, n_threads).unwrap();
     let splitted_b = split_ca(b, n_threads).unwrap();
     match (
-        left.null_count(),
-        right.null_count(),
+        left.has_validity(),
+        right.has_validity(),
         left.chunks.len(),
         right.chunks.len(),
     ) {
-        (0, 0, 1, 1) => {
+        (false, false, 1, 1) => {
             let keys_a = splitted_a
                 .iter()
                 .map(|ca| ca.cont_slice().unwrap())
@@ -516,7 +521,7 @@ where
                 .collect::<Vec<_>>();
             hash_join_tuples_inner(keys_a, keys_b, swap)
         }
-        (0, 0, _, _) => {
+        (false, false, _, _) => {
             let keys_a = splitted_a
                 .iter()
                 .map(|ca| ca.into_no_null_iter().collect::<Vec<_>>())
@@ -576,12 +581,12 @@ where
     let splitted_a = split_ca(left, n_threads).unwrap();
     let splitted_b = split_ca(right, n_threads).unwrap();
     match (
-        left.null_count(),
-        right.null_count(),
+        left.has_validity(),
+        right.has_validity(),
         left.chunks.len(),
         right.chunks.len(),
     ) {
-        (0, 0, 1, 1) => {
+        (false, false, 1, 1) => {
             let keys_a = splitted_a
                 .iter()
                 .map(|ca| ca.cont_slice().unwrap())
@@ -592,7 +597,7 @@ where
                 .collect::<Vec<_>>();
             hash_join_tuples_left(keys_a, keys_b)
         }
-        (0, 0, _, _) => {
+        (false, false, _, _) => {
             let keys_a = splitted_a
                 .iter()
                 .map(|ca| ca.into_no_null_iter().collect_trusted::<Vec<_>>())
@@ -687,9 +692,11 @@ where
                 num_group_join_inner(&ca, &other)
             }
             _ => {
-                let ca = self.cast::<UInt32Type>().unwrap();
-                let other = other.cast::<UInt32Type>().unwrap();
-                num_group_join_inner(&ca, &other)
+                let ca = self.cast(&DataType::UInt32).unwrap();
+                let ca = ca.u32().unwrap();
+                let other = other.cast(&DataType::UInt32).unwrap();
+                let other = other.u32().unwrap();
+                num_group_join_inner(ca, other)
             }
         }
     }
@@ -727,9 +734,11 @@ where
                 num_group_join_left(&ca, &other)
             }
             _ => {
-                let ca = self.cast::<UInt32Type>().unwrap();
-                let other = other.cast::<UInt32Type>().unwrap();
-                num_group_join_left(&ca, &other)
+                let ca = self.cast(&DataType::UInt32).unwrap();
+                let ca = ca.u32().unwrap();
+                let other = other.cast(&DataType::UInt32).unwrap();
+                let other = other.u32().unwrap();
+                num_group_join_left(ca, other)
             }
         }
     }
@@ -741,8 +750,8 @@ where
         let splitted_a = split_ca(a, n_partitions).unwrap();
         let splitted_b = split_ca(b, n_partitions).unwrap();
 
-        match (a.null_count(), b.null_count()) {
-            (0, 0) => {
+        match (a.has_validity(), b.has_validity()) {
+            (false, false) => {
                 let iters_a = splitted_a
                     .iter()
                     .map(|ca| ca.into_no_null_iter())
@@ -764,15 +773,19 @@ where
 
 impl HashJoin<BooleanType> for BooleanChunked {
     fn hash_join_inner(&self, other: &BooleanChunked) -> Vec<(u32, u32)> {
-        let ca = self.cast::<UInt32Type>().unwrap();
-        let other = other.cast::<UInt32Type>().unwrap();
-        ca.hash_join_inner(&other)
+        let ca = self.cast(&DataType::UInt32).unwrap();
+        let ca = ca.u32().unwrap();
+        let other = other.cast(&DataType::UInt32).unwrap();
+        let other = other.u32().unwrap();
+        ca.hash_join_inner(other)
     }
 
     fn hash_join_left(&self, other: &BooleanChunked) -> Vec<(u32, Option<u32>)> {
-        let ca = self.cast::<UInt32Type>().unwrap();
-        let other = other.cast::<UInt32Type>().unwrap();
-        ca.hash_join_left(&other)
+        let ca = self.cast(&DataType::UInt32).unwrap();
+        let ca = ca.u32().unwrap();
+        let other = other.cast(&DataType::UInt32).unwrap();
+        let other = other.u32().unwrap();
+        ca.hash_join_left(other)
     }
 
     fn hash_join_outer(&self, other: &BooleanChunked) -> Vec<(Option<u32>, Option<u32>)> {
@@ -782,8 +795,8 @@ impl HashJoin<BooleanType> for BooleanChunked {
         let splitted_a = split_ca(a, n_partitions).unwrap();
         let splitted_b = split_ca(b, n_partitions).unwrap();
 
-        match (a.null_count(), b.null_count()) {
-            (0, 0) => {
+        match (a.has_validity(), b.has_validity()) {
+            (false, false) => {
                 let iters_a = splitted_a
                     .iter()
                     .map(|ca| ca.into_no_null_iter())
@@ -855,8 +868,8 @@ impl HashJoin<Utf8Type> for Utf8Chunked {
         let splitted_a = split_ca(a, n_partitions).unwrap();
         let splitted_b = split_ca(b, n_partitions).unwrap();
 
-        match (a.null_count(), b.null_count()) {
-            (0, 0) => {
+        match (a.has_validity(), b.has_validity()) {
+            (false, false) => {
                 let iters_a = splitted_a
                     .iter()
                     .map(|ca| ca.into_no_null_iter())
@@ -919,18 +932,11 @@ where
                     }
                 }
             })
-            .collect::<ChunkedArray<T>>()
+            .collect_trusted::<ChunkedArray<T>>()
             .into_series()
     }
 }
 
-impl ZipOuterJoinColumn for Float32Chunked {}
-impl ZipOuterJoinColumn for Float64Chunked {}
-impl ZipOuterJoinColumn for ListChunked {}
-impl ZipOuterJoinColumn for CategoricalChunked {}
-#[cfg(feature = "object")]
-impl<T> ZipOuterJoinColumn for ObjectChunked<T> {}
-
 macro_rules! impl_zip_outer_join {
     ($chunkedtype:ident) => {
         impl ZipOuterJoinColumn for $chunkedtype {
@@ -965,12 +971,43 @@ macro_rules! impl_zip_outer_join {
 impl_zip_outer_join!(BooleanChunked);
 impl_zip_outer_join!(Utf8Chunked);
 
+impl ZipOuterJoinColumn for Float32Chunked {
+    fn zip_outer_join_column(
+        &self,
+        right_column: &Series,
+        opt_join_tuples: &[(Option<u32>, Option<u32>)],
+    ) -> Series {
+        self.apply_as_ints(|s| {
+            s.zip_outer_join_column(
+                &right_column.bit_repr_small().into_series(),
+                opt_join_tuples,
+            )
+        })
+    }
+}
+
+impl ZipOuterJoinColumn for Float64Chunked {
+    fn zip_outer_join_column(
+        &self,
+        right_column: &Series,
+        opt_join_tuples: &[(Option<u32>, Option<u32>)],
+    ) -> Series {
+        self.apply_as_ints(|s| {
+            s.zip_outer_join_column(
+                &right_column.bit_repr_large().into_series(),
+                opt_join_tuples,
+            )
+        })
+    }
+}
+
 impl DataFrame {
     /// Utility method to finish a join.
     pub(crate) fn finish_join(
         &self,
         mut df_left: DataFrame,
         mut df_right: DataFrame,
+        suffix: Option<String>,
     ) -> Result<DataFrame> {
         let mut left_names = HashSet::with_capacity_and_hasher(df_left.width(), RandomState::new());
 
@@ -985,9 +1022,10 @@ impl DataFrame {
                 rename_strs.push(series.name().to_owned())
             }
         });
+        let suffix = suffix.as_deref().unwrap_or("_right");
 
         for name in rename_strs {
-            df_right.rename(&name, &format!("{}_right", name))?;
+            df_right.rename(&name, &format!("{}{}", name, suffix))?;
         }
 
         df_left.hstack_mut(&df_right.columns)?;
@@ -1005,12 +1043,50 @@ impl DataFrame {
     }
 
     /// Generic join method. Can be used to join on multiple columns.
+    ///
+    /// # Example
+    ///
+    /// ```rust
+    /// use polars_core::df;
+    /// use polars_core::prelude::*;
+    ///
+    /// fn example() -> Result<()> {
+    ///     let df1: DataFrame = df!("Fruit" => &["Apple", "Banana", "Pear"],
+    ///                              "Phosphorus (mg/100g)" => &[11, 22, 12])?;
+    ///     let df2: DataFrame = df!("Name" => &["Apple", "Banana", "Pear"],
+    ///                              "Potassium (mg/100g)" => &[107, 358, 115])?;
+    ///
+    ///     let df3: DataFrame = df1.join(&df2, "Fruit", "Name", JoinType::Inner, None)?;
+    ///     assert_eq!(df3.shape(), (3, 3));
+    ///     println!("{}", df3);
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
+    ///
+    /// Output:
+    ///
+    /// ```text
+    /// shape: (3, 3)
+    /// +--------+----------------------+---------------------+
+    /// | Fruit  | Phosphorus (mg/100g) | Potassium (mg/100g) |
+    /// | ---    | ---                  | ---                 |
+    /// | str    | i32                  | i32                 |
+    /// +========+======================+=====================+
+    /// | Apple  | 11                   | 107                 |
+    /// +--------+----------------------+---------------------+
+    /// | Banana | 22                   | 358                 |
+    /// +--------+----------------------+---------------------+
+    /// | Pear   | 12                   | 115                 |
+    /// +--------+----------------------+---------------------+
+    /// ```
     pub fn join<'a, J, S1: Selection<'a, J>, S2: Selection<'a, J>>(
         &self,
         other: &DataFrame,
         left_on: S1,
         right_on: S2,
         how: JoinType,
+        suffix: Option<String>,
     ) -> Result<DataFrame> {
         #[cfg(feature = "cross_join")]
         if let JoinType::Cross = how {
@@ -1034,22 +1110,19 @@ impl DataFrame {
             return Err(PolarsError::ValueError("the dtype of the join keys don't match. first cast your columns to the correct dtype".into()));
         }
 
+        #[cfg(feature = "dtype-categorical")]
         for (l, r) in selected_left.iter().zip(&selected_right) {
             check_categorical_src(l, r)?
         }
 
         // Single keys
         if selected_left.len() == 1 {
+            let s_left = self.column(selected_left[0].name())?;
+            let s_right = other.column(selected_right[0].name())?;
             return match how {
-                JoinType::Inner => {
-                    self.inner_join(other, selected_left[0].name(), selected_right[0].name())
-                }
-                JoinType::Left => {
-                    self.left_join(other, selected_left[0].name(), selected_right[0].name())
-                }
-                JoinType::Outer => {
-                    self.outer_join(other, selected_left[0].name(), selected_right[0].name())
-                }
+                JoinType::Inner => self.inner_join_from_series(other, s_left, s_right, suffix),
+                JoinType::Left => self.left_join_from_series(other, s_left, s_right, suffix),
+                JoinType::Outer => self.outer_join_from_series(other, s_left, s_right, suffix),
                 #[cfg(feature = "asof_join")]
                 JoinType::AsOf => {
                     self.join_asof(other, selected_left[0].name(), selected_right[0].name())
@@ -1078,31 +1151,16 @@ impl DataFrame {
                 self.height()
             }
         }
-
-        // check if we have floats, in that case we have to coerce them
-        if selected_left
-            .iter()
-            .any(|s| matches!(s.dtype(), DataType::Float32 | DataType::Float64))
-        {
-            #[cfg(feature = "dtype-u64")]
-            for selected in &mut [&mut selected_left, &mut selected_right] {
-                selected.iter_mut().for_each(|s| match s.dtype().clone() {
-                    DataType::Float64 => *s = s.bit_repr_large().into_series(),
-                    DataType::Float32 => *s = s.bit_repr_small().into_series(),
-                    _ => {}
-                });
-            }
-            #[cfg(not(feature = "dtype-u64"))]
-            {
-                panic!("activate dtype-u64 feature to be able to join on floats")
-            }
-        }
+        // make sure that we don't have logical types.
+        // we don't overwrite the original selected as that might be used to create a column in the new df
+        let selected_left_physical = to_physical(&selected_left);
+        let selected_right_physical = to_physical(&selected_right);
 
         // multiple keys
         match how {
             JoinType::Inner => {
-                let left = DataFrame::new_no_checks(selected_left);
-                let right = DataFrame::new_no_checks(selected_right.clone());
+                let left = DataFrame::new_no_checks(selected_left_physical);
+                let right = DataFrame::new_no_checks(selected_right_physical);
                 let (left, right, swap) = det_hash_prone_order!(left, right);
                 let join_tuples = inner_join_multiple_keys(&left, &right, swap);
 
@@ -1115,11 +1173,11 @@ impl DataFrame {
                         )
                     },
                 );
-                self.finish_join(df_left, df_right)
+                self.finish_join(df_left, df_right, suffix)
             }
             JoinType::Left => {
-                let left = DataFrame::new_no_checks(selected_left);
-                let right = DataFrame::new_no_checks(selected_right.clone());
+                let left = DataFrame::new_no_checks(selected_left_physical);
+                let right = DataFrame::new_no_checks(selected_right_physical);
                 let join_tuples = left_join_multiple_keys(&left, &right);
 
                 let (df_left, df_right) = POOL.join(
@@ -1133,11 +1191,11 @@ impl DataFrame {
                         )
                     },
                 );
-                self.finish_join(df_left, df_right)
+                self.finish_join(df_left, df_right, suffix)
             }
             JoinType::Outer => {
-                let left = DataFrame::new_no_checks(selected_left.clone());
-                let right = DataFrame::new_no_checks(selected_right.clone());
+                let left = DataFrame::new_no_checks(selected_left_physical);
+                let right = DataFrame::new_no_checks(selected_right_physical);
 
                 let (left, right, swap) = det_hash_prone_order!(left, right);
                 let opt_join_tuples = outer_join_multiple_keys(&left, &right, swap);
@@ -1164,7 +1222,7 @@ impl DataFrame {
                     s.rename(s_left.name());
                     df_left.hstack_mut(&[s])?;
                 }
-                self.finish_join(df_left, df_right)
+                self.finish_join(df_left, df_right, suffix)
             }
             #[cfg(feature = "asof_join")]
             JoinType::AsOf => Err(PolarsError::ValueError(
@@ -1187,15 +1245,13 @@ impl DataFrame {
     ///     left.inner_join(right, "join_column_left", "join_column_right")
     /// }
     /// ```
-    pub fn inner_join(
+    pub fn inner_join<'a, J, S1: Selection<'a, J>, S2: Selection<'a, J>>(
         &self,
         other: &DataFrame,
-        left_on: &str,
-        right_on: &str,
+        left_on: S1,
+        right_on: S2,
     ) -> Result<DataFrame> {
-        let s_left = self.column(left_on)?;
-        let s_right = other.column(right_on)?;
-        self.inner_join_from_series(other, s_left, s_right)
+        self.join(other, left_on, right_on, JoinType::Inner, None)
     }
 
     pub(crate) fn inner_join_from_series(
@@ -1203,7 +1259,9 @@ impl DataFrame {
         other: &DataFrame,
         s_left: &Series,
         s_right: &Series,
+        suffix: Option<String>,
     ) -> Result<DataFrame> {
+        #[cfg(feature = "dtype-categorical")]
         check_categorical_src(s_left, s_right)?;
         let join_tuples = s_left.hash_join_inner(s_right);
 
@@ -1216,7 +1274,7 @@ impl DataFrame {
                     .take_iter_unchecked(join_tuples.iter().map(|(_left, right)| *right as usize))
             },
         );
-        self.finish_join(df_left, df_right)
+        self.finish_join(df_left, df_right, suffix)
     }
 
     /// Perform a left join on two DataFrames
@@ -1228,10 +1286,13 @@ impl DataFrame {
     ///     left.left_join(right, "join_column_left", "join_column_right")
     /// }
     /// ```
-    pub fn left_join(&self, other: &DataFrame, left_on: &str, right_on: &str) -> Result<DataFrame> {
-        let s_left = self.column(left_on)?;
-        let s_right = other.column(right_on)?;
-        self.left_join_from_series(other, s_left, s_right)
+    pub fn left_join<'a, J, S1: Selection<'a, J>, S2: Selection<'a, J>>(
+        &self,
+        other: &DataFrame,
+        left_on: S1,
+        right_on: S2,
+    ) -> Result<DataFrame> {
+        self.join(other, left_on, right_on, JoinType::Left, None)
     }
 
     pub(crate) fn left_join_from_series(
@@ -1239,7 +1300,9 @@ impl DataFrame {
         other: &DataFrame,
         s_left: &Series,
         s_right: &Series,
+        suffix: Option<String>,
     ) -> Result<DataFrame> {
+        #[cfg(feature = "dtype-categorical")]
         check_categorical_src(s_left, s_right)?;
         let opt_join_tuples = s_left.hash_join_left(s_right);
 
@@ -1253,7 +1316,7 @@ impl DataFrame {
                 )
             },
         );
-        self.finish_join(df_left, df_right)
+        self.finish_join(df_left, df_right, suffix)
     }
 
     /// Perform an outer join on two DataFrames
@@ -1265,22 +1328,22 @@ impl DataFrame {
     ///     left.outer_join(right, "join_column_left", "join_column_right")
     /// }
     /// ```
-    pub fn outer_join(
+    pub fn outer_join<'a, J, S1: Selection<'a, J>, S2: Selection<'a, J>>(
         &self,
         other: &DataFrame,
-        left_on: &str,
-        right_on: &str,
+        left_on: S1,
+        right_on: S2,
     ) -> Result<DataFrame> {
-        let s_left = self.column(left_on)?;
-        let s_right = other.column(right_on)?;
-        self.outer_join_from_series(other, s_left, s_right)
+        self.join(other, left_on, right_on, JoinType::Outer, None)
     }
     pub(crate) fn outer_join_from_series(
         &self,
         other: &DataFrame,
         s_left: &Series,
         s_right: &Series,
+        suffix: Option<String>,
     ) -> Result<DataFrame> {
+        #[cfg(feature = "dtype-categorical")]
         check_categorical_src(s_left, s_right)?;
         // Get the indexes of the joined relations
         let opt_join_tuples = s_left.hash_join_outer(s_right);
@@ -1305,7 +1368,7 @@ impl DataFrame {
         let mut s = s_left.zip_outer_join_column(s_right, &opt_join_tuples);
         s.rename(s_left.name());
         df_left.hstack_mut(&[s])?;
-        self.finish_join(df_left, df_right)
+        self.finish_join(df_left, df_right, suffix)
     }
 }
 
@@ -1313,7 +1376,6 @@ impl DataFrame {
 mod test {
     use crate::df;
     use crate::prelude::*;
-    use crate::toggle_string_cache;
 
     fn create_frames() -> (DataFrame, DataFrame) {
         let s0 = Series::new("days", &[0, 1, 2]);
@@ -1464,31 +1526,29 @@ mod test {
     #[test]
     #[cfg_attr(miri, ignore)]
     fn test_join_multiple_columns() {
-        let (df_a, df_b) = get_dfs();
+        let (mut df_a, mut df_b) = get_dfs();
 
         // First do a hack with concatenated string dummy column
         let mut s = df_a
             .column("a")
             .unwrap()
-            .cast::<Utf8Type>()
+            .cast(&DataType::Utf8)
             .unwrap()
             .utf8()
             .unwrap()
             + df_a.column("b").unwrap().utf8().unwrap();
         s.rename("dummy");
 
-        let mut df_a = df_a.clone();
         df_a.with_column(s).unwrap();
         let mut s = df_b
             .column("foo")
             .unwrap()
-            .cast::<Utf8Type>()
+            .cast(&DataType::Utf8)
             .unwrap()
             .utf8()
             .unwrap()
             + df_b.column("bar").unwrap().utf8().unwrap();
         s.rename("dummy");
-        let mut df_b = df_b.clone();
         df_b.with_column(s).unwrap();
 
         let joined = df_a.left_join(&df_b, "dummy", "dummy").unwrap();
@@ -1508,14 +1568,14 @@ mod test {
 
         // now check the join with multiple columns
         let joined = df_a
-            .join(&df_b, &["a", "b"], &["foo", "bar"], JoinType::Left)
+            .join(&df_b, &["a", "b"], &["foo", "bar"], JoinType::Left, None)
             .unwrap();
         let ca = joined.column("ham").unwrap().utf8().unwrap();
         dbg!(&df_a, &df_b);
         assert_eq!(Vec::from(ca), correct_ham);
         let joined_inner_hack = df_a.inner_join(&df_b, "dummy", "dummy").unwrap();
         let joined_inner = df_a
-            .join(&df_b, &["a", "b"], &["foo", "bar"], JoinType::Inner)
+            .join(&df_b, &["a", "b"], &["foo", "bar"], JoinType::Inner, None)
             .unwrap();
 
         dbg!(&joined_inner_hack, &joined_inner);
@@ -1526,7 +1586,7 @@ mod test {
 
         let joined_outer_hack = df_a.outer_join(&df_b, "dummy", "dummy").unwrap();
         let joined_outer = df_a
-            .join(&df_b, &["a", "b"], &["foo", "bar"], JoinType::Outer)
+            .join(&df_b, &["a", "b"], &["foo", "bar"], JoinType::Outer, None)
             .unwrap();
         assert!(joined_outer_hack
             .column("ham")
@@ -1536,18 +1596,20 @@ mod test {
 
     #[test]
     #[cfg_attr(miri, ignore)]
+    #[cfg(feature = "dtype-categorical")]
     fn test_join_categorical() {
+        use crate::toggle_string_cache;
         let _lock = crate::SINGLE_LOCK.lock();
         toggle_string_cache(true);
 
         let (mut df_a, mut df_b) = get_dfs();
 
-        df_a.may_apply("b", |s| s.cast_with_dtype(&DataType::Categorical))
+        df_a.may_apply("b", |s| s.cast(&DataType::Categorical))
             .unwrap();
-        df_b.may_apply("bar", |s| s.cast_with_dtype(&DataType::Categorical))
+        df_b.may_apply("bar", |s| s.cast(&DataType::Categorical))
             .unwrap();
 
-        let out = df_a.join(&df_b, "b", "bar", JoinType::Left).unwrap();
+        let out = df_a.join(&df_b, "b", "bar", JoinType::Left, None).unwrap();
         assert_eq!(out.shape(), (6, 5));
         let correct_ham = &[
             Some("let"),
@@ -1564,41 +1626,68 @@ mod test {
 
         // Test an error when joining on different string cache
         let (mut df_a, mut df_b) = get_dfs();
-        df_a.may_apply("b", |s| s.cast_with_dtype(&DataType::Categorical))
+        df_a.may_apply("b", |s| s.cast(&DataType::Categorical))
             .unwrap();
         // create a new cache
         toggle_string_cache(false);
         toggle_string_cache(true);
 
-        df_b.may_apply("bar", |s| s.cast_with_dtype(&DataType::Categorical))
+        df_b.may_apply("bar", |s| s.cast(&DataType::Categorical))
             .unwrap();
-        let out = df_a.join(&df_b, "b", "bar", JoinType::Left);
+        let out = df_a.join(&df_b, "b", "bar", JoinType::Left, None);
         assert!(out.is_err())
     }
 
     #[test]
     #[cfg_attr(miri, ignore)]
-    fn empty_df_join() {
+    fn empty_df_join() -> Result<()> {
         let empty: Vec<String> = vec![];
-        let left = DataFrame::new(vec![
+        let empty_df = DataFrame::new(vec![
             Series::new("key", &empty),
-            Series::new("lval", &empty),
+            Series::new("eval", &empty),
         ])
         .unwrap();
 
-        let right = DataFrame::new(vec![
+        let df = DataFrame::new(vec![
             Series::new("key", &["foo"]),
-            Series::new("rval", &[4]),
+            Series::new("aval", &[4]),
         ])
         .unwrap();
 
-        let res = left.inner_join(&right, "key", "key");
+        let out = empty_df.inner_join(&df, "key", "key").unwrap();
+        assert_eq!(out.height(), 0);
+        let out = empty_df.left_join(&df, "key", "key").unwrap();
+        assert_eq!(out.height(), 0);
+        let out = empty_df.outer_join(&df, "key", "key").unwrap();
+        assert_eq!(out.height(), 1);
+        df.left_join(&empty_df, "key", "key")?;
+        df.inner_join(&empty_df, "key", "key")?;
+        df.outer_join(&empty_df, "key", "key")?;
 
-        assert!(res.is_ok());
-        assert_eq!(res.unwrap().height(), 0);
-        right.left_join(&left, "key", "key").unwrap();
-        right.inner_join(&left, "key", "key").unwrap();
-        right.outer_join(&left, "key", "key").unwrap();
+        let empty: Vec<String> = vec![];
+        let _empty_df = DataFrame::new(vec![
+            Series::new("key", &empty),
+            Series::new("eval", &empty),
+        ])
+        .unwrap();
+
+        let df = df![
+            "key" => [1i32, 2],
+            "vals" => [1, 2],
+        ]?;
+
+        // https://github.com/pola-rs/polars/issues/1824
+        let empty: Vec<i32> = vec![];
+        let empty_df = DataFrame::new(vec![
+            Series::new("key", &empty),
+            Series::new("1val", &empty),
+            Series::new("2val", &empty),
+        ])?;
+
+        let out = df.left_join(&empty_df, "key", "key")?;
+        assert_eq!(out.shape(), (2, 4));
+
+        Ok(())
     }
 
     #[test]
@@ -1639,18 +1728,129 @@ mod test {
 
         // dtypes don't match, error
         assert!(df1
-            .join(&df2, vec!["a", "b"], vec!["a", "b"], JoinType::Left)
+            .join(&df2, vec!["a", "b"], vec!["a", "b"], JoinType::Left, None)
             .is_err());
         // length of join keys don't match error
         assert!(df1
-            .join(&df2, vec!["a"], vec!["a", "b"], JoinType::Left)
+            .join(&df2, vec!["a"], vec!["a", "b"], JoinType::Left, None)
             .is_err());
         Ok(())
     }
 
     #[test]
     #[cfg_attr(miri, ignore)]
-    #[cfg(feature = "dtype-u64")]
+    fn test_joins_with_duplicates() -> Result<()> {
+        // test joins with duplicates in both dataframes
+
+        let df_left = df![
+            "col1" => [1, 1, 2],
+            "int_col" => [1, 2, 3]
+        ]
+        .unwrap();
+
+        let df_right = df![
+            "join_col1" => [1, 1, 1, 1, 1, 3],
+            "dbl_col" => [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
+        ]
+        .unwrap();
+
+        let df_inner_join = df_left.inner_join(&df_right, "col1", "join_col1").unwrap();
+
+        assert_eq!(df_inner_join.height(), 10);
+        assert_eq!(df_inner_join.column("col1")?.null_count(), 0);
+        assert_eq!(df_inner_join.column("int_col")?.null_count(), 0);
+        assert_eq!(df_inner_join.column("dbl_col")?.null_count(), 0);
+
+        let df_left_join = df_left.left_join(&df_right, "col1", "join_col1").unwrap();
+
+        assert_eq!(df_left_join.height(), 11);
+        assert_eq!(df_left_join.column("col1")?.null_count(), 0);
+        assert_eq!(df_left_join.column("int_col")?.null_count(), 0);
+        assert_eq!(df_left_join.column("dbl_col")?.null_count(), 1);
+
+        let df_outer_join = df_left.outer_join(&df_right, "col1", "join_col1").unwrap();
+
+        assert_eq!(df_outer_join.height(), 12);
+        assert_eq!(df_outer_join.column("col1")?.null_count(), 0);
+        assert_eq!(df_outer_join.column("int_col")?.null_count(), 1);
+        assert_eq!(df_outer_join.column("dbl_col")?.null_count(), 1);
+
+        Ok(())
+    }
+
+    #[test]
+    #[cfg_attr(miri, ignore)]
+    fn test_multi_joins_with_duplicates() -> Result<()> {
+        // test joins with multiple join columns and duplicates in both
+        // dataframes
+
+        let df_left = df![
+            "col1" => [1, 1, 1],
+            "join_col2" => ["a", "a", "b"],
+            "int_col" => [1, 2, 3]
+        ]
+        .unwrap();
+
+        let df_right = df![
+            "join_col1" => [1, 1, 1, 1, 1, 2],
+            "col2" => ["a", "a", "a", "a", "a", "c"],
+            "dbl_col" => [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
+        ]
+        .unwrap();
+
+        let df_inner_join = df_left
+            .join(
+                &df_right,
+                &["col1", "join_col2"],
+                &["join_col1", "col2"],
+                JoinType::Inner,
+                None,
+            )
+            .unwrap();
+
+        assert_eq!(df_inner_join.height(), 10);
+        assert_eq!(df_inner_join.column("col1")?.null_count(), 0);
+        assert_eq!(df_inner_join.column("join_col2")?.null_count(), 0);
+        assert_eq!(df_inner_join.column("int_col")?.null_count(), 0);
+        assert_eq!(df_inner_join.column("dbl_col")?.null_count(), 0);
+
+        let df_left_join = df_left
+            .join(
+                &df_right,
+                &["col1", "join_col2"],
+                &["join_col1", "col2"],
+                JoinType::Left,
+                None,
+            )
+            .unwrap();
+
+        assert_eq!(df_left_join.height(), 11);
+        assert_eq!(df_left_join.column("col1")?.null_count(), 0);
+        assert_eq!(df_left_join.column("join_col2")?.null_count(), 0);
+        assert_eq!(df_left_join.column("int_col")?.null_count(), 0);
+        assert_eq!(df_left_join.column("dbl_col")?.null_count(), 1);
+
+        let df_outer_join = df_left
+            .join(
+                &df_right,
+                &["col1", "join_col2"],
+                &["join_col1", "col2"],
+                JoinType::Outer,
+                None,
+            )
+            .unwrap();
+
+        assert_eq!(df_outer_join.height(), 12);
+        assert_eq!(df_outer_join.column("col1")?.null_count(), 0);
+        assert_eq!(df_outer_join.column("join_col2")?.null_count(), 0);
+        assert_eq!(df_outer_join.column("int_col")?.null_count(), 1);
+        assert_eq!(df_outer_join.column("dbl_col")?.null_count(), 1);
+
+        Ok(())
+    }
+
+    #[test]
+    #[cfg_attr(miri, ignore)]
     fn test_join_floats() -> Result<()> {
         let df_a = df! {
             "a" => &[1.0, 2.0, 1.0, 1.0],
@@ -1664,11 +1864,50 @@ mod test {
             "ham" => &["let", "var", "const"]
         }?;
 
-        let out = df_a.join(&df_b, vec!["a", "c"], vec!["foo", "bar"], JoinType::Left)?;
+        let out = df_a.join(
+            &df_b,
+            vec!["a", "c"],
+            vec!["foo", "bar"],
+            JoinType::Left,
+            None,
+        )?;
         assert_eq!(
             Vec::from(out.column("ham")?.utf8()?),
             &[None, Some("var"), None, None]
         );
+
+        let out = df_a.join(
+            &df_b,
+            vec!["a", "c"],
+            vec!["foo", "bar"],
+            JoinType::Outer,
+            None,
+        )?;
+        assert_eq!(
+            out.dtypes(),
+            &[
+                DataType::Utf8,
+                DataType::Float64,
+                DataType::Float64,
+                DataType::Utf8
+            ]
+        );
+        Ok(())
+    }
+
+    #[test]
+    #[cfg_attr(miri, ignore)]
+    fn test_join_nulls() -> Result<()> {
+        let a = df![
+            "a" => [Some(1), None, None]
+        ]?;
+        let b = df![
+            "a" => [Some(1), None, None, None, None]
+        ]?;
+
+        let out = a.inner_join(&b, "a", "a")?;
+
+        assert_eq!(out.shape(), (9, 1));
         Ok(())
     }
 }
