@@ -1,21 +1,7 @@
 //! Everything you need to get started with Polars.
-pub use std::sync::Arc;
-
-pub use arrow::datatypes::{ArrowPrimitiveType, Field as ArrowField, Schema as ArrowSchema};
-
-pub use polars_arrow::vec::AlignedVec;
-pub(crate) use polars_arrow::{array::*, trusted_len::TrustedLen};
-
-#[cfg(feature = "object")]
-pub use crate::chunked_array::object::PolarsObject;
-pub use crate::chunked_array::ops::take::take_random::{
-    IntoTakeRandom, NumTakeRandomChunked, NumTakeRandomCont,
-};
-#[cfg(feature = "temporal")]
-pub use crate::chunked_array::temporal::conversion::*;
+pub(crate) use crate::chunked_array::to_array;
 pub(crate) use crate::frame::groupby::aggregations::*;
-#[cfg(feature = "checked_arithmetic")]
-pub use crate::series::arithmetic::checked::NumOpsDispatchChecked;
+pub(crate) use crate::utils::CustomIterTools;
 pub use crate::{
     chunked_array::{
         arithmetic::Pow,
@@ -24,13 +10,13 @@ pub use crate::{
             ListPrimitiveChunkedBuilder, ListUtf8ChunkedBuilder, NewChunkedArray,
             PrimitiveChunkedBuilder, Utf8ChunkedBuilder,
         },
-        comparison::NumComp,
         iterator::{IntoNoNullIterator, PolarsIterator},
-        ops::{aggregate::*, chunkops::ChunkOps, window::InitFold, *},
+        ops::{aggregate::*, chunkops::ChunkOps, *},
         ChunkedArray,
     },
     datatypes,
     datatypes::*,
+    df,
     error::{PolarsError, Result},
     frame::{hash_join::JoinType, DataFrame},
     series::{
@@ -41,3 +27,21 @@ pub use crate::{
     utils::IntoVec,
     vector_hasher::VecHash,
 };
+pub(crate) use arrow::array::*;
+pub use arrow::datatypes::{Field as ArrowField, Schema as ArrowSchema};
+pub use polars_arrow::prelude::{AlignedVec, LargeListArray, LargeStringArray};
+pub(crate) use polars_arrow::trusted_len::TrustedLen;
+pub use std::sync::Arc;
+
+#[cfg(feature = "object")]
+pub use crate::chunked_array::object::PolarsObject;
+#[cfg(feature = "temporal")]
+pub use crate::chunked_array::temporal::conversion::*;
+#[cfg(feature = "checked_arithmetic")]
+pub use crate::series::arithmetic::checked::NumOpsDispatchChecked;
+
+#[cfg(feature = "rank")]
+pub use crate::chunked_array::ops::unique::rank::RankMethod;
+
+#[cfg(feature = "rolling_window")]
+pub use crate::chunked_array::ops::rolling_window::RollingOptions;
