@@ -1,13 +1,18 @@
+mod boolean;
+#[cfg(feature = "dtype-categorical")]
+mod categorical;
 #[cfg(any(
-    feature = "dtype-date64",
-    feature = "dtype-date32",
-    feature = "dtype-duration-ms",
-    feature = "dtype-duration-ns",
-    feature = "dtype-time64-ns"
+    feature = "dtype-datetime",
+    feature = "dtype-date",
+    feature = "dtype-time"
 ))]
-pub mod dates;
+mod dates_time;
+mod floats;
+mod list;
 #[cfg(feature = "object")]
-pub mod object;
+mod object;
+mod utf8;
+
 #[cfg(feature = "object")]
 use std::any::Any;
 
@@ -15,6 +20,8 @@ use super::private;
 use super::IntoSeries;
 use super::SeriesTrait;
 use crate::chunked_array::comparison::*;
+#[cfg(feature = "rolling_window")]
+use crate::chunked_array::ops::rolling_window::RollingOptions;
 use crate::chunked_array::{
     ops::{
         aggregate::{ChunkAggSeries, VarAggSeries},
@@ -24,8 +31,6 @@ use crate::chunked_array::{
     AsSinglePtr, ChunkIdIter,
 };
 use crate::fmt::FmtList;
-#[cfg(feature = "asof_join")]
-use crate::frame::asof_join::JoinAsof;
 #[cfg(feature = "pivot")]
 use crate::frame::groupby::pivot::*;
 use crate::frame::groupby::*;
@@ -37,6 +42,7 @@ use ahash::RandomState;
 use arrow::array::ArrayRef;
 use std::borrow::Cow;
 use std::ops::Deref;
+use std::ops::{BitAnd, BitOr, BitXor};
 
 // Utility wrapper struct
 pub(crate) struct SeriesWrap<T>(pub T);
@@ -64,53 +70,44 @@ macro_rules! impl_dyn_series {
         }
 
         impl private::PrivateSeries for SeriesWrap<$ca> {
+            fn _field(&self) -> Cow<Field> {
+                Cow::Borrowed(self.0.ref_field())
+            }
+
+            fn _dtype(&self) -> &DataType {
+                self.0.ref_field().data_type()
+            }
+
             fn explode_by_offsets(&self, offsets: &[i64]) -> Series {
                 self.0.explode_by_offsets(offsets)
             }
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_sum(&self, options: RollingOptions) -> Result<Series> {
+                self.0.rolling_sum(options)
+            }
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_min(&self, options: RollingOptions) -> Result<Series> {
+                self.0.rolling_min(options)
+            }
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_max(&self, options: RollingOptions) -> Result<Series> {
+                self.0.rolling_max(options)
+            }
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_std(&self, options: RollingOptions) -> Result<Series> {
+                let s = self.cast(&DataType::Float64).unwrap();
+                s.f64().unwrap().rolling_std(options)
+            }
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_mean(&self, options: RollingOptions) -> Result<Series> {
+                let s = self.cast(&DataType::Float64).unwrap();
+                s.f64().unwrap().rolling_mean(options)
+            }
 
             #[cfg(feature = "rolling_window")]
-            fn _rolling_mean(
-                &self,
-                window_size: u32,
-                weight: Option<&[f64]>,
-                ignore_null: bool,
-                min_periods: u32,
-            ) -> Result<Series> {
-                ChunkWindow::rolling_mean(&self.0, window_size, weight, ignore_null, min_periods)
-                    .map(|ca| ca.into_series())
-            }
-            #[cfg(feature = "rolling_window")]
-            fn _rolling_sum(
-                &self,
-                window_size: u32,
-                weight: Option<&[f64]>,
-                ignore_null: bool,
-                min_periods: u32,
-            ) -> Result<Series> {
-                ChunkWindow::rolling_sum(&self.0, window_size, weight, ignore_null, min_periods)
-                    .map(|ca| ca.into_series())
-            }
-            #[cfg(feature = "rolling_window")]
-            fn _rolling_min(
-                &self,
-                window_size: u32,
-                weight: Option<&[f64]>,
-                ignore_null: bool,
-                min_periods: u32,
-            ) -> Result<Series> {
-                ChunkWindow::rolling_min(&self.0, window_size, weight, ignore_null, min_periods)
-                    .map(|ca| ca.into_series())
-            }
-            #[cfg(feature = "rolling_window")]
-            fn _rolling_max(
-                &self,
-                window_size: u32,
-                weight: Option<&[f64]>,
-                ignore_null: bool,
-                min_periods: u32,
-            ) -> Result<Series> {
-                ChunkWindow::rolling_max(&self.0, window_size, weight, ignore_null, min_periods)
-                    .map(|ca| ca.into_series())
+            fn _rolling_var(&self, options: RollingOptions) -> Result<Series> {
+                let s = self.cast(&DataType::Float64).unwrap();
+                s.f64().unwrap().rolling_var(options)
             }
 
             #[cfg(feature = "cum_agg")]
@@ -128,9 +125,14 @@ macro_rules! impl_dyn_series {
                 self.0.cumsum(reverse).into_series()
             }
 
+            #[cfg(feature = "cum_agg")]
+            fn _cumprod(&self, reverse: bool) -> Series {
+                self.0.cumprod(reverse).into_series()
+            }
+
             #[cfg(feature = "asof_join")]
             fn join_asof(&self, other: &Series) -> Result<Vec<Option<u32>>> {
-                self.0.join_asof(other.as_ref().as_ref())
+                self.0.join_asof(other)
             }
 
             fn set_sorted(&mut self, reverse: bool) {
@@ -299,6 +301,36 @@ macro_rules! impl_dyn_series {
                 self.0.interpolate().into_series()
             }
 
+            fn bitand(&self, other: &Series) -> Result<Series> {
+                let other = if other.len() == 1 {
+                    Cow::Owned(other.cast(self.dtype())?)
+                } else {
+                    Cow::Borrowed(other)
+                };
+                let other = self.0.unpack_series_matching_type(&other)?;
+                Ok(self.0.bitand(&other).into_series())
+            }
+
+            fn bitor(&self, other: &Series) -> Result<Series> {
+                let other = if other.len() == 1 {
+                    Cow::Owned(other.cast(self.dtype())?)
+                } else {
+                    Cow::Borrowed(other)
+                };
+                let other = self.0.unpack_series_matching_type(&other)?;
+                Ok(self.0.bitor(&other).into_series())
+            }
+
+            fn bitxor(&self, other: &Series) -> Result<Series> {
+                let other = if other.len() == 1 {
+                    Cow::Owned(other.cast(self.dtype())?)
+                } else {
+                    Cow::Borrowed(other)
+                };
+                let other = self.0.unpack_series_matching_type(&other)?;
+                Ok(self.0.bitxor(&other).into_series())
+            }
+
             fn rename(&mut self, name: &str) {
                 self.0.rename(name);
             }
@@ -310,10 +342,6 @@ macro_rules! impl_dyn_series {
                 self.0.name()
             }
 
-            fn field(&self) -> &Field {
-                self.0.ref_field()
-            }
-
             fn chunks(&self) -> &Vec<ArrayRef> {
                 self.0.chunks()
             }
@@ -325,7 +353,7 @@ macro_rules! impl_dyn_series {
                 if matches!(self.0.dtype(), DataType::Int8) {
                     unsafe { Ok(&*(self as *const dyn SeriesTrait as *const Int8Chunked)) }
                 } else {
-                    Err(PolarsError::DataTypeMisMatch(
+                    Err(PolarsError::SchemaMisMatch(
                         format!(
                             "cannot unpack Series: {:?} of type {:?} into i8",
                             self.name(),
@@ -341,7 +369,7 @@ macro_rules! impl_dyn_series {
                 if matches!(self.0.dtype(), DataType::Int16) {
                     unsafe { Ok(&*(self as *const dyn SeriesTrait as *const Int16Chunked)) }
                 } else {
-                    Err(PolarsError::DataTypeMisMatch(
+                    Err(PolarsError::SchemaMisMatch(
                         format!(
                             "cannot unpack Series: {:?} of type {:?} into i16",
                             self.name(),
@@ -356,7 +384,7 @@ macro_rules! impl_dyn_series {
                 if matches!(self.0.dtype(), DataType::Int32) {
                     unsafe { Ok(&*(self as *const dyn SeriesTrait as *const Int32Chunked)) }
                 } else {
-                    Err(PolarsError::DataTypeMisMatch(
+                    Err(PolarsError::SchemaMisMatch(
                         format!(
                             "cannot unpack Series: {:?} of type {:?} into i32",
                             self.name(),
@@ -371,7 +399,7 @@ macro_rules! impl_dyn_series {
                 if matches!(self.0.dtype(), DataType::Int64) {
                     unsafe { Ok(&*(self as *const dyn SeriesTrait as *const Int64Chunked)) }
                 } else {
-                    Err(PolarsError::DataTypeMisMatch(
+                    Err(PolarsError::SchemaMisMatch(
                         format!(
                             "cannot unpack Series: {:?} of type {:?} into i64",
                             self.name(),
@@ -386,7 +414,7 @@ macro_rules! impl_dyn_series {
                 if matches!(self.0.dtype(), DataType::Float32) {
                     unsafe { Ok(&*(self as *const dyn SeriesTrait as *const Float32Chunked)) }
                 } else {
-                    Err(PolarsError::DataTypeMisMatch(
+                    Err(PolarsError::SchemaMisMatch(
                         format!(
                             "cannot unpack Series: {:?} of type {:?} into f32",
                             self.name(),
@@ -401,7 +429,7 @@ macro_rules! impl_dyn_series {
                 if matches!(self.0.dtype(), DataType::Float64) {
                     unsafe { Ok(&*(self as *const dyn SeriesTrait as *const Float64Chunked)) }
                 } else {
-                    Err(PolarsError::DataTypeMisMatch(
+                    Err(PolarsError::SchemaMisMatch(
                         format!(
                             "cannot unpack Series: {:?} of type {:?} into f64",
                             self.name(),
@@ -416,7 +444,7 @@ macro_rules! impl_dyn_series {
                 if matches!(self.0.dtype(), DataType::UInt8) {
                     unsafe { Ok(&*(self as *const dyn SeriesTrait as *const UInt8Chunked)) }
                 } else {
-                    Err(PolarsError::DataTypeMisMatch(
+                    Err(PolarsError::SchemaMisMatch(
                         format!(
                             "cannot unpack Series: {:?} of type {:?} into u8",
                             self.name(),
@@ -431,7 +459,7 @@ macro_rules! impl_dyn_series {
                 if matches!(self.0.dtype(), DataType::UInt16) {
                     unsafe { Ok(&*(self as *const dyn SeriesTrait as *const UInt16Chunked)) }
                 } else {
-                    Err(PolarsError::DataTypeMisMatch(
+                    Err(PolarsError::SchemaMisMatch(
                         format!(
                             "cannot unpack Series: {:?} of type {:?} into u16",
                             self.name(),
@@ -446,7 +474,7 @@ macro_rules! impl_dyn_series {
                 if matches!(self.0.dtype(), DataType::UInt32) {
                     unsafe { Ok(&*(self as *const dyn SeriesTrait as *const UInt32Chunked)) }
                 } else {
-                    Err(PolarsError::DataTypeMisMatch(
+                    Err(PolarsError::SchemaMisMatch(
                         format!(
                             "cannot unpack Series: {:?} of type {:?} into u32",
                             self.name(),
@@ -461,7 +489,7 @@ macro_rules! impl_dyn_series {
                 if matches!(self.0.dtype(), DataType::UInt64) {
                     unsafe { Ok(&*(self as *const dyn SeriesTrait as *const UInt64Chunked)) }
                 } else {
-                    Err(PolarsError::DataTypeMisMatch(
+                    Err(PolarsError::SchemaMisMatch(
                         format!(
                             "cannot unpack Series: {:?} of type {:?} into u64",
                             self.name(),
@@ -472,147 +500,6 @@ macro_rules! impl_dyn_series {
                 }
             }
 
-            fn bool(&self) -> Result<&BooleanChunked> {
-                if matches!(self.0.dtype(), DataType::Boolean) {
-                    unsafe { Ok(&*(self as *const dyn SeriesTrait as *const BooleanChunked)) }
-                } else {
-                    Err(PolarsError::DataTypeMisMatch(
-                        format!(
-                            "cannot unpack Series: {:?} of type {:?} into bool",
-                            self.name(),
-                            self.dtype(),
-                        )
-                        .into(),
-                    ))
-                }
-            }
-
-            fn utf8(&self) -> Result<&Utf8Chunked> {
-                if matches!(self.0.dtype(), DataType::Utf8) {
-                    unsafe { Ok(&*(self as *const dyn SeriesTrait as *const Utf8Chunked)) }
-                } else {
-                    Err(PolarsError::DataTypeMisMatch(
-                        format!(
-                            "cannot unpack Series: {:?} of type {:?} into utf8",
-                            self.name(),
-                            self.dtype(),
-                        )
-                        .into(),
-                    ))
-                }
-            }
-
-            fn date32(&self) -> Result<&Date32Chunked> {
-                if matches!(self.0.dtype(), DataType::Date32) {
-                    unsafe { Ok(&*(self as *const dyn SeriesTrait as *const Date32Chunked)) }
-                } else {
-                    Err(PolarsError::DataTypeMisMatch(
-                        format!(
-                            "cannot unpack Series: {:?} of type {:?} into date32",
-                            self.name(),
-                            self.dtype(),
-                        )
-                        .into(),
-                    ))
-                }
-            }
-
-            fn date64(&self) -> Result<&Date64Chunked> {
-                if matches!(self.0.dtype(), DataType::Date64) {
-                    unsafe { Ok(&*(self as *const dyn SeriesTrait as *const Date64Chunked)) }
-                } else {
-                    Err(PolarsError::DataTypeMisMatch(
-                        format!(
-                            "cannot unpack Series: {:?} of type {:?} into date64",
-                            self.name(),
-                            self.dtype(),
-                        )
-                        .into(),
-                    ))
-                }
-            }
-
-            fn time64_nanosecond(&self) -> Result<&Time64NanosecondChunked> {
-                if matches!(self.0.dtype(), DataType::Time64(TimeUnit::Nanosecond)) {
-                    unsafe {
-                        Ok(&*(self as *const dyn SeriesTrait as *const Time64NanosecondChunked))
-                    }
-                } else {
-                    Err(PolarsError::DataTypeMisMatch(
-                        format!(
-                            "cannot unpack Series: {:?} of type {:?} into time64",
-                            self.name(),
-                            self.dtype(),
-                        )
-                        .into(),
-                    ))
-                }
-            }
-
-            fn duration_nanosecond(&self) -> Result<&DurationNanosecondChunked> {
-                if matches!(self.0.dtype(), DataType::Duration(TimeUnit::Nanosecond)) {
-                    unsafe {
-                        Ok(&*(self as *const dyn SeriesTrait as *const DurationNanosecondChunked))
-                    }
-                } else {
-                    Err(PolarsError::DataTypeMisMatch(
-                        format!(
-                            "cannot unpack Series: {:?} of type {:?} into duration_nanosecond",
-                            self.name(),
-                            self.dtype(),
-                        )
-                        .into(),
-                    ))
-                }
-            }
-
-            fn duration_millisecond(&self) -> Result<&DurationMillisecondChunked> {
-                if matches!(self.0.dtype(), DataType::Duration(TimeUnit::Millisecond)) {
-                    unsafe {
-                        Ok(&*(self as *const dyn SeriesTrait as *const DurationMillisecondChunked))
-                    }
-                } else {
-                    Err(PolarsError::DataTypeMisMatch(
-                        format!(
-                            "cannot unpack Series: {:?} of type {:?} into duration_millisecond",
-                            self.name(),
-                            self.dtype(),
-                        )
-                        .into(),
-                    ))
-                }
-            }
-
-            fn list(&self) -> Result<&ListChunked> {
-                if matches!(self.0.dtype(), DataType::List(_)) {
-                    unsafe { Ok(&*(self as *const dyn SeriesTrait as *const ListChunked)) }
-                } else {
-                    Err(PolarsError::DataTypeMisMatch(
-                        format!(
-                            "cannot unpack Series: {:?} of type {:?} into list",
-                            self.name(),
-                            self.dtype(),
-                        )
-                        .into(),
-                    ))
-                }
-            }
-
-            fn categorical(&self) -> Result<&CategoricalChunked> {
-                if matches!(self.0.dtype(), DataType::Categorical) {
-                    unsafe { Ok(&*(self as *const dyn SeriesTrait as *const CategoricalChunked)) }
-                } else {
-                    Err(PolarsError::DataTypeMisMatch(
-                        format!(
-                            "cannot unpack Series: {:?} of type {:?} into categorical",
-                            self.name(),
-                            self.dtype(),
-                        )
-                        .into(),
-                    ))
-                }
-            }
-
             fn append_array(&mut self, other: ArrayRef) -> Result<()> {
                 self.0.append_array(other)
             }
@@ -627,7 +514,7 @@ macro_rules! impl_dyn_series {
                     self.0.append(other.as_ref().as_ref());
                     Ok(())
                 } else {
-                    Err(PolarsError::DataTypeMisMatch(
+                    Err(PolarsError::SchemaMisMatch(
                         "cannot append Series; data types don't match".into(),
                     ))
                 }
@@ -704,8 +591,8 @@ macro_rules! impl_dyn_series {
                 ChunkExpandAtIndex::expand_at_index(&self.0, index, length).into_series()
             }
 
-            fn cast_with_dtype(&self, data_type: &DataType) -> Result<Series> {
-                self.0.cast_with_dtype(data_type)
+            fn cast(&self, data_type: &DataType) -> Result<Series> {
+                self.0.cast(data_type)
             }
 
             fn to_dummies(&self) -> Result<DataFrame> {
@@ -741,6 +628,10 @@ macro_rules! impl_dyn_series {
                 self.0.null_count()
             }
 
+            fn has_validity(&self) -> bool {
+                self.0.has_validity()
+            }
+
             fn unique(&self) -> Result<Series> {
                 ChunkUnique::unique(&self.0).map(|ca| ca.into_series())
             }
@@ -761,11 +652,6 @@ macro_rules! impl_dyn_series {
                 ArgAgg::arg_max(&self.0)
             }
 
-            fn arg_true(&self) -> Result<UInt32Chunked> {
-                let ca: &BooleanChunked = self.bool()?;
-                Ok(ca.arg_true())
-            }
-
             fn is_null(&self) -> BooleanChunked {
                 self.0.is_null()
             }
@@ -895,18 +781,20 @@ macro_rules! impl_dyn_series {
             fn mode(&self) -> Result<Series> {
                 Ok(self.0.mode()?.into_series())
             }
+
+            #[cfg(feature = "concat_str")]
+            fn str_concat(&self, delimiter: &str) -> Utf8Chunked {
+                self.0.str_concat(delimiter)
+            }
         }
     };
 }
 
-impl_dyn_series!(Float32Chunked);
-impl_dyn_series!(Float64Chunked);
 #[cfg(feature = "dtype-u8")]
 impl_dyn_series!(UInt8Chunked);
 #[cfg(feature = "dtype-u16")]
 impl_dyn_series!(UInt16Chunked);
 impl_dyn_series!(UInt32Chunked);
-#[cfg(feature = "dtype-u64")]
 impl_dyn_series!(UInt64Chunked);
 #[cfg(feature = "dtype-i8")]
 impl_dyn_series!(Int8Chunked);
@@ -914,10 +802,6 @@ impl_dyn_series!(Int8Chunked);
 impl_dyn_series!(Int16Chunked);
 impl_dyn_series!(Int32Chunked);
 impl_dyn_series!(Int64Chunked);
-impl_dyn_series!(CategoricalChunked);
-impl_dyn_series!(Utf8Chunked);
-impl_dyn_series!(ListChunked);
-impl_dyn_series!(BooleanChunked);
 
 macro_rules! impl_dyn_series_numeric {
     ($ca: ident) => {
@@ -942,7 +826,6 @@ impl_dyn_series_numeric!(UInt8Chunked);
 #[cfg(feature = "dtype-u16")]
 impl_dyn_series_numeric!(UInt16Chunked);
 impl_dyn_series_numeric!(UInt32Chunked);
-#[cfg(feature = "dtype-u64")]
 impl_dyn_series_numeric!(UInt64Chunked);
 #[cfg(feature = "dtype-i8")]
 impl_dyn_series_numeric!(Int8Chunked);
@@ -950,8 +833,7 @@ impl_dyn_series_numeric!(Int8Chunked);
 impl_dyn_series_numeric!(Int16Chunked);
 impl_dyn_series_numeric!(Int32Chunked);
 impl_dyn_series_numeric!(Int64Chunked);
-impl_dyn_series_numeric!(CategoricalChunked);
 
-impl private::PrivateSeriesNumeric for SeriesWrap<BooleanChunked> {}
 impl private::PrivateSeriesNumeric for SeriesWrap<Utf8Chunked> {}
 impl private::PrivateSeriesNumeric for SeriesWrap<ListChunked> {}
+impl private::PrivateSeriesNumeric for SeriesWrap<BooleanChunked> {}
