@@ -31,12 +31,19 @@ impl<T> PrivateSeries for SeriesWrap<ObjectChunked<T>>
 where
     T: PolarsObject,
 {
-    fn agg_first(&self, _groups: &[(u32, Vec<u32>)]) -> Series {
-        self.0.agg_first(_groups)
+    fn _field(&self) -> Cow<Field> {
+        Cow::Borrowed(self.0.ref_field())
+    }
+    fn agg_first(&self, groups: &[(u32, Vec<u32>)]) -> Series {
+        self.0.agg_first(groups)
     }
 
-    fn agg_last(&self, _groups: &[(u32, Vec<u32>)]) -> Series {
-        self.0.agg_last(_groups)
+    fn agg_last(&self, groups: &[(u32, Vec<u32>)]) -> Series {
+        self.0.agg_last(groups)
+    }
+
+    fn agg_list(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        self.0.agg_list(groups)
     }
 
     fn str_value(&self, index: usize) -> Cow<str> {
@@ -69,7 +76,7 @@ where
 {
     #[cfg(feature = "interpolate")]
     fn interpolate(&self) -> Series {
-        self.0.interpolate().into_series()
+        self.0.clone().into_series()
     }
 
     fn rename(&mut self, name: &str) {
@@ -84,10 +91,6 @@ where
         ObjectChunked::name(&self.0)
     }
 
-    fn field(&self) -> &Field {
-        ObjectChunked::ref_field(&self.0)
-    }
-
     fn dtype(&self) -> &DataType {
         ObjectChunked::dtype(&self.0)
     }
@@ -109,7 +112,7 @@ where
             ObjectChunked::append(&mut self.0, other.as_ref().as_ref());
             Ok(())
         } else {
-            Err(PolarsError::DataTypeMisMatch(
+            Err(PolarsError::SchemaMisMatch(
                 "cannot append Series; data types don't match".into(),
             ))
         }
@@ -169,16 +172,12 @@ where
         ChunkExpandAtIndex::expand_at_index(&self.0, index, length).into_series()
     }
 
-    fn cast_with_dtype(&self, _data_type: &DataType) -> Result<Series> {
+    fn cast(&self, _data_type: &DataType) -> Result<Series> {
         Err(PolarsError::InvalidOperation(
             "cannot cast array of type ObjectChunked to arrow datatype".into(),
         ))
     }
 
-    fn to_dummies(&self) -> Result<DataFrame> {
-        ToDummies::to_dummies(&self.0)
-    }
-
     fn value_counts(&self) -> Result<DataFrame> {
         ChunkUnique::value_counts(&self.0)
     }
@@ -186,23 +185,14 @@ where
     fn get(&self, index: usize) -> AnyValue {
         ObjectChunked::get_any_value(&self.0, index)
     }
-
-    fn sort_in_place(&mut self, reverse: bool) {
-        ChunkSort::sort_in_place(&mut self.0, reverse)
-    }
-
-    fn sort(&self, reverse: bool) -> Series {
-        ChunkSort::sort(&self.0, reverse).into_series()
-    }
-
-    fn argsort(&self, reverse: bool) -> UInt32Chunked {
-        ChunkSort::argsort(&self.0, reverse)
-    }
-
     fn null_count(&self) -> usize {
         ObjectChunked::null_count(&self.0)
     }
 
+    fn has_validity(&self) -> bool {
+        ObjectChunked::has_validity(&self.0)
+    }
+
     fn unique(&self) -> Result<Series> {
         ChunkUnique::unique(&self.0).map(|ca| ca.into_series())
     }
