@@ -1,10 +1,11 @@
 use crate::datatypes::UInt64Chunked;
 use crate::prelude::*;
+use crate::utils::arrow::array::Array;
 use crate::POOL;
-use ahash::RandomState;
-use arrow::array::ArrayRef;
+use ahash::{CallHasher, RandomState};
+use arrow::bitmap::utils::get_bit_unchecked;
 use hashbrown::{hash_map::RawEntryMut, HashMap};
-use itertools::Itertools;
+use polars_arrow::utils::CustomIterTools;
 use rayon::prelude::*;
 use std::convert::TryInto;
 use std::hash::{BuildHasher, BuildHasherDefault, Hash, Hasher};
@@ -26,10 +27,21 @@ pub trait VecHash {
     }
 }
 
+pub(crate) fn get_null_hash_value(random_state: RandomState) -> u64 {
+    // we just start with a large prime number and hash that twice
+    // to get a constant hash value for null/None
+    let mut hasher = random_state.build_hasher();
+    3188347919usize.hash(&mut hasher);
+    let first = hasher.finish();
+    let mut hasher = random_state.build_hasher();
+    first.hash(&mut hasher);
+    hasher.finish()
+}
+
 impl<T> VecHash for ChunkedArray<T>
 where
     T: PolarsIntegerType,
-    T::Native: Hash,
+    T::Native: Hash + CallHasher,
 {
     fn vec_hash(&self, random_state: RandomState) -> AlignedVec<u64> {
         // Note that we don't use the no null branch! This can break in unexpected ways.
@@ -41,46 +53,91 @@ where
         let mut av = AlignedVec::with_capacity(self.len());
 
         self.downcast_iter().for_each(|arr| {
-            av.extend(arr.into_iter().map(|opt_v| {
-                let mut hasher = random_state.build_hasher();
-                opt_v.hash(&mut hasher);
-                hasher.finish()
-            }))
+            av.extend(
+                arr.values()
+                    .as_slice()
+                    .iter()
+                    .map(|v| T::Native::get_hash(v, &random_state)),
+            );
         });
+
+        let null_h = get_null_hash_value(random_state);
+        let hashes = av.as_mut_slice();
+
+        let mut offset = 0;
+        self.downcast_iter().for_each(|arr| {
+            if let Some(validity) = arr.validity() {
+                let slice = validity.as_slice().0;
+                (0..validity.len())
+                    .map(|i| unsafe { get_bit_unchecked(slice, i) })
+                    .zip(&mut hashes[offset..])
+                    .for_each(|(valid, h)| {
+                        if !valid {
+                            *h = null_h;
+                        }
+                    })
+            }
+            offset += arr.len();
+        });
+
         av
     }
 
     fn vec_hash_combine(&self, random_state: RandomState, hashes: &mut [u64]) {
-        self.apply_to_slice(
-            |opt_v, h| {
-                let mut hasher = random_state.build_hasher();
-                opt_v.hash(&mut hasher);
-                boost_hash_combine(hasher.finish(), *h)
-            },
-            hashes,
-        )
+        let null_h = get_null_hash_value(random_state.clone());
+
+        let mut offset = 0;
+        self.downcast_iter().for_each(|arr| {
+            match arr.null_count() {
+                0 => arr
+                    .values()
+                    .as_slice()
+                    .iter()
+                    .zip(&mut hashes[offset..])
+                    .for_each(|(v, h)| {
+                        let l = T::Native::get_hash(v, &random_state);
+                        *h = boost_hash_combine(l, *h)
+                    }),
+                _ => arr
+                    .iter()
+                    .zip(&mut hashes[offset..])
+                    .for_each(|(opt_v, h)| match opt_v {
+                        Some(v) => {
+                            let l = T::Native::get_hash(v, &random_state);
+                            *h = boost_hash_combine(l, *h)
+                        }
+                        None => {
+                            *h = boost_hash_combine(null_h, *h);
+                        }
+                    }),
+            }
+            offset += arr.len();
+        });
     }
 }
 
 impl VecHash for Utf8Chunked {
     fn vec_hash(&self, random_state: RandomState) -> AlignedVec<u64> {
+        let null_h = get_null_hash_value(random_state.clone());
         let mut av = AlignedVec::with_capacity(self.len());
         self.downcast_iter().for_each(|arr| {
-            av.extend(arr.into_iter().map(|opt_v| {
-                let mut hasher = random_state.build_hasher();
-                opt_v.hash(&mut hasher);
-                hasher.finish()
+            av.extend(arr.into_iter().map(|opt_v| match opt_v {
+                Some(v) => str::get_hash(v, &random_state),
+                None => null_h,
             }))
         });
         av
     }
 
     fn vec_hash_combine(&self, random_state: RandomState, hashes: &mut [u64]) {
+        let null_h = get_null_hash_value(random_state.clone());
         self.apply_to_slice(
             |opt_v, h| {
-                let mut hasher = random_state.build_hasher();
-                opt_v.hash(&mut hasher);
-                boost_hash_combine(hasher.finish(), *h)
+                let l = match opt_v {
+                    Some(v) => str::get_hash(v, &random_state),
+                    None => null_h,
+                };
+                boost_hash_combine(l, *h)
             },
             hashes,
         )
@@ -113,35 +170,20 @@ impl VecHash for BooleanChunked {
 }
 
 impl VecHash for Float32Chunked {
-    fn vec_hash(&self, _random_state: RandomState) -> AlignedVec<u64> {
-        panic!("should already be coerced to u32")
+    fn vec_hash(&self, random_state: RandomState) -> AlignedVec<u64> {
+        self.bit_repr_small().vec_hash(random_state)
     }
 
-    fn vec_hash_combine(&self, _random_state: RandomState, _hashes: &mut [u64]) {
-        panic!("should already be coerced to u32")
+    fn vec_hash_combine(&self, random_state: RandomState, hashes: &mut [u64]) {
+        self.bit_repr_small().vec_hash_combine(random_state, hashes)
     }
 }
 impl VecHash for Float64Chunked {
-    fn vec_hash(&self, _random_state: RandomState) -> AlignedVec<u64> {
-        #[cfg(not(feature = "dtype-u64"))]
-        {
-            let mut av = AlignedVec::with_capacity(self.len());
-            self.downcast_iter().for_each(|arr| {
-                av.extend(arr.into_iter().map(|opt_v| {
-                    let mut hasher = _random_state.build_hasher();
-                    opt_v.map(|v| v.to_bits()).hash(&mut hasher);
-                    hasher.finish()
-                }))
-            });
-            av
-        }
-        #[cfg(feature = "dtype-u64")]
-        {
-            panic!("should already be coerced to u64")
-        }
+    fn vec_hash(&self, random_state: RandomState) -> AlignedVec<u64> {
+        self.bit_repr_large().vec_hash(random_state)
     }
-    fn vec_hash_combine(&self, _random_state: RandomState, _hashes: &mut [u64]) {
-        panic!("should already be coerced to u64")
+    fn vec_hash_combine(&self, random_state: RandomState, hashes: &mut [u64]) {
+        self.bit_repr_large().vec_hash_combine(random_state, hashes)
     }
 }
 
@@ -206,15 +248,15 @@ impl AsU64 for Option<u32> {
     fn as_u64(self) -> u64 {
         match self {
             Some(v) => v as u64,
-            // just a number
-            None => u64::MAX,
+            // just a number safe from overflow
+            None => u64::MAX >> 2,
         }
     }
 }
 
 impl AsU64 for Option<u64> {
     fn as_u64(self) -> u64 {
-        self.unwrap_or(u64::MAX)
+        self.unwrap_or(u64::MAX >> 2)
     }
 }
 
@@ -228,20 +270,17 @@ impl AsU64 for [u8; 9] {
 const BUILD_HASHER: RandomState = RandomState::with_seeds(0, 0, 0, 0);
 impl AsU64 for [u8; 17] {
     fn as_u64(self) -> u64 {
-        let mut h = BUILD_HASHER.build_hasher();
-        self.hash(&mut h);
-        h.finish()
+        <[u8]>::get_hash(&self, &BUILD_HASHER)
     }
 }
 
 impl AsU64 for [u8; 13] {
     fn as_u64(self) -> u64 {
-        let mut h = BUILD_HASHER.build_hasher();
-        self.hash(&mut h);
-        h.finish()
+        <[u8]>::get_hash(&self, &BUILD_HASHER)
     }
 }
 
+#[derive(Default)]
 pub struct IdHasher {
     hash: u64,
 }
@@ -276,12 +315,6 @@ impl Hasher for IdHasher {
     }
 }
 
-impl Default for IdHasher {
-    fn default() -> Self {
-        IdHasher { hash: 0 }
-    }
-}
-
 pub type IdBuildHasher = BuildHasherDefault<IdHasher>;
 
 #[derive(Debug)]
@@ -341,95 +374,37 @@ impl<'a> AsU64 for StrHash<'a> {
     }
 }
 
-impl From<u64> for NThreads {
-    fn from(n: u64) -> Self {
-        use NThreads::*;
-        match n {
-            2 => T2,
-            4 => T4,
-            8 => T8,
-            10 => T10,
-            12 => T12,
-            16 => T16,
-            20 => T20,
-            32 => T32,
-            40 => T40,
-            64 => T64,
-            _ => Other(n),
-        }
-    }
-}
-
-#[derive(Clone, Copy)]
-pub enum NThreads {
-    T2,
-    T4,
-    T8,
-    T10,
-    T12,
-    T16,
-    T20,
-    T32,
-    T40,
-    T64,
-    Other(u64),
-}
-
-impl NThreads {
-    #[inline]
-    fn modulo(&self, h: u64) -> u64 {
-        use NThreads::*;
-        match self {
-            T2 => h % 2,
-            T4 => h % 4,
-            T8 => h % 8,
-            T10 => h % 10,
-            T12 => h % 12,
-            T16 => h % 16,
-            T20 => h % 20,
-            T32 => h % 32,
-            T40 => h % 20,
-            T64 => h % 64,
-            Other(n) => h % n,
-        }
-    }
-}
-
-/// Check if a hash should be processed in that thread.
-#[inline]
-pub fn this_thread(h: u64, thread_no: u64, n_threads: NThreads) -> bool {
-    n_threads.modulo(h + thread_no) == 0
-}
-
 #[inline]
 /// For partitions that are a power of 2 we can use a bitshift instead of a modulo.
 pub(crate) fn this_partition(h: u64, thread_no: u64, n_partitions: u64) -> bool {
     // n % 2^i = n & (2^i - 1)
-    (h + thread_no) & (n_partitions - 1) == 0
+    (h.wrapping_add(thread_no)) & n_partitions.wrapping_sub(1) == 0
 }
 
 pub(crate) fn prepare_hashed_relation_threaded<T, I>(
     iters: Vec<I>,
-) -> Vec<HashMap<T, Vec<u32>, RandomState>>
+) -> Vec<HashMap<T, (bool, Vec<u32>), RandomState>>
 where
-    I: Iterator<Item = T> + Send,
+    I: Iterator<Item = T> + Send + TrustedLen,
     T: Send + Hash + Eq + Sync + Copy,
 {
-    let n_threads = iters.len();
+    let n_partitions = iters.len();
+    assert!(n_partitions.is_power_of_two());
+
     let (hashes_and_keys, build_hasher) = create_hash_and_keys_threaded_vectorized(iters, None);
 
     // We will create a hashtable in every thread.
     // We use the hash to partition the keys to the matching hashtable.
     // Every thread traverses all keys/hashes and ignores the ones that doesn't fall in that partition.
     POOL.install(|| {
-        (0..n_threads).into_par_iter().map(|thread_no| {
+        (0..n_partitions).into_par_iter().map(|partition_no| {
             let build_hasher = build_hasher.clone();
             let hashes_and_keys = &hashes_and_keys;
-            let thread_no = thread_no as u64;
-            let mut hash_tbl: HashMap<T, Vec<u32>, RandomState> =
+            let partition_no = partition_no as u64;
+            let mut hash_tbl: HashMap<T, (bool, Vec<u32>), RandomState> =
                 HashMap::with_hasher(build_hasher);
 
-            let n_threads = (n_threads as u64).into();
+            let n_threads = n_partitions as u64;
             let mut offset = 0;
             for hashes_and_keys in hashes_and_keys {
                 let len = hashes_and_keys.len();
@@ -440,7 +415,7 @@ where
                         let idx = idx as u32;
                         // partition hashes by thread no.
                         // So only a part of the hashes go to this hashmap
-                        if this_thread(*h, thread_no, n_threads) {
+                        if this_partition(*h, partition_no, n_threads) {
                             let idx = idx + offset;
                             let entry = hash_tbl
                                 .raw_entry_mut()
@@ -449,11 +424,11 @@ where
 
                             match entry {
                                 RawEntryMut::Vacant(entry) => {
-                                    entry.insert_hashed_nocheck(*h, *k, vec![idx]);
+                                    entry.insert_hashed_nocheck(*h, *k, (false, vec![idx]));
                                 }
                                 RawEntryMut::Occupied(mut entry) => {
                                     let (_k, v) = entry.get_key_value_mut();
-                                    v.push(idx);
+                                    v.1.push(idx);
                                 }
                             }
                         }
@@ -473,6 +448,7 @@ pub(crate) fn create_hash_and_keys_threaded_vectorized<I, T>(
 ) -> (Vec<Vec<(u64, T)>>, RandomState)
 where
     I: IntoIterator<Item = T> + Send,
+    I::IntoIter: TrustedLen,
     T: Send + Hash + Eq,
 {
     let build_hasher = build_hasher.unwrap_or_default();
@@ -487,7 +463,7 @@ where
                         val.hash(&mut hasher);
                         (hasher.finish(), val)
                     })
-                    .collect_vec()
+                    .collect_trusted::<Vec<_>>()
             })
             .collect()
     });
@@ -496,7 +472,7 @@ where
 
 // hash combine from c++' boost lib
 #[inline]
-fn boost_hash_combine(l: u64, r: u64) -> u64 {
+pub(crate) fn boost_hash_combine(l: u64, r: u64) -> u64 {
     l ^ r.wrapping_add(0x9e3779b9u64.wrapping_add(l << 6).wrapping_add(r >> 2))
 }
 
@@ -533,6 +509,10 @@ pub(crate) fn df_rows_to_hashes(
         keys.vec_hash_combine(build_hasher.clone(), hslice);
     }
 
-    let chunks = vec![Arc::new(hashes.into_primitive_array::<UInt64Type>(None)) as ArrayRef];
+    let chunks = vec![Arc::new(PrimitiveArray::from_data(
+        ArrowDataType::UInt64,
+        hashes.into(),
+        None,
+    )) as ArrayRef];
     (UInt64Chunked::new_from_chunks("", chunks), build_hasher)
 }
